@@ -652,14 +652,14 @@ module Senna
       proc { Senna::sen_values_close(handle) }
     end
 
-    def Values::open(str=nil, weight=1)
+    def Values::open(str=nil, weight=0)
       handle = Senna::sen_values_open
       ret = new(handle)
       ret.add(str, weight) unless str.nil?
       return ret
     end
 
-    def add(str, weight=1)
+    def add(str, weight=0)
       Senna::sen_values_add(@handle, str, str.length, weight)
     end
 
@@ -789,7 +789,7 @@ module Senna
       return new(handle, db)
     end
 
-    def Ctx::create(path, flags, encoding)
+    def Ctx::create(path, flags=0, encoding=ENC_DEFAULT)
       db = Senna::sen_db_create(path, flags, encoding)
       raise "sen_db_create(#{path}) failed" unless db
       handle = Senna::sen_ctx_open(db, 1)
