@@ -15,11 +15,12 @@
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 #include "senna_in.h"
+#include <stdio.h>
 #include <fcntl.h>
 #include <string.h>
 #include <sys/stat.h>
 
-#include "str.h"
+#include "ctx.h"
 #include "sym.h"
 #include "inv.h"
 
@@ -57,7 +58,8 @@ struct sen_inv_header {
   uint32_t total_chunk_size;
   uint32_t amax;
   uint32_t bmax;
-  uint32_t reserved[8];
+  uint32_t smax;
+  uint32_t reserved[7];
   uint16_t ainfo[SEN_INV_MAX_SEGMENT];
   uint16_t binfo[SEN_INV_MAX_SEGMENT];
   uint8_t chunks[1]; /* dummy */
@@ -139,7 +141,7 @@ buffer_segment_new(sen_inv *inv, uint16_t *segno)
   pseg = segment_get(inv);
   if (pseg < SEN_INV_MAX_SEGMENT) {
     inv->header->binfo[lseg] = pseg;
-    if (lseg > inv->header->bmax) { inv->header->bmax = lseg; }
+    if (lseg >= inv->header->bmax) { inv->header->bmax = lseg + 1; }
     return sen_success;
   } else {
     return sen_memory_exhausted;
@@ -147,36 +149,30 @@ buffer_segment_new(sen_inv *inv, uint16_t *segno)
 }
 
 void
-sen_inv_seg_expire(sen_inv *inv)
+sen_inv_seg_expire(sen_inv *inv, int32_t threshold)
 {
-  uint32_t expire_threshold = inv->header->initial_n_segments * 2;
-  return; // nop
+  uint16_t seg;
+  uint32_t th, nmaps;
   if (inv->v08p) { sen_inv_seg_expire08(inv); return; }
-  if (inv->seg->nmaps > expire_threshold) {
-    uint16_t seg;
-    for (seg = inv->header->bmax; seg; seg--) {
-      uint16_t pseg = inv->header->binfo[seg];
-      if (pseg != SEG_NOT_ASSIGNED) {
-        sen_io_mapinfo *info = &inv->seg->maps[pseg];
-        uint32_t *pnref = &inv->seg->nrefs[pseg];
-        if (info->map && !*pnref) {
-          sen_io_seg_expire(inv->seg, pseg);
-          if (inv->seg->nmaps <= expire_threshold) { return; }
-        }
-      }
-    }
-    for (seg = inv->header->amax; seg; seg--) {
-      uint16_t pseg = inv->header->ainfo[seg];
-      if (pseg != SEG_NOT_ASSIGNED) {
-        sen_io_mapinfo *info = &inv->seg->maps[pseg];
-        uint32_t *pnref = &inv->seg->nrefs[pseg];
-        if (info->map && !*pnref) {
-          sen_io_seg_expire(inv->seg, pseg);
-          if (inv->seg->nmaps <= expire_threshold) { return; }
-        }
-      }
+  th = (threshold < 0) ? (inv->header->initial_n_segments * 2) : (uint32_t) threshold;
+  if ((nmaps = inv->seg->nmaps) <= th) { return; }
+  for (seg = inv->header->bmax; seg && (inv->seg->nmaps > th); seg--) {
+    uint16_t pseg = inv->header->binfo[seg - 1];
+    if (pseg != SEG_NOT_ASSIGNED) {
+      sen_io_mapinfo *info = &inv->seg->maps[pseg];
+      uint32_t *pnref = &inv->seg->nrefs[pseg];
+      if (info->map && !*pnref) { sen_io_seg_expire(inv->seg, pseg, 0); }
     }
   }
+  for (seg = inv->header->amax; seg && (inv->seg->nmaps > th); seg--) {
+    uint16_t pseg = inv->header->ainfo[seg - 1];
+    if (pseg != SEG_NOT_ASSIGNED) {
+      sen_io_mapinfo *info = &inv->seg->maps[pseg];
+      uint32_t *pnref = &inv->seg->nrefs[pseg];
+      if (info->map && !*pnref) { sen_io_seg_expire(inv->seg, pseg, 0); }
+    }
+  }
+  SEN_LOG(sen_log_notice, "expired(%d) (%u -> %u)", threshold, nmaps, inv->seg->nmaps);
 }
 
 /* chunk */
@@ -305,7 +301,6 @@ buffer_new(sen_inv *inv, int size, uint32_t *pos,
   if (size + sizeof(buffer_header) + sizeof(buffer_term) > SEN_INV_SEGMENT_SIZE) {
     return SEG_NOT_ASSIGNED;
   }
-  // load_all_segments(inv); todo: ainfo and binfo should be inside the header
   for (seg = seg0; seg < segmax; seg += nseg0) {
     if (inv->header->binfo[seg] == SEG_NOT_ASSIGNED) { break; }
     if ((pseg = buffer_open_if_capable(inv, seg, size, &b)) != SEG_NOT_ASSIGNED) {
@@ -417,11 +412,11 @@ check_jump(buffer *b, buffer_rec *r, int j)
   if (r2->step == i) {
     SEN_LOG(sen_log_emerg, "cycle! %d(%d:%d)<->%d(%d:%d)", i, id.rid, id.sid, j, id2.rid, id2.sid);
     buffer_term_dump(b, tmp_bt);
-    return sen_other_error;
+    return sen_abnormal_error;
   }
   if (id2.rid < id.rid || (id2.rid == id.rid && id2.sid <= id.sid)) {
     SEN_LOG(sen_log_crit, "invalid jump! %d(%d:%d)(%d:%d)->%d(%d:%d)(%d:%d)", i, r->jump, r->step, id.rid, id.sid, j, r2->jump, r2->step, id2.rid, id2.sid);
-    return sen_other_error;
+    return sen_abnormal_error;
   }
   return sen_success;
 }
@@ -487,10 +482,10 @@ set_jump_r(buffer *b, buffer_rec *from, int to)
     if (BUFFER_REC_DELETED(r2)) { break; }
     if (j == (i = r->jump)) { break; }
     if (j == r->step) { break; }
-    if (check_jump(b, r, j)) { return sen_other_error; }
+    if (check_jump(b, r, j)) { return sen_internal_error; }
     r->jump = j;
     j = i;
-    if (!r->step) { return sen_other_error; }
+    if (!r->step) { return sen_abnormal_error; }
   }
   return sen_success;
 }
@@ -655,7 +650,7 @@ array_get(sen_inv *inv, uint32_t id)
   if ((pseg = inv->header->ainfo[seg]) == SEG_NOT_ASSIGNED) {
     if (segment_get_clear(inv, &pseg)) { return NULL; }
     inv->header->ainfo[seg] = pseg;
-    if (seg > inv->header->amax) { inv->header->amax = seg; }
+    if (seg >= inv->header->amax) { inv->header->amax = seg + 1; }
   }
   SEN_IO_SEG_REF(inv->seg, pseg, p)
   if (!p) { return NULL; }
@@ -674,21 +669,28 @@ sen_inv_updspec *
 sen_inv_updspec_open(uint32_t rid, uint32_t sid)
 {
   sen_inv_updspec *u;
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
   if (!(u = SEN_MALLOC(sizeof(sen_inv_updspec)))) { return NULL; }
   u->rid = rid;
   u->sid = sid;
   u->score = 0;
   u->tf = 0;
+  u->atf = 0;
   u->pos = NULL;
   u->tail = NULL;
   u->vnodes = NULL;
   return u;
 }
 
+#define SEN_INV_MAX_TF 0x1ffff
+
 sen_rc
 sen_inv_updspec_add(sen_inv_updspec *u, int pos, int32_t weight)
 {
   struct _sen_inv_pos *p;
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
+  u->atf++;
+  if (u->tf >= SEN_INV_MAX_TF) { return sen_success; }
   if (!(p = SEN_MALLOC(sizeof(struct _sen_inv_pos)))) {
     return sen_memory_exhausted;
   }
@@ -724,6 +726,7 @@ sen_inv_updspec_cmp(sen_inv_updspec *a, sen_inv_updspec *b)
 sen_rc
 sen_inv_updspec_close(sen_inv_updspec *u)
 {
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
   struct _sen_inv_pos *p = u->pos, *q;
   while (p) {
     q = p->next;
@@ -737,6 +740,7 @@ sen_inv_updspec_close(sen_inv_updspec *u)
 inline static uint8_t *
 encode_rec(sen_inv_updspec *u, unsigned int *size, int deletep)
 {
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
   uint8_t *br, *p;
   struct _sen_inv_pos *pp;
   uint32_t lpos, tf, score;
@@ -770,9 +774,12 @@ encode_rec(sen_inv_updspec *u, unsigned int *size, int deletep)
 inline static int
 sym_deletable(uint32_t tid, sen_set *h)
 {
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
   sen_inv_updspec **u;
   if (!h) { return 1; }
-  if (!sen_set_at(h, &tid, (void **) &u)) { return 1; }
+  if (!sen_set_at(h, &tid, (void **) &u)) {
+    return (ERRP(ctx, SEN_ERROR)) ? 0 : 1;
+  }
   if (!(*u)->tf || !(*u)->sid) { return 1; }
   return 0;
 }
@@ -802,6 +809,9 @@ inline static void
 sym_delete(sen_inv *inv, uint32_t tid, sen_set *h)
 {
   sis_deletable_arg arg = {inv, h};
+  if ((inv->lexicon->flags & SEN_INDEX_SHARED_LEXICON)) {
+    sen_sym_pocket_decr(inv->lexicon, tid);
+  }
   if (inv->lexicon->flags & SEN_SYM_WITH_SIS) {
     sen_sym_del_with_sis(inv->lexicon, tid, sis_deletable, &arg);
     /*
@@ -908,7 +918,7 @@ typedef struct {
 }
 
 inline static sen_rc
-buffer_flush(sen_inv *inv, uint32_t seg, sen_set *h)
+buffer_flush(sen_inv *inv, sen_ctx *ctx, uint32_t seg, sen_set *h)
 {
   buffer *sb, *db = NULL;
   sen_rc rc = sen_success;
@@ -945,7 +955,7 @@ buffer_flush(sen_inv *inv, uint32_t seg, sen_set *h)
   }
   //  sen_log("db=%p ds=%d sb=%p seg=%d", db, ds, sb, seg);
   if ((scn = sb->header.chunk) != CHUNK_NOT_ASSIGNED) {
-    sc = sen_io_win_map(inv->chunk, &sw, scn, 0, sb->header.chunk_size, SEN_IO_COPY);
+    sc = sen_io_win_map(inv->chunk, ctx, &sw, scn, 0, sb->header.chunk_size, SEN_IO_COPY);
     if (!sc) {
       SEN_LOG(sen_log_alert, "io_win_map(%d, %d) failed!", scn, sb->header.chunk_size);
       buffer_close(inv, pseg);
@@ -955,7 +965,7 @@ buffer_flush(sen_inv *inv, uint32_t seg, sen_set *h)
       return sen_memory_exhausted;
     }
   }
-  dc = sen_io_win_map(inv->chunk, &dw, dcn, 0, max_dest_chunk_size, SEN_IO_UPDATE);
+  dc = sen_io_win_map(inv->chunk, ctx, &dw, dcn, 0, max_dest_chunk_size, SEN_IO_UPDATE);
   if (!dc) {
     SEN_LOG(sen_log_alert, "io_win_map(%d, %d) failed!!", dcn, max_dest_chunk_size);
     buffer_close(inv, pseg);
@@ -1148,46 +1158,61 @@ buffer_flush(sen_inv *inv, uint32_t seg, sen_set *h)
         if (lid.flags & 2) { SEN_B_ENC(lid.score, tcp); }
       }
 
-      if (ndf) {
-        uint32_t o = tcp - tc;
-        SEN_B_ENC(o, dcp);
-        memcpy(dcp, tc, o);
-        dcp += o;
-        o = tpp - tp;
-        memcpy(dcp, tp, o);
-        dcp += o;
-      }
-      bt->size_in_chunk = (uint32_t)((dcp - dc) - bt->pos_in_chunk);
-      if (!ndf) {
+#define BTCLR {\
+  bt->tid = 0;\
+  bt->pos_in_chunk = 0;\
+  bt->size_in_chunk = 0;\
+  nterms_void++;\
+}
+#define BTSET {\
+  uint32_t o = tcp - tc;\
+  SEN_B_ENC(o, dcp);\
+  memcpy(dcp, tc, o);\
+  dcp += o;\
+  o = tpp - tp;\
+  memcpy(dcp, tp, o);\
+  dcp += o;\
+  bt->size_in_chunk = (uint32_t)((dcp - dc) - bt->pos_in_chunk);\
+}
+
+      {
         uint32_t *a;
-        if ((a = array_at(inv, bt->tid))) {
-          sen_sym_pocket_set(inv->lexicon, bt->tid, 0);
-          *a = 0;
-          sym_delete(inv, bt->tid, h);
-          array_unref(inv, bt->tid);
-        }
-        bt->tid = 0;
-        bt->pos_in_chunk = 0;
-        bt->size_in_chunk = 0;
-        nterms_void++;
-      } else if (ndf == 1 && lid.rid < 0x100000 && lid.sid < 0x800 && lid.tf == 1 && lid.score == 0) {
-        uint32_t pos_, o;
-        uint8_t *dcp_ = dc + bt->pos_in_chunk;
-        SEN_B_DEC(o, dcp_);
-        spp = dcp_ + o;
-        SEN_B_DEC(pos_, spp);
-        if (pos_ < 0x4000) {
-          uint32_t *a;
+        if (!ndf) {
           if ((a = array_at(inv, bt->tid))) {
-            sen_sym_pocket_set(inv->lexicon, bt->tid, pos_);
-            *a = (lid.rid << 12) + (lid.sid << 1) + 1;
+            if (!(inv->lexicon->flags & SEN_INDEX_SHARED_LEXICON)) {
+              sen_sym_pocket_set(inv->lexicon, bt->tid, 0);
+            }
+            *a = 0;
+            sym_delete(inv, bt->tid, h);
             array_unref(inv, bt->tid);
+            BTCLR;
+          } else {
+            BTSET;
           }
-          dcp = dc + bt->pos_in_chunk;
-          bt->tid = 0;
-          bt->pos_in_chunk = 0;
-          bt->size_in_chunk = 0;
-          nterms_void++;
+        } else if (ndf == 1 && lid.rid < 0x100000 && lid.sid < 0x800 && lid.tf == 1 && lid.score == 0) {
+          uint32_t pos_;
+          spp = tp;
+          SEN_B_DEC(pos_, spp);
+          if (inv->lexicon->flags & SEN_INDEX_SHARED_LEXICON) {
+            if (lid.sid == 1 && pos_ < 0x800 && (a = array_at(inv, bt->tid))) {
+              *a = (lid.rid << 12) + (pos_ << 1) + 1;
+              array_unref(inv, bt->tid);
+              BTCLR;
+            } else {
+              BTSET;
+            }
+          } else {
+            if (pos_ < 0x4000 && (a = array_at(inv, bt->tid))) {
+              sen_sym_pocket_set(inv->lexicon, bt->tid, pos_);
+              *a = (lid.rid << 12) + (lid.sid << 1) + 1;
+              array_unref(inv, bt->tid);
+              BTCLR;
+            } else {
+              BTSET;
+            }
+          }
+        } else {
+          BTSET;
         }
       }
     }
@@ -1259,7 +1284,7 @@ sen_inv_create(const char *path, sen_sym *lexicon, uint32_t initial_n_segments)
     header->binfo[i] = SEG_NOT_ASSIGNED;
   }
   header->initial_n_segments = initial_n_segments;
-  if (!(inv = SEN_MALLOC(sizeof(sen_inv)))) {
+  if (!(inv = SEN_GMALLOC(sizeof(sen_inv)))) {
     sen_io_close(seg);
     sen_io_close(chunk);
     return NULL;
@@ -1273,6 +1298,19 @@ sen_inv_create(const char *path, sen_sym *lexicon, uint32_t initial_n_segments)
   return inv;
 }
 
+sen_rc
+sen_inv_remove(const char *path)
+{
+  sen_rc rc;
+  char buffer[PATH_MAX];
+  if (!path || strlen(path) > PATH_MAX - 4) { return sen_invalid_argument; }
+  if ((rc = sen_sym_remove(path))) { goto exit; }
+  snprintf(buffer, PATH_MAX, "%s.c", path);
+  rc = sen_io_remove(buffer);
+exit :
+  return rc;
+}
+
 sen_inv *
 sen_inv_open(const char *path, sen_sym *lexicon)
 {
@@ -1300,7 +1338,7 @@ sen_inv_open(const char *path, sen_sym *lexicon)
     sen_io_close(chunk);
     return sen_inv_open08(path, lexicon);
   }
-  if (!(inv = SEN_MALLOC(sizeof(sen_inv)))) {
+  if (!(inv = SEN_GMALLOC(sizeof(sen_inv)))) {
     sen_io_close(seg);
     sen_io_close(chunk);
     return NULL;
@@ -1320,7 +1358,7 @@ sen_inv_close(sen_inv *inv)
   if (!inv) { return sen_invalid_argument; }
   if ((rc = sen_io_close(inv->seg))) { return rc; }
   if ((rc = sen_io_close(inv->chunk))) { return rc; }
-  SEN_FREE(inv);
+  SEN_GFREE(inv);
   return rc;
 }
 
@@ -1347,10 +1385,11 @@ sen_inv_info(sen_inv *inv, off_t *seg_size, off_t *chunk_size)
 sen_rc
 sen_inv_update(sen_inv *inv, uint32_t key, sen_inv_updspec *u, sen_set *h, int hint)
 {
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
   sen_rc rc = sen_success;
   buffer *b;
   uint8_t *bs;
-  uint16_t pseg;
+  uint16_t pseg = 0;
   buffer_rec *br = NULL;
   buffer_term *bt;
   uint32_t pos = 0, size, *a;
@@ -1359,6 +1398,7 @@ sen_inv_update(sen_inv *inv, uint32_t key, sen_inv_updspec *u, sen_set *h, int h
   }
   // sen_log("key=%d tf=%d pos0=%d rid=%d", key, u->tf, u->pos->pos, u->rid);
   if (!u->tf || !u->sid) { return sen_inv_delete(inv, key, u, h); }
+  if (u->sid > inv->header->smax) { inv->header->smax = u->sid; }
   if (!(a = array_get(inv, key))) { return sen_memory_exhausted; }
   if (!(bs = encode_rec(u, &size, 0))) { rc = sen_memory_exhausted; goto exit; }
   for (;;) {
@@ -1374,7 +1414,7 @@ sen_inv_update(sen_inv *inv, uint32_t key, sen_inv_updspec *u, sen_set *h, int h
           SEN_LOG(sen_log_debug, "flushing *a=%d seg=%d(%p) free=%d",
                   *a, *a >> W_OF_SEGMENT, b, b->header.buffer_free);
           buffer_close(inv, pseg);
-          if ((rc = buffer_flush(inv, pos >> W_OF_SEGMENT, h))) { goto exit; }
+          if ((rc = buffer_flush(inv, ctx, pos >> W_OF_SEGMENT, h))) { goto exit; }
           if (*a != pos) {
             SEN_LOG(sen_log_debug, "sen_inv_update: *a changed %d->%d", *a, pos);
             continue;
@@ -1403,13 +1443,23 @@ sen_inv_update(sen_inv *inv, uint32_t key, sen_inv_updspec *u, sen_set *h, int h
         sen_inv_updspec u2;
         uint32_t size2 = 0, v = *a;
         struct _sen_inv_pos pos2;
-        pos2.pos = sen_sym_pocket_get(inv->lexicon, key);
-        pos2.next = NULL;
-        u2.pos = &pos2;
-        u2.rid = BIT31_12(v);
-        u2.sid = BIT11_01(v);
-        u2.tf = 1;
-        u2.score = 0;
+        if (inv->lexicon->flags & SEN_INDEX_SHARED_LEXICON) {
+          pos2.pos = BIT11_01(v);
+          pos2.next = NULL;
+          u2.pos = &pos2;
+          u2.rid = BIT31_12(v);
+          u2.sid = 1;
+          u2.tf = 1;
+          u2.score = 0;
+        } else {
+          pos2.pos = sen_sym_pocket_get(inv->lexicon, key);
+          pos2.next = NULL;
+          u2.pos = &pos2;
+          u2.rid = BIT31_12(v);
+          u2.sid = BIT11_01(v);
+          u2.tf = 1;
+          u2.score = 0;
+        }
         if (u2.rid != u->rid || u2.sid != u->sid) {
           uint8_t *bs2 = encode_rec(&u2, &size2, 0);
           if (!bs2) {
@@ -1439,37 +1489,53 @@ sen_inv_update(sen_inv *inv, uint32_t key, sen_inv_updspec *u, sen_set *h, int h
     }
     break;
   }
+  if (!*a && (inv->lexicon->flags & SEN_INDEX_SHARED_LEXICON)) {
+    sen_sym_pocket_incr(inv->lexicon, key);
+  }
   if (!br) {
-    if (u->rid < 0x100000 && u->sid < 0x800 &&
-        u->tf == 1 && u->score == 0 && u->pos->pos < 0x4000) {
-      sen_sym_pocket_set(inv->lexicon, key, u->pos->pos);
-      *a = (u->rid << 12) + (u->sid << 1) + 1;
-      goto exit;
+    if (inv->lexicon->flags & SEN_INDEX_SHARED_LEXICON) {
+      if (u->rid < 0x100000 && u->sid == 1 &&
+          u->tf == 1 && u->score == 0 && u->pos->pos < 0x800) {
+        *a = (u->rid << 12) + (u->pos->pos << 1) + 1;
+        goto exit;
+      }
     } else {
-      pseg = buffer_new(inv, size, &pos, &bt, &br, &b, hint);
-      if (pseg == SEG_NOT_ASSIGNED) { goto exit; }
-      bt->tid = key;
-      bt->size_in_chunk = 0;
-      bt->pos_in_chunk = 0;
-      bt->size_in_buffer = 0;
-      bt->pos_in_buffer = 0;
+      if (u->rid < 0x100000 && u->sid < 0x800 &&
+          u->tf == 1 && u->score == 0 && u->pos->pos < 0x4000) {
+        sen_sym_pocket_set(inv->lexicon, key, u->pos->pos);
+        *a = (u->rid << 12) + (u->sid << 1) + 1;
+        goto exit;
+      }
     }
+    pseg = buffer_new(inv, size, &pos, &bt, &br, &b, hint);
+    if (pseg == SEG_NOT_ASSIGNED) { goto exit; }
+    bt->tid = key;
+    bt->size_in_chunk = 0;
+    bt->pos_in_chunk = 0;
+    bt->size_in_buffer = 0;
+    bt->pos_in_buffer = 0;
   }
   rc = buffer_put(b, bt, br, bs, u, size);
   buffer_close(inv, pseg);
   if (!*a || (*a & 1)) {
     *a = pos;
-    sen_sym_pocket_set(inv->lexicon, key, 0);
+    if (!(inv->lexicon->flags & SEN_INDEX_SHARED_LEXICON)) {
+      sen_sym_pocket_set(inv->lexicon, key, 0);
+    }
   }
 exit :
   array_unref(inv, key);
   if (bs) { SEN_FREE(bs); }
+  if (u->tf != u->atf) {
+    SEN_LOG(sen_log_warning, "too many postings(%d) on '%s'. discarded %d.", u->atf, _sen_sym_key(inv->lexicon, key), u->atf - u->tf);
+  }
   return rc;
 }
 
 sen_rc
 sen_inv_delete(sen_inv *inv, uint32_t key, sen_inv_updspec *u, sen_set *h)
 {
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
   sen_rc rc = sen_success;
   buffer *b;
   uint16_t pseg;
@@ -1505,7 +1571,7 @@ sen_inv_delete(sen_inv *inv, uint32_t key, sen_inv_updspec *u, sen_set *h)
       uint32_t _a = *a;
       SEN_LOG(sen_log_debug, "flushing! b=%p free=%d, seg(%d)", b, b->header.buffer_free, *a >> W_OF_SEGMENT);
       buffer_close(inv, pseg);
-      if ((rc = buffer_flush(inv, *a >> W_OF_SEGMENT, h))) { goto exit; }
+      if ((rc = buffer_flush(inv, ctx, *a >> W_OF_SEGMENT, h))) { goto exit; }
       if (*a != _a) {
         SEN_LOG(sen_log_debug, "sen_inv_delete: *a changed %d->%d)", *a, _a);
         continue;
@@ -1551,6 +1617,7 @@ sen_inv_initial_n_segments(sen_inv *inv)
 sen_inv_cursor *
 sen_inv_cursor_open(sen_inv *inv, uint32_t key, int with_pos)
 {
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
   sen_inv_cursor *c  = NULL;
   uint32_t pos, *a;
   if (inv->v08p) {
@@ -1561,14 +1628,20 @@ sen_inv_cursor_open(sen_inv *inv, uint32_t key, int with_pos)
   if (!(c = SEN_MALLOC(sizeof(sen_inv_cursor)))) { goto exit; }
   memset(c, 0, sizeof(sen_inv_cursor));
   c->inv = inv;
+  c->iw.ctx = ctx;
   c->with_pos = (uint16_t) with_pos;
   if (pos & 1) {
     c->stat = 0;
     c->pb.rid = BIT31_12(pos);
-    c->pb.sid = BIT11_01(pos);
     c->pb.tf = 1;
     c->pb.score = 0;
-    c->pb.pos = sen_sym_pocket_get(inv->lexicon, key);
+    if (inv->lexicon->flags & SEN_INDEX_SHARED_LEXICON) {
+      c->pb.sid = 1;
+      c->pb.pos = BIT11_01(pos);
+    } else {
+      c->pb.sid = BIT11_01(pos);
+      c->pb.pos = sen_sym_pocket_get(inv->lexicon, key);
+    }
   } else {
     uint32_t chunk;
     buffer_term *bt;
@@ -1579,7 +1652,7 @@ sen_inv_cursor_open(sen_inv *inv, uint32_t key, int with_pos)
       goto exit;
     }
     if (bt->size_in_chunk && (chunk = c->buf->header.chunk) != CHUNK_NOT_ASSIGNED) {
-      c->cp = sen_io_win_map(inv->chunk, &c->iw,
+      c->cp = sen_io_win_map(inv->chunk, ctx, &c->iw,
                              chunk, bt->pos_in_chunk, bt->size_in_chunk, sen_io_rdonly);
       if (!c->cp) {
         buffer_close(inv, c->buffer_pseg);
@@ -1619,10 +1692,15 @@ sen_inv_cursor_openv1(sen_inv *inv, uint32_t key)
   if (pos & 1) {
     c->stat = 0;
     c->pb.rid = BIT31_12(pos);
-    c->pb.sid = BIT11_01(pos);
     c->pb.tf = 1;
     c->pb.score = 0;
-    c->pb.pos = sen_sym_pocket_get(inv->lexicon, key);
+    if (inv->lexicon->flags & SEN_INDEX_SHARED_LEXICON) {
+      c->pb.sid = 1;
+      c->pb.pos = BIT11_01(pos);
+    } else {
+      c->pb.sid = BIT11_01(pos);
+      c->pb.pos = sen_sym_pocket_get(inv->lexicon, key);
+    }
   } else {
     buffer_term *bt;
     c->pb.rid = 0; c->pb.sid = 0;
@@ -1647,6 +1725,7 @@ exit :
 sen_rc
 sen_inv_cursor_openv2(sen_inv_cursor **cursors, int ncursors)
 {
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
   sen_rc rc = sen_success;
   int i, j = 0;
   sen_inv_cursor *c;
@@ -1658,7 +1737,7 @@ sen_inv_cursor_openv2(sen_inv_cursor **cursors, int ncursors)
       iws[j++] = &c->iw;
     }
   }
-  if (j) { rc = sen_io_win_mapv(iws, j); }
+  if (j) { rc = sen_io_win_mapv(iws, ctx, j); }
   for (i = 0; i < ncursors; i++) {
     c = cursors[i];
     if (c->iw.addr) {
@@ -1709,7 +1788,7 @@ sen_inv_cursor_next(sen_inv_cursor *c)
           SEN_B_DEC(c->pb.sid, c->bp);
           if (lrid > c->pb.rid || (lrid == c->pb.rid && lsid >= c->pb.sid)) {
             SEN_LOG(sen_log_crit, "brokend!! (%d:%d) -> (%d:%d)", lrid, lsid, c->pb.rid, c->pb.sid);
-            return sen_other_error;
+            return sen_abnormal_error;
           }
           c->nextb = br->step;
           SEN_B_DEC(c->pb.tf, c->bp);
@@ -1754,14 +1833,14 @@ sen_inv_cursor_next(sen_inv_cursor *c)
           if (c->pc.tf && c->pc.sid) { c->post = &c->pc; break; }
         } else {
           c->post = NULL;
-          return sen_other_error;
+          return sen_abnormal_error;
         }
       }
     }
   } else {
     if (c->stat & SOLE_DOC_USED) {
       c->post = NULL;
-      return sen_other_error;
+      return sen_abnormal_error;
     } else {
       c->post = &c->pb;
       c->stat |= SOLE_DOC_USED;
@@ -1786,7 +1865,7 @@ sen_inv_cursor_next_pos(sen_inv_cursor *c)
           SEN_B_DEC(gap, c->cpp);
           c->pc.pos += gap;
         } else {
-          rc = sen_other_error;
+          rc = sen_abnormal_error;
         }
       } else if (c->post == &c->pb) {
         if (c->pb.rest) {
@@ -1794,14 +1873,14 @@ sen_inv_cursor_next_pos(sen_inv_cursor *c)
           SEN_B_DEC(gap, c->bp);
           c->pb.pos += gap;
         } else {
-          rc = sen_other_error;
+          rc = sen_abnormal_error;
         }
       } else {
-        rc = sen_other_error;
+        rc = sen_abnormal_error;
       }
     } else {
       if (c->stat & SOLE_POS_USED) {
-        rc = sen_other_error;
+        rc = sen_abnormal_error;
       } else {
         c->stat |= SOLE_POS_USED;
       }
@@ -1813,6 +1892,7 @@ sen_inv_cursor_next_pos(sen_inv_cursor *c)
 sen_rc
 sen_inv_cursor_close(sen_inv_cursor *c)
 {
+  sen_ctx *ctx = c->iw.ctx;
   if (c->inv->v08p) {
     return sen_inv_cursor_close08(c);
   }
@@ -1864,6 +1944,7 @@ sen_inv_entry_info(sen_inv *inv, unsigned key, unsigned *a, unsigned *pocket,
   buffer_term *bt;
   uint32_t *ap;
   uint16_t pseg;
+  ERRCLR(NULL);
   if (inv->v08p) {
     return sen_inv_entry_info08(inv, key, a, pocket,
                                 chunk, chunk_size, buffer_free,
@@ -1891,3 +1972,18 @@ sen_inv_entry_info(sen_inv *inv, unsigned key, unsigned *a, unsigned *pocket,
   buffer_close(inv, pseg);
   return 4;
 }
+
+const char *
+sen_inv_path(sen_inv *inv)
+{
+  return sen_io_path(inv->seg);
+}
+
+uint32_t
+sen_inv_max_section(sen_inv *inv)
+{
+  if (inv->v08p) {
+    return SEN_SYM_MAX_ID;
+  }
+  return inv->header->smax;
+}
