@@ -4,34 +4,34 @@
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.
-  
+
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
-  
+
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
-#ifndef _SEN_INV_H
-#define _SEN_INV_H
+#ifndef SEN_INV_H
+#define SEN_INV_H
 
-#ifndef _SENNA_H
+#ifndef SENNA_H
 #include "senna_in.h"
-#endif /* _SENNA_H */
+#endif /* SENNA_H */
 
-#ifndef _SEN_SET_H
+#ifndef SEN_SET_H
 #include "set.h"
-#endif /* _SEN_SET_H */
+#endif /* SEN_SET_H */
 
-#ifndef _SEN_IO_H
+#ifndef SEN_IO_H
 #include "io.h"
-#endif /* _SEN_IO_H */
+#endif /* SEN_IO_H */
 
-#ifndef _SEN_STORE_H
+#ifndef SEN_STORE_H
 #include "store.h"
-#endif /* _SEN_STORE_H */
+#endif /* SEN_STORE_H */
 
 #ifdef  __cplusplus
 extern "C" {
@@ -50,7 +50,8 @@ struct _sen_inv_updspec {
   uint32_t rid;
   uint32_t sid;
   int32_t score;
-  int32_t tf;
+  int32_t tf;                 /* number of postings successfully stored to index */
+  int32_t atf;                /* actual number of postings */
   struct _sen_inv_pos *pos;
   struct _sen_inv_pos *tail;
   sen_vgram_vnode *vnodes;
@@ -62,6 +63,7 @@ sen_inv *sen_inv_create(const char *path, sen_sym *lexicon,
                          uint32_t initial_n_segments);
 sen_inv *sen_inv_open(const char *path, sen_sym *lexicon);
 sen_rc sen_inv_close(sen_inv *inv);
+sen_rc sen_inv_remove(const char *path);
 sen_rc sen_inv_info(sen_inv *inv, off_t *seg_size, off_t *chunk_size);
 sen_rc sen_inv_update(sen_inv *inv, uint32_t key, sen_inv_updspec *u, sen_set *h,
                       int hint);
@@ -75,7 +77,7 @@ int sen_inv_updspec_cmp(sen_inv_updspec *a, sen_inv_updspec *b);
 
 uint32_t sen_inv_estimate_size(sen_inv *inv, uint32_t key);
 
-void sen_inv_seg_expire(sen_inv *inv);
+void sen_inv_seg_expire(sen_inv *inv, int32_t threshold);
 
 typedef struct {
   sen_id rid;
@@ -117,11 +119,13 @@ sen_rc sen_inv_cursor_openv2(sen_inv_cursor **cursors, int ncursors);
 sen_rc sen_inv_cursor_next(sen_inv_cursor *c);
 sen_rc sen_inv_cursor_next_pos(sen_inv_cursor *c);
 sen_rc sen_inv_cursor_close(sen_inv_cursor *c);
+uint32_t sen_inv_max_section(sen_inv *inv);
 
 int sen_inv_check(sen_inv *inv);
+const char *sen_inv_path(sen_inv *inv);
 
 #ifdef __cplusplus
 }
 #endif
 
-#endif /* _SEN_INV_H */
+#endif /* SEN_INV_H */
