@@ -23,9 +23,8 @@
 #include <fcntl.h>
 #include <string.h>
 #include <sys/stat.h>
-#include <errno.h>
 
-#include "str.h"
+#include "ctx.h"
 #include "io.h"
 #include "cache.h"
 
@@ -84,8 +83,15 @@ inline static int sen_munmap(HANDLE *fmo, void *start, size_t length);
 #else /* defined(WIN32) && defined(WIN32_FMO_EACH) */
 inline static void * sen_mmap(fileinfo *fi, off_t offset, size_t length);
 inline static int sen_munmap(void *start, size_t length);
-#define SEN_MMAP(fmo,fi,offset,length) (sen_mmap((fi), (offset), (length)))
 #define SEN_MUNMAP(fmo,start,length) (sen_munmap((start), (length)))
+#ifdef USE_FAIL_MALLOC
+inline static void * sen_fail_mmap(fileinfo *fi, off_t offset, size_t length,
+                                   const char* file, int line, const char *func);
+#define SEN_MMAP(fmo,fi,offset,length) \
+  (sen_fail_mmap((fi), (offset), (length), __FILE__, __LINE__, __FUNCTION__))
+#else /* USE_FAIL_MALLOC */
+#define SEN_MMAP(fmo,fi,offset,length) (sen_mmap((fi), (offset), (length)))
+#endif /* USE_FAIL_MALLOC */
 #endif  /* defined(WIN32) && defined(WIN32_FMO_EACH) */
 inline static int sen_msync(void *start, size_t length);
 inline static sen_rc sen_pread(fileinfo *fi, void *buf, size_t count, off_t offset);
@@ -108,38 +114,39 @@ sen_io_create(const char *path, uint32_t header_size, uint32_t segment_size,
   max_nfiles = (unsigned int)(
     ((uint64_t)segment_size * (max_segment + bs) + SEN_IO_FILE_SIZE - 1)
     / SEN_IO_FILE_SIZE);
-  if (!(fis = SEN_MALLOC(sizeof(fileinfo) * max_nfiles))) { return NULL; }
-  sen_fileinfo_init(fis, max_nfiles);
-  if (!sen_open(fis, path, O_RDWR|O_CREAT|O_TRUNC, SEN_IO_FILE_SIZE)) {
-    if ((header = SEN_MMAP(&fis->fmo, fis, 0, b))) {
-      header->header_size = header_size;
-      header->segment_size = segment_size;
-      header->max_segment = max_segment;
-      memcpy(header->idstr, SEN_IO_IDSTR, 16);
-      sen_msync(header, b);
-      if ((io = SEN_MALLOC(sizeof(sen_io)))) {
-        if ((io->maps = SEN_MALLOC(sizeof(sen_io_mapinfo) * max_segment))) {
-          memset(io->maps, 0, sizeof(sen_io_mapinfo) * max_segment);
-          strncpy(io->path, path, PATH_MAX);
-          io->header = header;
-          io->nrefs = (uint32_t *)(((byte *) header) + IO_HEADER_SIZE);
-          io->user_header = ((byte *) io->nrefs) + max_segment * sizeof(uint32_t);
-          io->base = b;
-          io->base_seg = bs;
-          io->mode = mode;
-          io->cache_size = cache_size;
-          io->fis = fis;
-          io->nmaps = 0;
-          io->count = 0;
-          return io;
+  if ((fis = SEN_GMALLOCN(fileinfo, max_nfiles))) {
+    sen_fileinfo_init(fis, max_nfiles);
+    if (!sen_open(fis, path, O_RDWR|O_CREAT|O_TRUNC, SEN_IO_FILE_SIZE)) {
+      if ((header = (io_header *)SEN_MMAP(&fis->fmo, fis, 0, b))) {
+        header->header_size = header_size;
+        header->segment_size = segment_size;
+        header->max_segment = max_segment;
+        memcpy(header->idstr, SEN_IO_IDSTR, 16);
+        sen_msync(header, b);
+        if ((io = SEN_GMALLOCN(sen_io, 1))) {
+          if ((io->maps = SEN_GMALLOCN(sen_io_mapinfo, max_segment))) {
+            memset(io->maps, 0, sizeof(sen_io_mapinfo) * max_segment);
+            strncpy(io->path, path, PATH_MAX);
+            io->header = header;
+            io->nrefs = (uint32_t *)(((byte *) header) + IO_HEADER_SIZE);
+            io->user_header = ((byte *) io->nrefs) + max_segment * sizeof(uint32_t);
+            io->base = b;
+            io->base_seg = bs;
+            io->mode = mode;
+            io->cache_size = cache_size;
+            io->fis = fis;
+            io->nmaps = 0;
+            io->count = 0;
+            return io;
+          }
+          SEN_GFREE(io);
         }
-        SEN_FREE(io);
+        SEN_MUNMAP(&fis->fmo, header, b);
       }
-      SEN_MUNMAP(&fis->fmo, header, b);
+      sen_close(fis);
     }
-    sen_close(fis);
+    SEN_GFREE(fis);
   }
-  SEN_FREE(fis);
   return NULL;
 }
 
@@ -157,7 +164,7 @@ sen_io_open(const char *path, sen_io_mode mode, uint32_t cache_size)
   {
     io_header h;
     int fd = open(path, O_RDWR);
-    if (fd == -1) { return NULL; }
+    if (fd == -1) { GSERR(path); return NULL; }
     if (fstat(fd, &s) != -1 && s.st_size >= sizeof(io_header)) {
       if (read(fd, &h, sizeof(io_header)) == sizeof(io_header)) {
         if (memcmp(h.idstr, SEN_IO_IDSTR, 16)) {
@@ -182,17 +189,17 @@ sen_io_open(const char *path, sen_io_mode mode, uint32_t cache_size)
   max_nfiles = (unsigned int)(
     ((uint64_t)segment_size * (max_segment + bs) + SEN_IO_FILE_SIZE - 1)
     / SEN_IO_FILE_SIZE);
-  if (!(fis = SEN_MALLOC(sizeof(fileinfo) * max_nfiles))) { return NULL; }
+  if (!(fis = SEN_GMALLOCN(fileinfo, max_nfiles))) { return NULL; }
   sen_fileinfo_init(fis, max_nfiles);
   if (!sen_open(fis, path, O_RDWR, SEN_IO_FILE_SIZE)) {
     io_header *header;
     if ((header = SEN_MMAP(&fis->fmo, fis, 0, b))) {
-      if ((io = SEN_MALLOC(sizeof(sen_io)))) {
-        if ((io->maps = SEN_CALLOC(sizeof(sen_io_mapinfo) * max_segment))) {
+      if ((io = SEN_GMALLOC(sizeof(sen_io)))) {
+        if ((io->maps = SEN_GCALLOC(sizeof(sen_io_mapinfo) * max_segment))) {
           strncpy(io->path, path, PATH_MAX);
           io->header = header;
           if (v08p) {
-            io->nrefs = SEN_CALLOC(sizeof(uint32_t) * max_segment);
+            io->nrefs = SEN_GCALLOC(sizeof(uint32_t) * max_segment);
             io->user_header = ((byte *) header) + sizeof(io_header08);
           } else {
             io->nrefs = (uint32_t *)(((byte *) header) + IO_HEADER_SIZE);
@@ -208,15 +215,15 @@ sen_io_open(const char *path, sen_io_mode mode, uint32_t cache_size)
             io->count = 0;
             return io;
           }
-          SEN_FREE(io->maps);
+          SEN_GFREE(io->maps);
         }
-        SEN_FREE(io);
+        SEN_GFREE(io);
       }
       SEN_MUNMAP(&fis->fmo, header, b);
     }
     sen_close(fis);
   }
-  SEN_FREE(fis);
+  SEN_GFREE(fis);
   return NULL;
 }
 
@@ -240,9 +247,9 @@ sen_io_close(sen_io *io)
   }
   SEN_MUNMAP(&io->fis->fmo, io->header, io->base);
   for (fi = io->fis, i = max_nfiles; i; fi++, i--) { sen_close(fi); }
-  SEN_FREE(io->fis);
-  SEN_FREE(io->maps);
-  SEN_FREE(io);
+  SEN_GFREE(io->fis);
+  SEN_GFREE(io->maps);
+  SEN_GFREE(io);
   return sen_success;
 }
 
@@ -292,7 +299,11 @@ sen_io_size(sen_io *io, off_t *size)
     / SEN_IO_FILE_SIZE);
   for (fno = 0; fno < max_nfiles; fno++) {
     gen_pathname(io->path, buffer, fno);
-    if (!stat(buffer, &s)) { tsize += s.st_size; }
+    if (stat(buffer, &s)) {
+      GSERR(buffer);
+    } else {
+      tsize += s.st_size;
+    }
   }
   *size = tsize;
   return sen_success;
@@ -302,23 +313,38 @@ sen_rc
 sen_io_remove(const char *path)
 {
   struct stat s;
-  if (!stat(path, &s) && !unlink(path)) {
+  if (stat(path, &s)) {
+    SEN_LOG(sen_log_info, "stat failed '%s' (%s)", path, strerror(errno));
+    return sen_file_operation_error;
+  } else if (unlink(path)) {
+    GSERR(path);
+    return sen_file_operation_error;
+  } else {
     int fno;
     char buffer[PATH_MAX];
     for (fno = 1; ; fno++) {
       gen_pathname(path, buffer, fno);
-      if (stat(buffer, &s) || unlink(buffer)) { break; }
+      if (!stat(buffer, &s)) {
+        if (unlink(buffer)) { GSERR(buffer); }
+      } else {
+        break;
+      }
     }
     return sen_success;
   }
-  return sen_file_operation_error;
 }
 
 sen_rc
 sen_io_rename(const char *old_name, const char *new_name)
 {
   struct stat s;
-  if (!stat(old_name, &s) && !rename(old_name, new_name)) {
+  if (stat(old_name, &s)) {
+    SEN_LOG(sen_log_info, "stat failed '%s' (%s)", old_name, strerror(errno));
+    return sen_file_operation_error;
+  } else if (rename(old_name, new_name)) {
+    GSERR(old_name);
+    return sen_file_operation_error;
+  } else {
     int fno;
     char old_buffer[PATH_MAX];
     char new_buffer[PATH_MAX];
@@ -326,7 +352,7 @@ sen_io_rename(const char *old_name, const char *new_name)
       gen_pathname(old_name, old_buffer, fno);
       if (!stat(old_buffer, &s)) {
         gen_pathname(new_name, new_buffer, fno);
-        rename(old_buffer, new_buffer);
+        if (rename(old_buffer, new_buffer)) { GSERR(old_buffer); }
       } else {
         break;
       }
@@ -337,7 +363,7 @@ sen_io_rename(const char *old_name, const char *new_name)
 }
 
 void *
-sen_io_win_map(sen_io *io, sen_io_win *iw, uint32_t segment,
+sen_io_win_map(sen_io *io, sen_ctx *ctx, sen_io_win *iw, uint32_t segment,
                uint32_t offset, uint32_t size, sen_io_rw_mode mode)
 {
   byte *p;
@@ -346,6 +372,7 @@ sen_io_win_map(sen_io *io, sen_io_win *iw, uint32_t segment,
   uint32_t nseg, bseg;
   uint32_t segment_size = io->header->segment_size;
   uint32_t segments_per_file = SEN_IO_FILE_SIZE / segment_size;
+  iw->ctx = ctx;
   iw->diff = 0;
   if (offset >= segment_size) {
     segment += offset / segment_size;
@@ -395,6 +422,7 @@ sen_io_win_map(sen_io *io, sen_io_win *iw, uint32_t segment,
       }
     } else {
       SEN_LOG(sen_log_alert, "nseg == 0! in sen_io_win_map(%p, %u, %u, %u)", io, segment, offset, size);
+      return NULL;
       // if (!(p = sen_io_seg_ref(io, segment))) { return NULL; }
     }
     iw->addr = p + offset;
@@ -419,7 +447,7 @@ sen_io_win_map(sen_io *io, sen_io_win *iw, uint32_t segment,
 
 #ifdef USE_AIO
 sen_rc
-sen_io_win_mapv(sen_io_win **list, int nent)
+sen_io_win_mapv(sen_io_win **list, sen_ctx *ctx, int nent)
 {
   int i;
   sen_io_win *iw;
@@ -457,14 +485,14 @@ retry:
         pos = (bseg % segments_per_file) * segment_size + offset + base;
         if (!size || !io || segment + nseg > io->header->max_segment ||
             fno != (bseg + nseg - 1) / segments_per_file) {
-          return sen_other_error;
+          return sen_abnormal_error;
         }
         fi = &io->fis[fno];
         if (!sen_opened(fi)) {
             char path[PATH_MAX];
             gen_pathname(io->path, path, fno);
             if (sen_open(fi, path, O_RDWR|O_CREAT|O_DIRECT, SEN_IO_FILE_SIZE)) {
-                return sen_other_error;
+                return sen_internal_error;
             }
         }
         {
@@ -507,7 +535,7 @@ retry:
 
                 /* allocate aligned memory */
                 if (posix_memalign(&p, MEM_ALIGN, vsize) != 0) {
-                    return sen_other_error;
+                    return sen_external_error;
                 }
                 iocb[count].aio_buf = p;
                 iocb[count].aio_nbytes = vsize;
@@ -529,8 +557,8 @@ retry:
             iw->pos = pos;
         } /* End  AIO + DIO + cache hack */
     } else {
-        if (!sen_io_win_map(iw->io, iw, iw->segment, iw->offset, iw->size, iw->mode)) {
-            return sen_other_error;
+        if (!sen_io_win_map(iw->io, ctx, iw, iw->segment, iw->offset, iw->size, iw->mode)) {
+            return sen_internal_error;
         }
     }
   }
@@ -542,7 +570,7 @@ retry:
           /* aio_read () */
           if (lio_listio (LIO_WAIT, iocbs, count, NULL) < 0) {
               perror ("lio_listio");
-              return sen_other_error;
+              return sen_external_error;
           }
           for (c=0;c<count;c++) {
               /* cache data is now VALID */
@@ -574,6 +602,7 @@ sen_io_win_unmap(sen_io_win *iw)
 {
   sen_rc rc = sen_success;
   sen_io *io = iw->io;
+  sen_ctx *ctx = iw->ctx;
   uint32_t segment_size = io->header->segment_size;
   uint32_t segments_per_file = SEN_IO_FILE_SIZE / segment_size;
   int nseg = iw->nseg;
@@ -687,7 +716,7 @@ sen_io_seg_ref(sen_io *io, uint32_t segno)
         SEN_LOG(sen_log_crit, "deadlock detected! in sen_io_seg_ref(%p, %u)", io, segno);
         return NULL;
       }
-      usleep(1);
+      usleep(1000);
       continue;
     }
     if (!info->map) {
@@ -697,7 +726,7 @@ sen_io_seg_ref(sen_io *io, uint32_t segno)
           SEN_LOG(sen_log_crit, "deadlock detected!! in sen_io_seg_ref(%p, %u)", io, segno);
           return NULL;
         }
-        usleep(1);
+        usleep(1000);
         continue;
       } else {
         SEG_MAP(io, segno, info);
@@ -722,7 +751,7 @@ sen_io_seg_unref(sen_io *io, uint32_t segno)
 }
 
 sen_rc
-sen_io_seg_expire(sen_io *io, uint32_t segno)
+sen_io_seg_expire(sen_io *io, uint32_t segno, uint32_t nretry)
 {
   uint32_t retry, *pnref;
   sen_io_mapinfo *info;
@@ -737,31 +766,28 @@ sen_io_seg_expire(sen_io *io, uint32_t segno)
       SEN_ATOMIC_ADD_EX(pnref, -1, nref);
       if (retry >= SEN_IO_MAX_RETRY) {
         SEN_LOG(sen_log_crit, "deadlock detected! in sen_io_seg_expire(%p, %u, %u)", io, segno, nref);
-        return sen_other_error;
+        return sen_abnormal_error;
       }
-      usleep(1);
-      continue;
     } else {
       SEN_ATOMIC_ADD_EX(pnref, SEN_IO_MAX_REF, nref);
       if (nref > 1) {
         SEN_ATOMIC_ADD_EX(pnref, -(SEN_IO_MAX_REF + 1), nref);
         if (retry >= SEN_IO_MAX_RETRY) {
           SEN_LOG(sen_log_crit, "deadlock detected!! in sen_io_seg_expire(%p, %u, %u)", io, segno, nref);
-          return sen_other_error;
+          return sen_abnormal_error;
         }
-        usleep(1);
-        continue;
       } else {
         uint32_t nmaps;
         SEN_MUNMAP(&info->fmo, info->map, io->header->segment_size);
         info->map = NULL;
         SEN_ATOMIC_ADD_EX(pnref, -(SEN_IO_MAX_REF + 1), nref);
         SEN_ATOMIC_ADD_EX(&io->nmaps, -1, nmaps);
-        break;
+        return sen_success;
       }
     }
+    if (retry >= nretry) { return sen_abnormal_error; }
+    usleep(1000);
   }
-  return sen_success;
 }
 
 /** mmap abstraction **/
@@ -786,13 +812,19 @@ sen_open(fileinfo *fi, const char *path, int flags, size_t maxsize)
   fi->fh = CreateFile(path, GENERIC_READ | GENERIC_WRITE,
                       FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
                       OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0);
-  if (fi->fh == INVALID_HANDLE_VALUE) { return sen_file_operation_error; }
+  if (fi->fh == INVALID_HANDLE_VALUE) {
+    GERR(sen_file_operation_error, "CreateFile failed");
+    return sen_file_operation_error;
+  }
   if ((flags & O_TRUNC)) {
     CloseHandle(fi->fh);
     fi->fh = CreateFile(path, GENERIC_READ | GENERIC_WRITE,
                         FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
                         TRUNCATE_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
-    if (fi->fh == INVALID_HANDLE_VALUE) { return sen_file_operation_error; }
+    if (fi->fh == INVALID_HANDLE_VALUE) {
+      GERR(sen_file_operation_error, "CreateFile failed");
+      return sen_file_operation_error;
+    }
   }
   MUTEX_INIT(fi->mutex);
   return sen_success;
@@ -808,12 +840,16 @@ sen_mmap(HANDLE *fmo, fileinfo *fi, off_t offset, size_t length)
   *fmo = CreateFileMapping(fi->fh, NULL, PAGE_READWRITE, 0, offset + length, NULL);
   if (!*fmo) { return NULL; }
   res = MapViewOfFile(*fmo, FILE_MAP_WRITE, 0, (DWORD)offset, (SIZE_T)length);
-  if (res) {
-    mmap_size += length;
-  } else {
-    SEN_LOG(sen_log_alert, "MapViewOfFile failed #%d <%d>", GetLastError(), mmap_size);
+  if (!res) {
+    sen_index_expire();
+    res = MapViewOfFile(*fmo, FILE_MAP_WRITE, 0, (DWORD)offset, (SIZE_T)length);
+    if (!res) {
+      GMERR("MapViewOfFile failed #%d <%d>", GetLastError(), mmap_size);
+      return NULL;
+    }
   }
   /* MUTEX_UNLOCK(fi->mutex); */
+  mmap_size += length;
   return res;
 }
 
@@ -860,14 +896,20 @@ sen_open(fileinfo *fi, const char *path, int flags, size_t maxsize)
   fi->fh = CreateFile(path, GENERIC_READ | GENERIC_WRITE,
                       FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
                       OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0);
-  if (fi->fh == INVALID_HANDLE_VALUE) { return sen_file_operation_error; }
+  if (fi->fh == INVALID_HANDLE_VALUE) {
+    GERR(sen_file_operation_error, "CreateFile failed");
+    return sen_file_operation_error;
+  }
   if ((flags & O_TRUNC)) {
     CloseHandle(fi->fh);
     /* unable to assign OPEN_ALWAYS and TRUNCATE_EXISTING at once */
     fi->fh = CreateFile(path, GENERIC_READ | GENERIC_WRITE,
                         FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
                         TRUNCATE_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
-    if (fi->fh == INVALID_HANDLE_VALUE) { return sen_file_operation_error; }
+    if (fi->fh == INVALID_HANDLE_VALUE) {
+      GERR(sen_file_operation_error, "CreateFile failed");
+      return sen_file_operation_error;
+    }
   }
   /* signature may be wrong.. */
   fi->fmo = OpenFileMapping(FILE_MAP_ALL_ACCESS, FALSE, NULL);
@@ -895,6 +937,7 @@ sen_open(fileinfo *fi, const char *path, int flags, size_t maxsize)
     SEN_LOG(sen_log_alert, "failed to get FileMappingObject #%d", GetLastError());
   }
   CloseHandle(fi->fh);
+  GERR(sen_file_operation_error, "OpenFileMapping failed");
   return sen_file_operation_error;
 }
 
@@ -919,11 +962,15 @@ sen_mmap(fileinfo *fi, off_t offset, size_t length)
   }
   */
   res = MapViewOfFile(fi->fmo, FILE_MAP_WRITE, 0, (DWORD)offset, (SIZE_T)length);
-  if (res) {
-    mmap_size += length;
-  } else {
-    SEN_LOG(sen_log_alert, "MapViewOfFile failed #%d  <%d>", GetLastError(), mmap_size);
+  if (!res) {
+    sen_index_expire();
+    res = MapViewOfFile(fi->fmo, FILE_MAP_WRITE, 0, (DWORD)offset, (SIZE_T)length);
+    if (!res) {
+      GMERR("MapViewOfFile failed #%d  <%d>", GetLastError(), mmap_size);
+      return NULL;
+    }
   }
+  mmap_size += length;
   return res;
 }
 
@@ -1028,7 +1075,12 @@ sen_pwrite(fileinfo *fi, void *buf, size_t count, off_t offset)
 inline static unsigned int
 sen_get_pagesize(void)
 {
-  return (unsigned int)sysconf(_SC_PAGESIZE);
+  long v = sysconf(_SC_PAGESIZE);
+  if (v == -1) {
+    GSERR("_SC_PAGESIZE");
+    return 0;
+  }
+  return (unsigned int) v;
 }
 
 inline static sen_rc
@@ -1036,9 +1088,13 @@ sen_open(fileinfo *fi, const char *path, int flags, size_t maxsize)
 {
   struct stat st;
   if ((fi->fd = open(path, flags, 0666)) == -1) {
+    GSERR(path);
+    return sen_file_operation_error;
+  }
+  if (fstat(fi->fd, &st) == -1) {
+    GSERR(path);
     return sen_file_operation_error;
   }
-  fstat(fi->fd, &st);
   fi->dev = st.st_dev;
   fi->inode = st.st_ino;
   return sen_success;
@@ -1060,7 +1116,10 @@ inline static sen_rc
 sen_close(fileinfo *fi)
 {
   if (fi->fd != -1) {
-    close(fi->fd);
+    if (close(fi->fd) == -1) {
+      GSERR("close");
+      return sen_file_operation_error;
+    }
     fi->fd = -1;
   }
   return sen_success;
@@ -1079,17 +1138,39 @@ sen_mmap(fileinfo *fi, off_t offset, size_t length)
   }
   res = mmap(NULL, length, PROT_READ|PROT_WRITE, MAP_SHARED, fi->fd, offset);
   if (MAP_FAILED == res) {
-    SEN_LOG(sen_log_alert, "mmap(%zu,%d,%d)=%s <%zu>", length, fi->fd, offset, strerror(errno), mmap_size);
+    sen_index_expire();
+    res = mmap(NULL, length, PROT_READ|PROT_WRITE, MAP_SHARED, fi->fd, offset);
+    if (MAP_FAILED == res) {
+      GMERR("mmap(%zu,%d,%d)=%s <%zu>", length, fi->fd, offset, strerror(errno), mmap_size);
+    }
     return NULL;
   }
   mmap_size += length;
   return res;
 }
 
+#ifdef USE_FAIL_MALLOC
+inline static void *
+sen_fail_mmap(fileinfo *fi, off_t offset, size_t length,
+              const char* file, int line, const char *func)
+{
+  if (fail_malloc_check(length, file, line, func)) {
+    return sen_mmap(fi, offset, length);
+  } else {
+    sen_index_expire();
+    GMERR("fail_mmap(%zu,%d,%d) (%s:%d@%s) <%zu>",
+          length, fi->fd, offset, file, line, func, mmap_size);
+    return NULL;
+  }
+}
+#endif /* USE_FAIL_MALLOC */
+
 inline static int
 sen_msync(void *start, size_t length)
 {
-  return msync(start, length, MS_SYNC);
+  int r = msync(start, length, MS_SYNC);
+  if (r == -1) { GSERR("msync"); }
+  return r;
 }
 
 inline static int
@@ -1108,7 +1189,13 @@ sen_munmap(void *start, size_t length)
 inline static sen_rc
 sen_pread(fileinfo *fi, void *buf, size_t count, off_t offset)
 {
-  if (pread(fi->fd, buf, count, offset) != count) {
+  ssize_t r = pread(fi->fd, buf, count, offset);
+  if (r != count) {
+    if (r == -1) {
+      GSERR("pread");
+    } else {
+      GERR(sen_file_operation_error, "pread returned %d != %d", r, count);
+    }
     return sen_file_operation_error;
   }
   return sen_success;
@@ -1117,7 +1204,13 @@ sen_pread(fileinfo *fi, void *buf, size_t count, off_t offset)
 inline static sen_rc
 sen_pwrite(fileinfo *fi, void *buf, size_t count, off_t offset)
 {
-  if (pwrite(fi->fd, buf, count, offset) != count) {
+  ssize_t r = pwrite(fi->fd, buf, count, offset);
+  if (r != count) {
+    if (r == -1) {
+      GSERR("pwrite");
+    } else {
+      GERR(sen_file_operation_error, "pwrite returned %d != %d", r, count);
+    }
     return sen_file_operation_error;
   }
   return sen_success;
