@@ -14,12 +14,12 @@
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
-#ifndef _SEN_IO_H
-#define _SEN_IO_H
+#ifndef SEN_IO_H
+#define SEN_IO_H
 
-#ifndef _SENNA_H
+#ifndef SENNA_H
 #include "senna_in.h"
-#endif /* _SENNA_H */
+#endif /* SENNA_H */
 
 #ifdef	__cplusplus
 extern "C" {
@@ -54,6 +54,7 @@ typedef struct _sen_io sen_io;
 
 typedef struct {
   sen_io *io;
+  sen_ctx *ctx;
   sen_io_rw_mode mode;
   uint32_t segment;
   uint32_t offset;
@@ -101,14 +102,14 @@ sen_rc sen_io_size(sen_io *io, off_t *size);
 sen_rc sen_io_rename(const char *old_name, const char *new_name);
 void *sen_io_header(sen_io *io);
 
-void *sen_io_win_map(sen_io *io, sen_io_win *iw, uint32_t segment,
+void *sen_io_win_map(sen_io *io, sen_ctx *ctx, sen_io_win *iw, uint32_t segment,
 		     uint32_t offset, uint32_t size, sen_io_rw_mode mode);
-sen_rc sen_io_win_mapv(sen_io_win **list, int nent);
+sen_rc sen_io_win_mapv(sen_io_win **list, sen_ctx *ctx, int nent);
 sen_rc sen_io_win_unmap(sen_io_win *iw);
 
 void * sen_io_seg_ref(sen_io *io, uint32_t segno);
 sen_rc sen_io_seg_unref(sen_io *io, uint32_t segno);
-sen_rc sen_io_seg_expire(sen_io *io, uint32_t segno);
+sen_rc sen_io_seg_expire(sen_io *io, uint32_t segno, uint32_t nretry);
 
 #define SEN_IO_MAX_REF 0x80000000
 #define SEN_IO_MAX_RETRY 0x10000
@@ -138,7 +139,7 @@ void sen_io_seg_map_(sen_io *io, uint32_t segno, sen_io_mapinfo *info);
         *pnref = 0; /* force reset */ \
 	break;\
       }\
-      usleep(1);\
+      usleep(1000);\
       continue;\
     }\
     if (!info->map) {\
@@ -148,7 +149,7 @@ void sen_io_seg_map_(sen_io *io, uint32_t segno, sen_io_mapinfo *info);
 	  SEN_LOG(sen_log_crit, "deadlock detected!! in SEN_IO_SEG_REF(%p, %u, %u)", io, segno, nref);\
 	  break;\
 	}\
-	usleep(1);\
+	usleep(1000);\
 	continue;\
       } else {\
 	sen_io_seg_map_(io, segno, info);\
@@ -188,13 +189,17 @@ void sen_io_seg_map_(sen_io *io, uint32_t segno, sen_io_mapinfo *info);
       if (nref) {\
 	SEN_ATOMIC_ADD_EX(pnref, -1, nref);\
 	if (retry >= SEN_IO_MAX_RETRY) {\
-	  SEN_LOG(sen_log_crit, "deadlock detected!! in sen_io_seg_ref(%p, %u)", io, segno);\
+	  SEN_LOG(sen_log_crit, "deadlock detected!! in SEN_IO_SEG_MAP(%p, %u)", io, segno);\
 	  break;\
 	}\
-	usleep(1);\
+	usleep(1000);\
 	continue;\
       } else {\
 	sen_io_seg_map_(io, segno, info);\
+        if (!info->map) {\
+          SEN_ATOMIC_ADD_EX(pnref, -1, nref);\
+          SEN_LOG(sen_log_crit, "mmap failed!! in SEN_IO_SEG_MAP(%p, %u, %u)", io, segno, nref);\
+        }\
         break;\
       }\
     }\
@@ -324,4 +329,4 @@ const char *sen_io_path(sen_io *io);
   while (size--) { *p2++ = *p1++; } \
 }
 
-#endif /* _SEN_IO_H */
+#endif /* SEN_IO_H */
