@@ -17,7 +17,6 @@
 #include "senna_in.h"
 #include <string.h>
 #include <ctype.h>
-#include "str.h"
 #include "lex.h"
 
 /* ngram */
@@ -26,12 +25,15 @@ inline static sen_lex *
 sen_ngram_open(sen_sym *sym, sen_nstr *nstr, uint8_t flags)
 {
   sen_lex *lex;
+  sen_ctx *ctx = nstr->ctx;
   if (!(lex = SEN_MALLOC(sizeof(sen_lex)))) { return NULL; }
   lex->sym = sym;
 #ifndef NO_MECAB
   lex->mecab = NULL;
 #endif /* NO_MECAB */
   lex->buf = NULL;
+  lex->token = NULL;
+  lex->tlen = 0;
   lex->pos = -1;
   lex->skip = 1;
   lex->tail = 0;
@@ -48,16 +50,27 @@ sen_ngram_open(sen_sym *sym, sen_nstr *nstr, uint8_t flags)
   return lex;
 }
 
+#define LEX_TOKEN(lex,str,len) do {\
+  if ((lex)->tlen < len) {\
+    char *buf = SEN_REALLOC((lex)->token, (len) + 1);\
+    if (!(buf)) { (lex)->status = sen_lex_done; return SEN_SYM_NIL; }\
+    (lex)->token = buf;\
+    (lex)->tlen = len;\
+  }\
+  memcpy((lex)->token, str, len);\
+  (lex)->token[len] = '\0';\
+} while (0)
+
 inline static sen_id
 sen_ngram_next(sen_lex *lex)
 {
-  char *buf;
   sen_id tid;
   sen_sym *sym = lex->sym;
+  sen_ctx *ctx = lex->nstr->ctx;
   uint_least8_t *cp = NULL;
   int32_t len = 0, pos;
   const unsigned char *p, *q, *r;
-  if (lex->status) { return SEN_SYM_NIL; }
+  if (lex->status == sen_lex_done) { return SEN_SYM_NIL; }
   lex->force_prefix = 0;
   for (p = lex->next, pos = lex->pos + lex->skip; *p; p = r, pos++) {
     if (lex->nstr->ctypes) { cp = lex->nstr->ctypes + pos; }
@@ -72,19 +85,11 @@ sen_ngram_next(sen_lex *lex)
       {
         size_t blen = r - p;
         if (!blen) {
-          lex->status = sen_lex_not_found;
-          return SEN_SYM_NIL;
-        }
-        if (!(buf = SEN_MALLOC(blen + 1))) { return SEN_SYM_NIL; }
-        memcpy(buf, p, blen);
-        buf[blen] = '\0';
-        // sen_log("alpha(%s)", buf);
-        tid = (lex->flags & SEN_LEX_ADD) ? sen_sym_get(sym, buf) : sen_sym_at(sym, buf);
-        SEN_FREE(buf);
-        if (tid == SEN_SYM_NIL) {
-          lex->status = sen_lex_not_found;
+          lex->status = sen_lex_done;
           return SEN_SYM_NIL;
         }
+        LEX_TOKEN(lex, p, blen);
+        tid = (lex->flags & SEN_LEX_ADD) ? sen_sym_get(sym, lex->token) : sen_sym_at(sym, lex->token);
         lex->skip = len;
       }
     } else if (lex->uni_digit && SEN_NSTR_CTYPE(*cp) == sen_str_digit) {
@@ -98,19 +103,11 @@ sen_ngram_next(sen_lex *lex)
       {
         size_t blen = r - p;
         if (!blen) {
-          lex->status = sen_lex_not_found;
-          return SEN_SYM_NIL;
-        }
-        if (!(buf = SEN_MALLOC(blen + 1))) { return SEN_SYM_NIL; }
-        memcpy(buf, p, blen);
-        buf[blen] = '\0';
-        // sen_log("digit(%s) len=%d", buf, len);
-        tid = (lex->flags & SEN_LEX_ADD) ? sen_sym_get(sym, buf) : sen_sym_at(sym, buf);
-        SEN_FREE(buf);
-        if (tid == SEN_SYM_NIL) {
-          lex->status = sen_lex_not_found;
+          lex->status = sen_lex_done;
           return SEN_SYM_NIL;
         }
+        LEX_TOKEN(lex, p, blen);
+        tid = (lex->flags & SEN_LEX_ADD) ? sen_sym_get(sym, lex->token) : sen_sym_at(sym, lex->token);
         lex->skip = len;
       }
     } else if (lex->uni_symbol && SEN_NSTR_CTYPE(*cp) == sen_str_symbol) {
@@ -124,19 +121,11 @@ sen_ngram_next(sen_lex *lex)
       {
         size_t blen = r - p;
         if (!blen) {
-          lex->status = sen_lex_not_found;
-          return SEN_SYM_NIL;
-        }
-        if (!(buf = SEN_MALLOC(blen + 1))) { return SEN_SYM_NIL; }
-        memcpy(buf, p, blen);
-        buf[blen] = '\0';
-        // sen_log("symbol(%s)", buf);
-        tid = (lex->flags & SEN_LEX_ADD) ? sen_sym_get(sym, buf) : sen_sym_at(sym, buf);
-        SEN_FREE(buf);
-        if (tid == SEN_SYM_NIL) {
-          lex->status = sen_lex_not_found;
+          lex->status = sen_lex_done;
           return SEN_SYM_NIL;
         }
+        LEX_TOKEN(lex, p, blen);
+        tid = (lex->flags & SEN_LEX_ADD) ? sen_sym_get(sym, lex->token) : sen_sym_at(sym, lex->token);
         lex->skip = len;
       }
     } else {
@@ -145,7 +134,10 @@ sen_ngram_next(sen_lex *lex)
       {
         const unsigned char *key = NULL;
         if ((tid = sen_sym_common_prefix_search(sym, p))) {
-          key = _sen_sym_key(sym, tid);
+          if (!(key = _sen_sym_key(sym, tid))) {
+            lex->status = sen_lex_not_found;
+            return SEN_SYM_NIL;
+          }
           len = sen_str_len(key, lex->encoding, NULL);
         }
         r = p + sen_str_charlen(p, lex->encoding);
@@ -157,7 +149,7 @@ sen_ngram_next(sen_lex *lex)
       }
 #endif /* PRE_DEFINED_UNSPLIT_WORDS */
       if (!(cl = sen_str_charlen((char *)p, lex->encoding))) {
-        lex->status = sen_lex_not_found;
+        lex->status = sen_lex_done;
         return SEN_SYM_NIL;
       }
       r = p + cl;
@@ -183,19 +175,11 @@ sen_ngram_next(sen_lex *lex)
       {
         size_t blen = q - p;
         if (!blen) {
-          lex->status = sen_lex_not_found;
-          return SEN_SYM_NIL;
-        }
-        if (!(buf = SEN_MALLOC(blen + 1))) { return SEN_SYM_NIL; }
-        memcpy(buf, p, blen);
-        buf[blen] = '\0';
-        // sen_log("other(%s), pos=%d, %d", buf, pos, cp - lex->nstr->ctypes);
-        tid = (lex->flags & SEN_LEX_ADD) ? sen_sym_get(sym, buf) : sen_sym_at(sym, buf);
-        SEN_FREE(buf);
-        if (tid == SEN_SYM_NIL) {
-          lex->status = sen_lex_not_found;
+          lex->status = sen_lex_done;
           return SEN_SYM_NIL;
         }
+        LEX_TOKEN(lex, p, blen);
+        tid = (lex->flags & SEN_LEX_ADD) ? sen_sym_get(sym, lex->token) : sen_sym_at(sym, lex->token);
         lex->skip = 1;
       }
     }
@@ -205,7 +189,11 @@ sen_ngram_next(sen_lex *lex)
     lex->next = r;
     // printf("tid=%d pos=%d tail=%d (%s) %s\n", tid, lex->pos, lex->tail, _sen_sym_key(sym, tid), r);
     // printf("tid=%d pos=%d tail=%d (%s)\n", tid, lex->pos, lex->tail, _sen_sym_key(sym, tid));
-    if (!*r) { lex->status = sen_lex_done; }
+    if (!tid) {
+      lex->status = sen_lex_not_found;
+    } else {
+      if (!*r) { lex->status = sen_lex_done; }
+    }
     return tid;
   }
   lex->status = sen_lex_done;
@@ -234,6 +222,7 @@ sen_mecab_open(sen_sym *sym, sen_nstr *nstr, uint8_t flags)
   unsigned int bufsize, maxtrial = 10, len;
   char *buf, *s, *p;
   sen_lex *lex;
+  sen_ctx *ctx = nstr->ctx;
   if (!(lex = SEN_MALLOC(sizeof(sen_lex)))) { return NULL; }
   lex->sym = sym;
   // sen_log("(%s)", str);
@@ -244,6 +233,8 @@ sen_mecab_open(sen_sym *sym, sen_nstr *nstr, uint8_t flags)
   }
   lex->mecab = sole_mecab;
   lex->buf = NULL;
+  lex->token = NULL;
+  lex->tlen = 0;
   // if (!(lex->mecab = mecab_new3())) {
   lex->pos = -1;
   lex->offset = 0;
@@ -283,13 +274,13 @@ sen_mecab_open(sen_sym *sym, sen_nstr *nstr, uint8_t flags)
 inline static sen_id
 sen_mecab_next(sen_lex *lex)
 {
-  char *buf;
   sen_id tid;
   sen_sym *sym = lex->sym;
+  sen_ctx *ctx = lex->nstr->ctx;
   uint32_t size;
   int32_t len, offset = lex->offset + lex->len;
   const unsigned char *p;
-  if (lex->status) { return SEN_SYM_NIL; }
+  if (lex->status == sen_lex_done) { return SEN_SYM_NIL; }
   for (p = lex->next, len = 0;;) {
     size_t cl;
     if (!(cl = sen_str_charlen((char *)p, lex->encoding)) ||
@@ -304,20 +295,18 @@ sen_mecab_next(sen_lex *lex)
     return SEN_SYM_NIL;
   }
   size = (uint32_t)(p - lex->next);
-  if (!(buf = SEN_MALLOC(size + 1))) { return SEN_SYM_NIL; }
-  memcpy(buf, lex->next, size);
-  buf[size] = '\0';
-  // sen_log("size=%d (%s)", size, buf);
-  tid = (lex->flags & SEN_LEX_ADD) ? sen_sym_get(sym, buf) : sen_sym_at(sym, buf);
-  SEN_FREE(buf);
-  if (tid == SEN_SYM_NIL) {
-    lex->status = sen_lex_not_found;
-  } else {
+  LEX_TOKEN(lex, lex->next, size);
+  tid = (lex->flags & SEN_LEX_ADD) ? sen_sym_get(sym, lex->token) : sen_sym_at(sym, lex->token);
+  {
     int cl;
     while ((cl = sen_isspace(p, lex->encoding))) { p += cl; }
     lex->next = p;
     lex->offset = offset;
     lex->len = len;
+  }
+  if (tid == SEN_SYM_NIL) {
+    lex->status = sen_lex_not_found;
+  } else {
     if (!*p) { lex->status = sen_lex_done; }
   }
   lex->pos++;
@@ -333,6 +322,7 @@ sen_delimited_open(sen_sym *sym, sen_nstr *nstr, uint8_t flags)
 {
   int cl;
   sen_lex *lex;
+  sen_ctx *ctx = nstr->ctx;
   const char *p;
   if (!(lex = SEN_MALLOC(sizeof(sen_lex)))) { return NULL; }
   lex->sym = sym;
@@ -340,6 +330,8 @@ sen_delimited_open(sen_sym *sym, sen_nstr *nstr, uint8_t flags)
   lex->mecab = NULL;
 #endif /* NO_MECAB */
   lex->buf = NULL;
+  lex->token = NULL;
+  lex->tlen = 0;
   lex->pos = -1;
   lex->skip = 1;
   lex->tail = 0;
@@ -361,13 +353,13 @@ sen_delimited_open(sen_sym *sym, sen_nstr *nstr, uint8_t flags)
 inline static sen_id
 sen_delimited_next(sen_lex *lex)
 {
-  char *buf;
   sen_id tid;
   sen_sym *sym = lex->sym;
+  sen_ctx *ctx = lex->nstr->ctx;
   uint32_t size;
   int32_t len, offset = lex->offset + lex->len;
   const unsigned char *p;
-  if (lex->status) { return SEN_SYM_NIL; }
+  if (lex->status == sen_lex_done) { return SEN_SYM_NIL; }
   for (p = lex->next, len = 0;;) {
     size_t cl;
     if (!(cl = sen_str_charlen((char *)p, lex->encoding)) ||
@@ -382,20 +374,18 @@ sen_delimited_next(sen_lex *lex)
     return SEN_SYM_NIL;
   }
   size = (uint32_t)(p - lex->next);
-  if (!(buf = SEN_MALLOC(size + 1))) { return SEN_SYM_NIL; }
-  memcpy(buf, lex->next, size);
-  buf[size] = '\0';
-  // sen_log("size=%d (%s)", size, buf);
-  tid = (lex->flags & SEN_LEX_ADD) ? sen_sym_get(sym, buf) : sen_sym_at(sym, buf);
-  SEN_FREE(buf);
-  if (tid == SEN_SYM_NIL) {
-    lex->status = sen_lex_not_found;
-  } else {
+  LEX_TOKEN(lex, lex->next, size);
+  tid = (lex->flags & SEN_LEX_ADD) ? sen_sym_get(sym, lex->token) : sen_sym_at(sym, lex->token);
+  {
     int cl;
     while ((cl = sen_isspace(p, lex->encoding))) { p += cl; }
     lex->next = p;
     lex->offset = offset;
     lex->len = len;
+  }
+  if (tid == SEN_SYM_NIL) {
+    lex->status = sen_lex_not_found;
+  } else {
     if (!*p) { lex->status = sen_lex_done; }
   }
   lex->pos++;
@@ -409,7 +399,7 @@ sen_lex_init(void)
 {
 #ifndef NO_MECAB
   // char *arg[] = {"", "-Owakati"};
-  // return mecab_load_dictionary(2, arg) ? sen_success : sen_other_error;
+  // return mecab_load_dictionary(2, arg) ? sen_success : sen_external_error;
   sole_mecab = NULL;
   MUTEX_INIT(sole_mecab_lock);
 #endif /* NO_MECAB */
@@ -490,12 +480,17 @@ sen_lex_next(sen_lex *lex)
 sen_rc
 sen_lex_close(sen_lex *lex)
 {
-  if (!lex) { return sen_invalid_argument; }
-  if (lex->nstr) { sen_nstr_close(lex->nstr); }
-  // if (lex->mecab) { mecab_destroy(lex->mecab); }
-  if (lex->buf) { SEN_FREE(lex->buf); }
-  SEN_FREE(lex);
-  return sen_success;
+  if (lex) {
+    sen_ctx *ctx = lex->nstr->ctx;
+    if (lex->nstr) { sen_nstr_close(lex->nstr); }
+    // if (lex->mecab) { mecab_destroy(lex->mecab); }
+    if (lex->buf) { SEN_FREE(lex->buf); }
+    if (lex->token) { SEN_REALLOC(lex->token, 0); }
+    SEN_FREE(lex);
+    return sen_success;
+  } else {
+    return sen_invalid_argument;
+  }
 }
 
 sen_rc
@@ -514,12 +509,12 @@ sen_lex_validate(sen_sym *sym)
     SOLE_MECAB_CONFIRM;
     if (!sole_mecab) {
       SEN_LOG(sen_log_alert, "mecab_new failed on sen_lex_validate");
-      return sen_other_error;
+      return sen_external_error;
     }
     di = mecab_dictionary_info(sole_mecab);
     if (!di || !di->charset) {
       SEN_LOG(sen_log_alert, "mecab_dictionary_info failed on sen_lex_validate");
-      return sen_other_error;
+      return sen_external_error;
     }
     switch (di->charset[0]) {
       case 'u':
@@ -534,13 +529,13 @@ sen_lex_validate(sen_sym *sym)
         break;
       default:
         SEN_LOG(sen_log_alert, "unknown encoding %s on sen_lex_validate", di->charset);
-        return sen_other_error;
+        return sen_external_error;
     }
     if (enc != sym->encoding) {
       SEN_LOG(sen_log_alert,
               "dictionary encoding %s is differ from sym encoding %s",
               di->charset, sen_enctostr(sym->encoding));
-      return sen_other_error;
+      return sen_abnormal_error;
     }
   }
 #endif /* USE_MECAB_DICINFO */
