@@ -14,12 +14,23 @@
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
-#ifndef _SEN_QL_H
-#define _SEN_QL_H
+#ifndef SEN_QL_H
+#define SEN_QL_H
 
-#include <setjmp.h>
+#ifndef SEN_STORE_H
 #include "store.h"
+#endif /* SEN_STORE_H */
+
+#ifndef SEN_COM_H
 #include "com.h"
+#endif /* SEN_COM_H */
+
+#ifndef SEN_CTX_H
+#include "ctx.h"
+#endif /* SEN_CTX_H */
+
+#define __USE_ISOC99
+#include <math.h>
 
 #ifdef __cplusplus
 extern "C" {
@@ -32,10 +43,9 @@ extern "C" {
 #define sen_ql_records       0x12
 #define sen_ql_bulk          0x13
 #define sen_ql_int           0x14
-#define sen_ql_uint          0x15
-#define sen_ql_int64         0x16
+#define sen_ql_time          0x15
 #define sen_ql_float         0x17
-#define sen_ql_native_method 0x18
+#define sen_ql_snip          0x18
 #define sen_ql_query         0x19
 #define sen_ql_op            0x1a
 #define sen_ql_syntax        0x1b
@@ -43,57 +53,9 @@ extern "C" {
 #define sen_ql_promise       0x1d
 #define sen_ql_closure       0x1e
 #define sen_ql_continuation  0x1f
+#define sen_ql_index         0x30
 #define sen_ql_list          0x40
 
-typedef struct _sen_obj sen_obj;
-typedef struct _sen_ql_co sen_ql_co;
-
-typedef sen_obj *sen_ql_method_func(sen_ctx *, sen_obj *, sen_ql_co *);
-
-struct _sen_obj {
-  uint8_t type;
-  uint8_t flags;
-  uint16_t nrefs;
-  sen_id class;
-  union {
-    struct {
-      sen_id self;
-      sen_ql_method_func *func;
-    } o;
-    struct {
-      sen_records *records; // todo : unify with o;
-    } r;
-    struct {
-      char *value;
-      uint32_t size;
-    } b;
-    struct {
-      sen_obj *car;
-      sen_obj *cdr;
-    } l;
-    struct {
-      int32_t i;
-    } i;
-    struct {
-      double d;
-    } d;
-    struct {
-      sen_query *query;
-    } q;
-    struct {
-      int8_t op;
-      int8_t mode;
-      int16_t weight;
-      int32_t option;
-    } op;
-  } u;
-};
-
-#define SEN_QL_HEAD 1
-#define SEN_QL_TAIL 2
-
-#define SEN_QL_MORE 1
-
 #define SEN_QL_CO_BEGIN(c) if (c) { switch((c)->last) { case 0:
 #define SEN_QL_CO_WAIT(c,d) \
   (c)->last = __LINE__; (c)->data = (d); return NULL;\
@@ -106,18 +68,13 @@ struct _sen_obj {
 #define SEN_QL_GET_MODE(c) ((c)->feed_mode)
 #define SEN_QL_GET_STAT(c) ((c)->stat)
 
-struct _sen_ql_co {
-  uint16_t mode;
-  uint16_t last;
-  void *data;
-};
-
 enum {
   sen_ql_atonce = 0,
   sen_ql_step
 };
 
 #define SEN_QL_TOPLEVEL  0x00
+#define SEN_QL_QUITTING  0x0f
 #define SEN_QL_EVAL      0x40
 #define SEN_QL_NATIVE    0x41
 #define SEN_QL_WAIT_EXPR 0xc0
@@ -126,62 +83,6 @@ enum {
 
 #define SEN_QL_WAITINGP(c) ((c)->stat & 0x80)
 
-struct _sen_ctx {
-
-  jmp_buf error_jmp;
-
-  uint32_t seqno;
-  uint32_t lseqno;
-  uint32_t nbinds;
-  uint32_t nunbinds;
-  uint8_t feed_mode;
-  uint8_t stat;
-
-  uint8_t batchmode;
-  uint8_t gc_verbose;
-
-  sen_encoding encoding;
-
-  int tok;
-  char *cur;
-  char *str_end;
-  sen_obj **pht;
-
-  sen_obj arg;
-  sen_db *db;
-  sen_set *objects;
-  sen_set *symbols;
-  sen_ql_method_func *doing;
-  sen_obj *phs;            /* list of placeholders */
-  sen_ql_co co;
-
-  sen_obj *args;      /* register for arguments of function */
-  sen_obj *envir;      /* stack register for current environment */
-  sen_obj *code;      /* register for current code */
-  sen_obj *dump;      /* stack register for next evaluation */
-  sen_obj *value;
-  sen_obj *global_env;
-  uint8_t op;
-
-  char *inbuf;
-  sen_rbuf outbuf;
-  sen_rbuf subbuf;
-  unsigned int bufcur;
-
-  void (*output)(sen_ctx *, int, void *);
-
-  sen_com_sqtp *com;
-  unsigned int com_status;
-  unsigned int com_info;
-
-  union {
-    void *ptr;
-    int fd;
-    uint32_t u32;
-    uint64_t u64;
-  } data;
-};
-
 extern sen_obj *sen_ql_nil;  /* special cell representing empty cell */
 extern sen_obj *sen_ql_t;    /* special cell representing #t */
 extern sen_obj *sen_ql_f;    /* special cell representing #f */
@@ -199,51 +100,47 @@ extern sen_obj *sen_ql_f;    /* special cell representing #f */
 #define SEN_OBJ_REFERER 32
 #define SEN_OBJ_NATIVE 64
 
-
 void sen_ql_init_globals(sen_ctx *c);
 
 sen_obj *sen_ql_at(sen_ctx *c, const char *key);
-sen_obj *sen_obj_new(sen_ctx *c);
-sen_obj *sen_obj_alloc(sen_ctx *c, uint32_t size);
-sen_obj *sen_obj_clear(sen_obj *o);
-void sen_ql_def_db_methods(sen_ctx *c);
-
-sen_rc sen_ctx_recv_handler_set(sen_ctx *c, void (*func)(sen_ctx *, int, void *),
-                                void *func_arg);
-
-void sen_ctx_stream_out_func(sen_ctx *c, int flags, void *stream);
+void sen_ql_def_db_funcs(sen_ctx *c);
 
 #define SEN_OBJ_INSPECT_ESC 1
 #define SEN_OBJ_INSPECT_SYM_AS_STR 2
 
 void sen_obj_inspect(sen_ctx *c, sen_obj *obj, sen_rbuf *buf, int flags);
-void sen_ql_def_native_method(sen_ctx *c, const char *name,
-                                  sen_ql_method_func *func);
+void sen_ql_def_native_func(sen_ctx *c, const char *name,
+                                  sen_ql_native_func *func);
 const char *_sen_obj_key(sen_ctx *c, sen_obj *obj);
 sen_obj *sen_ql_feed(sen_ctx *c, char *str, uint32_t str_size, int mode);
-sen_obj *sen_ql_sexp_parser(sen_ctx *c, const char *str, uint32_t str_size,
-                               sen_obj **phs);
-sen_rc sen_obj2int(sen_obj *o);
+sen_obj *sen_ql_eval(sen_ctx *c, sen_obj *code, sen_obj *objs);
+sen_rc sen_obj2int(sen_ctx *c, sen_obj *o);
 
 sen_obj *sen_ql_mk_symbol(sen_ctx *c, const char *name);
 sen_obj *sen_ql_mk_obj(sen_ctx *c, sen_id cls, sen_id self);
 void sen_ql_bind_symbol(sen_db_store *dbs, sen_obj *symbol);
 sen_obj *sen_ql_mk_string(sen_ctx *c, const char *str, unsigned int len);
 
+sen_obj *sen_obj_new(sen_ctx *c);
+sen_obj *sen_obj_alloc(sen_ctx *c, uint32_t size);
+void sen_obj_clear(sen_ctx *c, sen_obj *o);
+sen_obj *sen_obj_cons(sen_ctx *ctx, sen_obj *a, sen_obj *b);
+
 void sen_ql_init_const(void);
 
 #define SEN_OBJ2VALUE(o,v,s) ((v) = (o)->u.b.value, (s) = (o)->u.b.size)
 #define SEN_VALUE2OBJ(o,v,s) ((o)->u.b.value = (v), (o)->u.b.size = (s))
 
-#define NILP(c) ((c) == NIL || !(c) || (c)->type == sen_ql_void)
+#define VOIDP(c) ((c) == NIL || !(c) || (c)->type == sen_ql_void)
 #define OBJECTP(c) ((c)->type == sen_ql_object)
 #define RECORDSP(c) ((c)->type == sen_ql_records)
+#define SNIPP(c) ((c)->type == sen_ql_snip)
 #define BULKP(c) ((c)->type == sen_ql_bulk)
 #define LISTP(c) (((c)->type & sen_ql_list))
 #define INTP(c) ((c)->type == sen_ql_int)
 #define CLASSP(c) ((c)->type == sen_db_class)
 #define RAW_CLASSP(c) ((c)->type == sen_db_raw_class)
-#define NATIVE_METHODP(c) ((c)->type == sen_ql_native_method)
+#define NATIVE_FUNCP(c) ((c)->flags & SEN_OBJ_NATIVE)
 #define QUERYP(c) ((c)->type == sen_ql_query)
 #define OBJ_SLOTP(c) ((c)->type == sen_db_obj_slot)
 #define RA_SLOTP(c) ((c)->type == sen_db_ra_slot)
@@ -251,22 +148,62 @@ void sen_ql_init_const(void);
 #define IDX_SLOTP(c) ((c)->type == sen_db_idx_slot)
 #define SLOTP(c) (sen_db_obj_slot <= (c)->type && (c)->type <= sen_db_pslot)
 #define OPP(c) ((c)->type == sen_ql_op)
+#define STRVALUE(c) ((c)->u.b.value)
+#define NUMBERP(c) ((c)->type == sen_ql_int || (c)->type == sen_ql_float)
+#define SYMBOLP(c) ((c)->flags & SEN_OBJ_SYMBOL)
+#define SYMNAME(c) (SEN_SET_STRKEY_BY_VAL(c))
+#define KEYWORDP(c) (*SYMNAME(c) == ':')
 
-#define SETNIL(c) ((c)->type = sen_ql_void)
 #define SETINT(c,v) ((c)->type = sen_ql_int, (c)->u.i.i = (v))
+#define SETFLOAT(c,v) ((c)->type = sen_ql_float, (c)->u.d.d = (v))
 #define SETBULK(c,v,s) ((c)->type = sen_ql_bulk, (c)->u.b.value = (v), (c)->u.b.size = (s))
+#define SETTIME(c,v) ((c)->type = sen_ql_time, memcpy(&(c)->u.tv, v, sizeof(sen_timeval)))
+
+#define CAR(c)          ((c)->u.l.car)
+#define CDR(c)          ((c)->u.l.cdr)
+#define CAAR(c)         CAR(CAR(c))
+#define CADR(c)         CAR(CDR(c))
+#define CDAR(c)         CDR(CAR(c))
+#define CDDR(c)         CDR(CDR(c))
+#define CADDR(c)        CAR(CDDR(c))
+#define CADAR(p)        CAR(CDAR(p))
+#define CADAAR(p)       CAR(CDR(CAAR(p)))
+#define CADDDR(p)       CAR(CDR(CDDR(p)))
+#define CDDDDR(p)       CDR(CDR(CDDR(p)))
+
+#define IVALUE(p)       ((p)->u.i.i)
+#define FVALUE(p)       ((p)->u.d.d)
+
+#define SEN_OBJ_NEW(ctx,o) do {\
+  if (!((o) = sen_obj_new(ctx))) { QLERR("obj_new failed"); }\
+} while(0)
+
+#define SEN_RBUF2OBJ(ctx,rbuf,o) do {\
+  SEN_RBUF_PUTC((rbuf), '\0');\
+  SEN_OBJ_NEW(ctx, (o));\
+  (o)->flags = SEN_OBJ_ALLOCATED;\
+  (o)->type = sen_ql_bulk;\
+  (o)->u.b.value = (rbuf)->head;\
+  (o)->u.b.size = SEN_RBUF_VSIZE(rbuf) - 1;\
+} while(0)
+
+#define CONS(a,b) (sen_obj_cons(ctx, a, b))
 
-#define CAR(c) ((c)->u.l.car)
-#define CDR(c) ((c)->u.l.cdr)
-#define CAAR(c) (CAR(CAR(c)))
-#define CADR(c) (CAR(CDR(c)))
-#define CDAR(c) (CDR(CAR(c)))
-#define CDDR(c) (CDR(CDR(c)))
-#define CADDR(c) (CAR(CDDR(c)))
 #define POP(x,c) ((x) = CAR(c), (c) = CDR(c), (x))
 
+#define SKIPSPACE(c) do {\
+  unsigned int len;\
+  while ((c)->cur < (c)->str_end && sen_isspace((c)->cur, (c)->encoding)) {\
+    if (!(len = sen_str_charlen_nonnull((c)->cur, (c)->str_end, (c)->encoding))) {\
+      (c)->cur = (c)->str_end;\
+      break;\
+    }\
+    (c)->cur += len;\
+  }\
+} while (0)
+
 #ifdef __cplusplus
 }
 #endif
 
-#endif /* _SEN_QL_H */
+#endif /* SEN_QL_H */
