@@ -5,7 +5,7 @@
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.
 
-  This library is distributed in the hope that it will be useful,
+nnn  This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
@@ -79,8 +79,10 @@ cons(sen_query *q, cell *car, cell *cdr)
     c->type = sen_ql_list;
     c->u.l.car = car;
     c->u.l.cdr = cdr;
+    return c;
+  } else {
+    return NIL;
   }
-  return c;
 }
 
 inline static cell *
@@ -94,8 +96,10 @@ token_new(sen_query *q, const char *start, const char *end)
     c->u.b.value = (char *)start;
     c->u.b.size = len;
     q->cur_expr++;
+    return c;
+  } else {
+    return NIL;
   }
-  return c;
 }
 
 inline static cell *
@@ -108,8 +112,10 @@ op_new(sen_query *q, int8_t op, int16_t weight, int8_t mode, int32_t option)
     c->u.op.weight = weight;
     c->u.op.mode = mode;
     c->u.op.option = option;
+    return c;
+  } else {
+    return NIL;
   }
-  return c;
 }
 
 inline static void
@@ -129,18 +135,28 @@ skip_space(sen_query *q)
 inline static cell *
 get_phrase(sen_query *q)
 {
-  const char *start = q->cur, *end;
-  for (end = q->cur;; end++) {
-    if (end >= q->str_end) {
-      q->cur = end;
+  char *d = (char *)q->cur;
+  const char *start, *s;
+  start = s = q->cur;
+  while (1) {
+    unsigned int len;
+    if (s >= q->str_end) {
+      q->cur = s;
       break;
     }
-    if (*end == SEN_QUERY_QUOTER) {
-      q->cur = end + 1;
-      break;
+    len = sen_str_charlen_nonnull(s, q->str_end, q->encoding);
+    if (len == 1) {
+      if (*s == SEN_QUERY_QUOTER) {
+        q->cur = s + 1;
+        break;
+      } else if (*s == SEN_QUERY_ESCAPE && s + 1 < q->str_end) {
+        s++;
+        len = sen_str_charlen_nonnull(s, q->str_end, q->encoding);
+      }
     }
+    while (len--) { *d++ = *s++; }
   }
-  return token_new(q, start, end);
+  return token_new(q, start, d);
 }
 
 inline static cell *
@@ -209,7 +225,7 @@ get_token(sen_query *q)
 {
   cell *token = NIL;
   sen_sel_operator op = q->default_op;
-  do {
+  {
     int weight = DEFAULT_WEIGHT, prefixp = 0, mode = -1, option = 0;
     skip_space(q);
     if (q->cur_expr >= q->max_exprs ||
@@ -266,7 +282,7 @@ get_token(sen_query *q)
       }
       break;
     }
-  } while (!token);
+  }
   return cons(q, token, NIL);
 }
 
@@ -283,6 +299,7 @@ get_expr(sen_query *q)
 static const char *
 get_weight_vector(sen_query *query, const char *source)
 {
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
   const char *p;
 
   if (!query->opt.weight_vector &&
@@ -394,6 +411,7 @@ sen_query *
 sen_query_open(const char *str, unsigned int str_len,
                sen_sel_operator default_op, int max_exprs, sen_encoding encoding)
 {
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
   sen_query *q;
   int max_cells = max_exprs * 4;
   if (!(q = SEN_MALLOC(sizeof(sen_query) + max_cells * sizeof(cell)))) {
@@ -409,7 +427,7 @@ sen_query_open(const char *str, unsigned int str_len,
   q->max_cells = max_cells;
   q->cur_cell = 0;
   q->cur_expr = 0;
-  q->escalation_threshold = sen_query_escalation_threshold;
+  q->escalation_threshold = SENNA_DEFAULT_QUERY_ESCALATION_THRESHOLD;
   q->escalation_decaystep = DEFAULT_DECAYSTEP;
   q->weight_offset = 0;
   q->opt.weight_vector = NULL;
@@ -426,6 +444,7 @@ sen_query_open(const char *str, unsigned int str_len,
 unsigned int
 sen_query_rest(sen_query *q, const char ** const rest)
 {
+  if (!q) { return 0; }
   if (rest) {
     *rest = q->cur;
   }
@@ -435,6 +454,7 @@ sen_query_rest(sen_query *q, const char ** const rest)
 sen_rc
 sen_query_close(sen_query *q)
 {
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
   if (!q) { return sen_invalid_argument; }
   if (q->opt.weight_vector) {
     SEN_FREE(q->opt.weight_vector);
@@ -476,6 +496,9 @@ exec_query(sen_index *i, sen_query *q, cell *c, sen_records *r, sen_sel_operator
         if (!q->opt.weight_vector) {
           q->opt.vector_size = ope->u.op.weight + q->weight_offset;
         }
+        if (ope->u.op.mode == sen_sel_similar) {
+          q->opt.max_interval = q->default_mode;
+        }
       } else {
         q->opt.mode = q->default_mode;
         q->opt.max_interval = DEFAULT_MAX_INTERVAL;
@@ -557,8 +580,8 @@ query_term_rec(sen_query* q, cell* c, query_term_callback func, void *func_arg)
   if (BULKP(c)) {
     return func(c->u.b.value, c->u.b.size, func_arg);
   }
-  for (token = c->u.l.car; token; token = token->u.l.cdr) {
-    if (!query_term_rec(q, token, func, func_arg)) {
+  for (token = c; LISTP(token); token = CDR(token)) {
+    if (!query_term_rec(q, CAR(token), func, func_arg)) {
       return 0; /* abort */
     }
   }
@@ -700,6 +723,7 @@ scan_query(sen_query *q, sen_nstr *nstr, sen_id section, cell *c, snip_cond **sc
 static sen_rc
 alloc_snip_conds(sen_query *q)
 {
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
   if (!(q->snip_conds = SEN_CALLOC(sizeof(snip_cond) * q->cur_expr))) {
     SEN_LOG(sen_log_alert, "snip_cond allocation failed");
     return sen_memory_exhausted;
