@@ -14,15 +14,15 @@
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
-#ifndef _SENNA_IN_H
-#define _SENNA_IN_H
+#ifndef SENNA_IN_H
+#define SENNA_IN_H
 #include "config.h"
 
 #ifdef USE_AIO
 /* #define __USE_XOPEN2K 1 */
 #define _XOPEN_SOURCE 600
 #else
-#if !defined(__APPLE__) && !defined(__FreeBSD)
+#if !defined(__APPLE__) && !defined(__FreeBSD__)
 #define _XOPEN_SOURCE 500
 #endif
 #endif /* USE_AIO */
@@ -62,6 +62,7 @@
 #endif /* HAVE_SYS_RESOURCE_H */
 
 #ifdef WIN32
+
 #pragma warning(disable: 4996)
 #include <io.h>
 #include <basetsd.h>
@@ -70,6 +71,7 @@
 #include <windows.h>
 #include <stddef.h>
 #include <windef.h>
+#include <float.h>
 #define PATH_MAX (MAX_PATH - 1)
 #define inline _inline
 #define snprintf _snprintf
@@ -86,6 +88,7 @@
 #define close _close
 #define usleep Sleep
 #define sleep(x) Sleep((x) * 1000)
+#define fpclassify _fpclass
 #define uint8_t UINT8
 #define int8_t INT8
 #define int_least8_t INT8
@@ -100,7 +103,20 @@
 #undef MSG_WAITALL
 #define MSG_WAITALL 0
 #define PATH_SEPARATOR "\\"
+#define CASE_FP_NAN case _FPCLASS_SNAN: case _FPCLASS_QNAN:
+#define CASE_FP_INFINITE case _FPCLASS_NINF: case _FPCLASS_PINF:
+#define SHUT_RDWR SD_BOTH
+
+#define islessgreater(x,y)  (!_isnan(x) && !_isnan(y) && ((x) < (y) || (x) > (y)))
+#define isless(x,y)         (!_isnan(x) && !_isnan(y) && ((x) < (y)))
+#define isgreater(x,y)      (!_isnan(x) && !_isnan(y) && ((x) > (y)))
+typedef SOCKET sen_sock;
+#define sen_sock_close closesocket
+
+#define CALLBACK __stdcall
+
 #else /* WIN32 */
+
 #ifndef PATH_MAX
 #if defined(MAXPATHLEN)
 #define PATH_MAX MAXPATHLEN
@@ -115,16 +131,36 @@ typedef char int_least8_t;
 typedef unsigned char uint_least8_t;
 #endif /* UINT_LEAST8_MAX */
 #define PATH_SEPARATOR "/"
+#define CASE_FP_NAN case FP_NAN:
+#define CASE_FP_INFINITE case FP_INFINITE:
+typedef int sen_sock;
+#define sen_sock_close close
+#define CALLBACK
 #endif /* WIN32 */
 
+#ifndef INT32_MAX
+#define INT32_MAX (2147483647)
+#endif /* INT32_MAX */
+
+#ifndef INT32_MIN
+#define INT32_MIN (-2147483648)
+#endif /* INT32_MIN */
+
 #ifdef HAVE_PTHREAD_H
 #include <pthread.h>
+typedef pthread_t sen_thread;
 typedef pthread_mutex_t sen_mutex;
+#define THREAD_CREATE(thread,func,arg) (pthread_create(&(thread), NULL, (func), (arg)))
 #define MUTEX_INIT(m) pthread_mutex_init(&m, NULL)
 #define MUTEX_LOCK(m) pthread_mutex_lock(&m)
 #define MUTEX_UNLOCK(m) pthread_mutex_unlock(&m)
 #define MUTEX_DESTROY(m)
 
+typedef pthread_cond_t sen_cond;
+#define COND_INIT(c) pthread_cond_init(&c, NULL)
+#define COND_SIGNAL(c) pthread_cond_signal(&c)
+#define COND_WAIT(c,m) pthread_cond_wait(&c, &m)
+
 typedef pthread_key_t sen_thread_key;
 #define THREAD_KEY_CREATE pthread_key_create
 #define THREAD_KEY_DELETE pthread_key_delete
@@ -141,14 +177,24 @@ typedef int sen_thread_key;
 #define THREAD_GETSPECIFIC(key,value)
 
 #ifdef WIN32
+typedef uintptr_t sen_thread;
 typedef CRITICAL_SECTION sen_mutex;
+#define THREAD_CREATE(thread,func,arg) (((thread)=_beginthreadex(NULL, 0, (func), (arg), 0, NULL)) == NULL)
 #define MUTEX_INIT(m) InitializeCriticalSection(&m)
 #define MUTEX_LOCK(m) EnterCriticalSection(&m)
 #define MUTEX_UNLOCK(m) LeaveCriticalSection(&m)
 #define MUTEX_DESTROY(m) DeleteCriticalSection(&m)
+
 #else /* WIN32 */
 /* todo */
 #endif /* WIN32 */
+
+typedef int sen_cond;
+#define COND_INIT(c) ((c) = 0)
+#define COND_SIGNAL(c)
+#define COND_WAIT(c,m) do { MUTEX_UNLOCK(m); usleep(1000); MUTEX_LOCK(m); } while (0)
+/* todo : must be enhanced! */
+
 #endif /* HAVE_PTHREAD_H */
 
 #ifdef __GNUC__
@@ -156,11 +202,16 @@ typedef CRITICAL_SECTION sen_mutex;
 #if (defined(__i386__) || defined(__x86_64__)) /* ATOMIC ADD */
 #define SEN_ATOMIC_ADD_EX(p,i,r) \
   __asm__ __volatile__ ("lock; xaddl %0,%1" : "=r"(r), "=m"(*p) : "0"(i), "m" (*p))
+#define SEN_BIT_SCAN_REV(v,r) \
+  __asm__ __volatile__ ("\tmovl %1,%%eax\n\tbsr %%eax,%0\n" : "=r"(r) : "r"(v) : "%eax")
 #elif (defined(__PPC__) || defined(__ppc__)) /* ATOMIC ADD */
 #define SEN_ATOMIC_ADD_EX(p,i,r) \
   __asm__ __volatile__ ("\n1:\n\tlwarx %0, 0, %1\n\tadd %0, %0, %2\n\tstwcx. %0, 0, %1\n\tbne- 1b\n\tsub %0, %0, %2" : "=&r" (r) : "r" (p), "r" (i) : "cc", "memory");
+/* todo */
+#define SEN_BIT_SCAN_REV(v,r)   for (r = 31; r && !((1 << r) & v); r--)
 #else /* ATOMIC ADD */
 /* todo */
+#define SEN_BIT_SCAN_REV(v,r)   for (r = 31; r && !((1 << r) & v); r--)
 #endif /* ATOMIC ADD */
 
 #ifdef __i386__ /* ATOMIC 64BIT SET */
@@ -182,16 +233,18 @@ typedef CRITICAL_SECTION sen_mutex;
   *(p) = (v);
 #endif /* ATOMIC 64BIT SET */
 
+/* todo */
+#define SEN_BIT_SCAN_REV(v,r)   for (r = 31; r && !((1 << r) & v); r--)
+
 #else /* __GNUC__ */
 /* todo */
+#define SEN_BIT_SCAN_REV(v,r)   for (r = 31; r && !((1 << r) & v); r--)
 #endif /* __GNUC__ */
 
 typedef uint8_t byte;
 
-#ifndef _SENNA_H
+#ifndef SENNA_H
 #include "senna.h"
-#endif /* _SENNA_H */
+#endif /* SENNA_H */
 
-extern int sen_query_escalation_threshold;
-
-#endif /* _SENNA_IN_H */
+#endif /* SENNA_IN_H */
