@@ -19,6 +19,7 @@
 #include <stddef.h>
 #include <assert.h>
 #include "snip.h"
+#include "ctx.h"
 
 #if !defined MAX
 #define MAX(a, b) ((a) > (b) ? (a) : (b))
@@ -108,7 +109,7 @@ sen_bm_preBmBc(const unsigned char *x, size_t m, size_t *bmBc)
     cond->end_offset = offset; \
     cond->found = found + shift; \
     cond->found_alpha_head = found_alpha_head; \
-    /* printf("bm: cond:%p found:%d last_found:%d st_off:%d ed_off:%d\n",cond, cond->found,cond->last_found,cond->start_offset,cond->end_offset); */ \
+    /* printf("bm: cond:%p found:%zd last_found:%zd st_off:%zd ed_off:%zd\n", cond, cond->found,cond->last_found,cond->start_offset,cond->end_offset); */ \
     return; \
   }
 
@@ -264,9 +265,11 @@ sen_snip_cond_init(snip_cond *sc, const char *keyword, unsigned int keyword_len,
     sen_snip_cond_close(sc);
     return sen_invalid_argument;
   }
-  sen_bm_preBmBc((unsigned char *)sc->keyword->norm, norm_blen, sc->bmBc);
-  sc->shift = sc->bmBc[(unsigned char)sc->keyword->norm[norm_blen - 1]];
-  sc->bmBc[(unsigned char)sc->keyword->norm[norm_blen - 1]] = 0;
+  if (norm_blen != 1) {
+    sen_bm_preBmBc((unsigned char *)sc->keyword->norm, norm_blen, sc->bmBc);
+    sc->shift = sc->bmBc[(unsigned char)sc->keyword->norm[norm_blen - 1]];
+    sc->bmBc[(unsigned char)sc->keyword->norm[norm_blen - 1]] = 0;
+  }
   return sen_success;
 }
 
@@ -289,11 +292,12 @@ sen_snip_add_cond(sen_snip *snip,
                   const char *closetag, unsigned int closetag_len)
 {
   sen_rc rc;
-  snip_cond *cond = snip->cond + snip->cond_len;
+  snip_cond *cond;
 
-  if (!keyword || !*keyword || snip->cond_len >= MAX_SNIP_COND_COUNT) {
+  if (!snip || !keyword || !*keyword || snip->cond_len >= MAX_SNIP_COND_COUNT) {
     return sen_invalid_argument;
   }
+  cond = snip->cond + snip->cond_len;
   if ((rc = sen_snip_cond_init(cond, keyword, keyword_len,
                                snip->encoding, snip->flags))) {
     return rc;
@@ -350,6 +354,7 @@ sen_snip_open(sen_encoding encoding, int flags, unsigned int width,
               const char *defaultclosetag, unsigned int defaultclosetag_len,
               sen_snip_mapping *mapping)
 {
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
   sen_snip *ret = NULL;
   if (!(ret = SEN_MALLOC(sizeof(sen_snip)))) {
     SEN_LOG(sen_log_alert, "sen_snip allocation failed on sen_snip_open");
@@ -379,14 +384,15 @@ sen_snip_open(sen_encoding encoding, int flags, unsigned int width,
 static sen_rc
 exec_clean(sen_snip *snip)
 {
-  snip_cond *cond;
+  snip_cond *cond, *cond_end;
   if (snip->nstr) {
     sen_nstr_close(snip->nstr);
     snip->nstr = NULL;
   }
   snip->tag_count = 0;
   snip->snip_count = 0;
-  for (cond = snip->cond; (unsigned int) (cond - snip->cond) < snip->cond_len; cond++) {
+  for (cond = snip->cond, cond_end = cond + snip->cond_len;
+       cond < cond_end; cond++) {
     sen_snip_cond_reinit(cond);
   }
   return sen_success;
@@ -395,16 +401,18 @@ exec_clean(sen_snip *snip)
 sen_rc
 sen_snip_close(sen_snip *snip)
 {
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
   snip_cond *cond, *cond_end;
   if (!snip) {
     return sen_invalid_argument;
   }
-  cond = cond_end = snip->cond;
-  cond_end += snip->cond_len;
-  for (; cond < cond_end; cond++) {
+  if (snip->nstr) {
+    sen_nstr_close(snip->nstr);
+  }
+  for (cond = snip->cond, cond_end = cond + snip->cond_len;
+       cond < cond_end; cond++) {
     sen_snip_cond_close(cond);
   }
-  exec_clean(snip);
   SEN_FREE(snip);
   return sen_success;
 }
@@ -414,13 +422,11 @@ sen_snip_exec(sen_snip *snip, const char *string, unsigned int string_len,
               unsigned int *nresults, unsigned int *max_tagged_len)
 {
   size_t i;
-
-  exec_clean(snip);
-
-  *nresults = 0;
-  if (!string || !*string) {
+  if (!snip || !string || !*string) {
     return sen_invalid_argument;
   }
+  exec_clean(snip);
+  *nresults = 0;
   if (snip->flags & SEN_SNIP_NORMALIZE) {
     snip->nstr =
       sen_nstr_open(string, string_len, snip->encoding,
@@ -541,11 +547,13 @@ sen_snip_exec(sen_snip *snip, const char *string, unsigned int string_len,
       snip_result->start_offset =
         sen_snip_find_firstbyte(string, snip->encoding, snip_result->start_offset, 1);
 
-      snip_result->end_offset = MIN(string_len, snip_result->start_offset + snip->width);
-
-      snip_result->end_offset =
-        sen_snip_find_firstbyte(string, snip->encoding, snip_result->end_offset, -1);
-
+      snip_result->end_offset = snip_result->start_offset + snip->width;
+      if (snip_result->end_offset < string_len) {
+        snip_result->end_offset =
+          sen_snip_find_firstbyte(string, snip->encoding, snip_result->end_offset, -1);
+      } else {
+        snip_result->end_offset = string_len;
+      }
       last_last_end_offset = snip_result->end_offset;
 
       tagged_len +=
