@@ -18,10 +18,70 @@
 #include "str.h"
 #include "inv.h"
 #include "sym.h"
-#include "str.h"
 #include "store.h"
+#include "ctx.h"
 #include <string.h>
 
+/* simple array */
+
+void
+sen_array_init(sen_array *a, uint16_t element_size, uint16_t flags)
+{
+  a->element_size = element_size;
+  a->flags = flags;
+  MUTEX_INIT(a->lock);
+  memset(a->elements, 0, sizeof(a->elements));
+}
+
+void
+sen_array_fin(sen_array *a)
+{
+  int i;
+  for (i = 0; i < 8; i++) {
+    void **e = &a->elements[i];
+    if (*e) {
+      SEN_GFREE(*e);
+      *e = NULL;
+    }
+  }
+}
+
+void *
+sen_array_at(sen_array *a, sen_id id)
+{
+  int l, i, o;
+  byte **e;
+  SEN_BIT_SCAN_REV(id, l);
+  i = l >> 2;
+  o = 1 << (i << 2);
+  e = (byte **)&a->elements[i];
+  if (!*e) {
+    MUTEX_LOCK(a->lock);
+    if (!*e) {
+      if (a->flags & sen_array_clear) {
+        *e = SEN_GCALLOC(0x0f * o * a->element_size);
+      } else {
+        *e = SEN_GMALLOC(0x0f * o * a->element_size);
+      }
+    }
+    MUTEX_UNLOCK(a->lock);
+    if (!*e) { return NULL; }
+  }
+  return *e + (id - o) * a->element_size;
+}
+
+#define SEN_ARRAY_EACH(a,k,v,block) do {\
+  int _ei;\
+  for (_ei = 0, (k) = 1; _ei < 8; _ei++) {\
+    int _ej = 0x0f * (1 << (_ei << 2));\
+    if (((v) = (a)->elements[_ei])) {\
+      for (; _ej--; (k)++, (v) = (void *)((byte *)(v) + (a)->element_size)) block\
+    } else {\
+      (k) += _ej;\
+    }\
+  }\
+} while (0)
+
 /* rectangular arrays */
 
 #define SEN_RA_IDSTR "SENNA:RA:01.000"
@@ -50,7 +110,7 @@ sen_ra_create(const char *path, unsigned int element_size)
   memcpy(header->idstr, SEN_RA_IDSTR, 16);
   header->element_size = actual_size;
   header->curr_max = 0;
-  if (!(ra = SEN_MALLOC(sizeof(sen_ra)))) {
+  if (!(ra = SEN_GMALLOC(sizeof(sen_ra)))) {
     sen_io_close(io);
     return NULL;
   }
@@ -78,7 +138,7 @@ sen_ra_open(const char *path)
     sen_io_close(io);
     return NULL;
   }
-  if (!(ra = SEN_MALLOC(sizeof(sen_ra)))) {
+  if (!(ra = SEN_GMALLOC(sizeof(sen_ra)))) {
     sen_io_close(io);
     return NULL;
   }
@@ -106,7 +166,7 @@ sen_ra_close(sen_ra *ra)
   sen_rc rc;
   if (!ra) { return sen_invalid_argument; }
   rc = sen_io_close(ra->io);
-  SEN_FREE(ra);
+  SEN_GFREE(ra);
   return rc;
 }
 
@@ -236,7 +296,7 @@ sen_ja_create(const char *path, unsigned int max_element_size)
   header->max_segments = max_segments;
   header->segments[0] = JA_SEG_ESEG;
   header->esegs[0] = 0;
-  if (!(ja = SEN_MALLOC(sizeof(sen_ja)))) {
+  if (!(ja = SEN_GMALLOC(sizeof(sen_ja)))) {
     sen_io_close(io);
     return NULL;
   }
@@ -259,7 +319,7 @@ sen_ja_open(const char *path)
     sen_io_close(io);
     return NULL;
   }
-  if (!(ja = SEN_MALLOC(sizeof(sen_ja)))) {
+  if (!(ja = SEN_GMALLOC(sizeof(sen_ja)))) {
     sen_io_close(io);
     return NULL;
   }
@@ -272,6 +332,7 @@ sen_rc
 sen_ja_info(sen_ja *ja, unsigned int *max_element_size)
 {
   if (!ja) { return sen_invalid_argument; }
+  if (max_element_size) { *max_element_size = ja->header->max_element_size; }
   return sen_success;
 }
 
@@ -281,7 +342,7 @@ sen_ja_close(sen_ja *ja)
   sen_rc rc;
   if (!ja) { return sen_invalid_argument; }
   rc = sen_io_close(ja->io);
-  SEN_FREE(ja);
+  SEN_GFREE(ja);
   return rc;
 }
 
@@ -311,9 +372,13 @@ sen_ja_put(sen_ja *ja, sen_id id, const void *value, int value_len, int flags)
       memcpy(buf, value, value_len);
     }
   } else {
-    if ((rc = sen_ja_alloc(ja, value_len, &einfo, &buf))) { return rc; }
-    // printf("put id=%d, value_len=%d value=%p ei=%p(%d:%d)\n", id, value_len, buf, &einfo, einfo.u.s.pos, einfo.u.s.tail[0]);
-    memcpy(buf, value, value_len);
+    if (value_len) {
+      if ((rc = sen_ja_alloc(ja, value_len, &einfo, &buf))) { return rc; }
+      // printf("put id=%d, value_len=%d value=%p ei=%p(%d:%d)\n", id, value_len, buf, &einfo, einfo.u.s.pos, einfo.u.s.tail[0]);
+      memcpy(buf, value, value_len);
+    } else {
+      memset(&einfo, 0, sizeof(sen_ja_einfo));
+    }
   }
   return sen_ja_replace(ja, id, &einfo);
 }
@@ -448,13 +513,31 @@ sen_ja_replace(sen_ja *ja, sen_id id, sen_ja_einfo *ei)
   if (!einfo) { return sen_memory_exhausted; }
   eback = einfo[pos];
   einfo[pos] = *ei;
-  // todo: SEN_ATOMIC_SET64
+  // SEN_SET_64BIT(&einfo[pos], *ei);
   sen_ja_free(ja, &eback);
   return sen_success;
 }
 
 /**** db ****/
 
+sen_rc
+sen_db_lock(sen_db *s, int timeout)
+{
+  return sen_sym_lock(s->keys, timeout);
+}
+
+sen_rc
+sen_db_unlock(sen_db *s)
+{
+  return sen_sym_unlock(s->keys);
+}
+
+sen_rc
+sen_db_clear_lock(sen_db *s)
+{
+  return sen_sym_clear_lock(s->keys);
+}
+
 inline static void
 gen_pathname(const char *path, char *buffer, int fno)
 {
@@ -471,9 +554,9 @@ gen_pathname(const char *path, char *buffer, int fno)
 sen_db_store *
 sen_db_store_by_id(sen_db *s, sen_id id)
 {
-  sen_db_store *slot;
   const char *name;
-  if (sen_set_at(s->stores, &id, (void **) &slot)) { return slot; }
+  sen_db_store *slot = sen_array_at(&s->stores, id);
+  if (slot->type) { return slot; }
   if (!(name = _sen_sym_key(s->keys, id))) { return NULL; }
   return sen_db_store_open(s, name);
 }
@@ -492,20 +575,49 @@ sen_db_slot_class(sen_db *s, const char *slot)
   return sen_db_store_open(s, buf);
 }
 
+sen_rc
+sen_db_idx_slot_build(sen_db *s, sen_db_store *store)
+{
+  sen_index *index = store->u.i.index;
+  sen_id id = SEN_SYM_NIL, maxid = sen_sym_curr_id(index->keys);
+  sen_db_store *target = sen_db_store_by_id(s, store->triggers->target);
+  if (!target) { return sen_invalid_argument; }
+  while (++id <= maxid) {
+    uint32_t vs;
+    const char *key = _sen_sym_key(index->keys, id);
+    void *vp = (void *)sen_ja_ref(target->u.v.ja, id, &vs);
+    if (key && vp && vs && sen_index_upd(index, key, NULL, 0, vp, vs)) {
+      SEN_LOG(sen_log_error, "sen_index_upd failed. id=%d key=(%s)", id, key);
+    }
+  }
+  return sen_success;
+}
+
 sen_db_store *
 sen_db_store_open(sen_db *s, const char *name)
 {
   sen_id id;
-  sen_set_eh *ep;
   uint32_t spec_len;
-  sen_db_store *e;
+  sen_db_store *e, *k, *l, *r = NULL;
   char buffer[PATH_MAX];
   const sen_db_store_spec *spec;
-  if (!s || !(id = sen_sym_at(s->keys, name))) { return NULL; }
-  if (sen_set_at(s->stores, &id, (void **) &e)) { return e; }
-  if (!(spec = sen_ja_ref(s->values, id, &spec_len))) { return NULL; }
-  ep = sen_set_get(s->stores, &id, (void **) &e);
-  e->type = spec->type;
+  if (!s || !(id = sen_sym_at(s->keys, name))) {
+    return NULL;
+  }
+  if (!(e = sen_array_at(&s->stores, id))) { return NULL; }
+  if (e->type) { return e; }
+
+  if (!(spec = sen_ja_ref(s->values, id, &spec_len))) {
+    return NULL;
+  }
+  if (spec->type == sen_db_idx_slot) {
+    if (!(k = sen_db_store_by_id(s, spec->u.s.class)) ||
+        !(l = sen_db_slot_class(s, name))) {
+      sen_ja_unref(s->values, id);
+      return NULL;
+    }
+  }
+  MUTEX_LOCK(s->lock);
   e->db = s;
   e->id = id;
   e->triggers = NULL;
@@ -531,14 +643,9 @@ sen_db_store_open(sen_db *s, const char *name)
     break;
   case sen_db_idx_slot :
     e->u.i.class = spec->u.s.class;
-    {
-      sen_db_store *l, *k;
-      if (!(k = sen_db_store_by_id(s, spec->u.s.class))) { goto exit; }
-      if (!(l = sen_db_slot_class(s, name))) { goto exit; }
-      if (!(e->u.i.index =
-            sen_index_open_with_keys_lexicon(buffer, k->u.c.keys, l->u.c.keys))) {
-        goto exit;
-      }
+    if (!(e->u.i.index =
+          sen_index_open_with_keys_lexicon(buffer, k->u.c.keys, l->u.c.keys))) {
+      goto exit;
     }
     break;
   case sen_db_rel1 :
@@ -551,9 +658,9 @@ sen_db_store_open(sen_db *s, const char *name)
   {
     int i;
     for (i = 0; i < spec->n_triggers; i++) {
-      sen_db_store *target = sen_db_store_by_id(s, spec->triggers[i].target);
+      sen_id target = spec->triggers[i].target;
       if (target) {
-        sen_db_trigger *r = SEN_MALLOC(sizeof(sen_db_trigger));
+        sen_db_trigger *r = SEN_GMALLOC(sizeof(sen_db_trigger));
         if (!r) { goto exit; }
         r->next = e->triggers;
         r->type = spec->triggers[i].type;
@@ -562,36 +669,62 @@ sen_db_store_open(sen_db *s, const char *name)
       }
     }
   }
-  sen_ja_unref(s->values, id);
-  return e;
+  e->type = spec->type;
+  r = e;
 exit :
-  sen_set_del(s->stores, ep);
   sen_ja_unref(s->values, id);
-  return NULL;
+  MUTEX_UNLOCK(s->lock);
+  return r;
 }
 
 sen_db_store *
 sen_db_store_create(sen_db *s, const char *name, sen_db_store_spec *spec)
 {
   sen_id id;
-  sen_set_eh *ep;
-  sen_db_store *e;
+  sen_db_store *e, *k, *l, *r = NULL;
   char buffer[PATH_MAX];
   if (strlen(name) >= SEN_SYM_MAX_KEY_SIZE) {
-    SEN_LOG(sen_log_error, "too long store name (%s)", name);
+    GERR(sen_invalid_argument, "too long store name (%s)", name);
     return NULL;
   }
   if (strchr(name, '.') &&
-      ((spec->type == sen_db_raw_class) ||
-       (spec->type == sen_db_class))) {
-    SEN_LOG(sen_log_error, "class name must not include '.' (%s)", name);
+      ((spec->type == sen_db_raw_class) || (spec->type == sen_db_class))) {
+    GERR(sen_invalid_argument, "class name must not include '.' (%s)", name);
     return NULL;
   }
-  if (!(id = sen_sym_get(s->keys, name))) { return NULL; }
-  if (!(ep = sen_set_get(s->stores, &id, (void **) &e))) { goto exit; }
+  if (spec->type == sen_db_idx_slot) {
+    if (!(k = sen_db_store_by_id(s, spec->u.s.class)) ||
+        !(l = sen_db_slot_class(s, name))) {
+      return NULL;
+    }
+  }
+
+  if (sen_db_lock(s, -1)) {
+    SEN_LOG(sen_log_crit, "sen_db_store_create: lock failed");
+    return NULL;
+  }
+
+  if (sen_sym_at(s->keys, name)) {
+    sen_db_unlock(s);
+    GERR(sen_invalid_argument, "sen_db_store_create: existing name (%s)", name);
+    return NULL;
+  }
+
+  if (!(id = sen_sym_nextid(s->keys, name))) {
+    sen_db_unlock(s);
+    return NULL;
+  }
+
+  MUTEX_LOCK(s->lock);
+
+  if (!(e = sen_array_at(&s->stores, id))) { goto exit; }
+  if (e->type) {
+    SEN_LOG(sen_log_crit, "sen_db corrupt: %s", name);
+    goto exit;
+  }
+
   spec->n_triggers = 0;
   if (sen_ja_put(s->values, id, spec, SEN_DB_STORE_SPEC_SIZE(0), 0)) { goto exit; }
-  e->type = spec->type;
   e->db = s;
   e->id = id;
   e->triggers = NULL;
@@ -620,15 +753,10 @@ sen_db_store_create(sen_db *s, const char *name, sen_db_store_spec *spec)
     break;
   case sen_db_idx_slot :
     e->u.i.class = spec->u.s.class;
-    {
-      sen_db_store *l, *k;
-      if (!(k = sen_db_store_by_id(s, spec->u.s.class))) { goto exit; }
-      if (!(l = sen_db_slot_class(s, name))) { goto exit; }
-      if (!(e->u.i.index =
-            sen_index_create_with_keys_lexicon(buffer, k->u.c.keys, l->u.c.keys,
-                                               spec->u.s.size))) {
-        goto exit;
-      }
+    if (!(e->u.i.index =
+          sen_index_create_with_keys_lexicon(buffer, k->u.c.keys, l->u.c.keys,
+                                             spec->u.s.size))) {
+      goto exit;
     }
     break;
   case sen_db_rel1 :
@@ -638,12 +766,16 @@ sen_db_store_create(sen_db *s, const char *name, sen_db_store_spec *spec)
   default :
     goto exit;
   }
-  return e;
+  if ((id != sen_sym_get(s->keys, name))) {
+    SEN_LOG(sen_log_crit, "sen_db id unmatch: %d", id);
+  }
+  e->type = spec->type;
+  r = e;
 exit :
-  if (ep) { sen_set_del(s->stores, ep); }
-  sen_sym_del(s->keys, name);
+  MUTEX_UNLOCK(s->lock);
+  sen_db_unlock(s);
   // todo : sen_ja_put(s->values, id, NULL, 0, 0);
-  return NULL;
+  return r;
 }
 
 sen_rc
@@ -654,44 +786,103 @@ sen_db_store_add_trigger(sen_db_store *e, sen_db_store_rel_spec *t)
   uint32_t spec_len, newspec_len;
   const sen_db_store_spec *spec;
   sen_db_store_spec *newspec;
+  sen_db_store *target = sen_db_store_by_id(s, t->target);
+  if (!target) { return sen_invalid_argument; }
+
+  if (sen_db_lock(s, -1)) {
+    SEN_LOG(sen_log_crit, "sen_db_store_add_trigger: lock failed");
+    return sen_abnormal_error;
+  }
   if (!(spec = sen_ja_ref(s->values, e->id, &spec_len))) {
+    sen_db_unlock(s);
     return sen_invalid_argument;
   }
   newspec_len = SEN_DB_STORE_SPEC_SIZE(spec->n_triggers + 1);
-  if (!(newspec = SEN_MALLOC(newspec_len))) { return sen_memory_exhausted; }
+  if (!(newspec = SEN_GMALLOC(newspec_len))) {
+    sen_db_unlock(s);
+    return sen_memory_exhausted;
+  }
   memcpy(newspec, spec, spec_len);
   memcpy(&newspec->triggers[spec->n_triggers], t, sizeof(sen_db_store_rel_spec));
   newspec->n_triggers++;
   sen_ja_unref(s->values, e->id);
-  if ((rc = sen_ja_put(s->values, e->id, newspec, newspec_len, 0))) {
-    goto exit;
-  }
+  rc = sen_ja_put(s->values, e->id, newspec, newspec_len, 0);
+  sen_db_unlock(s);
+  SEN_GFREE(newspec);
+  if (rc) { return rc; }
   {
-    sen_db_store *target = sen_db_store_by_id(s, t->target);
-    if (target) {
-      sen_db_trigger *r = SEN_MALLOC(sizeof(sen_db_trigger));
-      if (!r) { rc = sen_memory_exhausted; goto exit; }
+    sen_db_trigger *r = SEN_GMALLOC(sizeof(sen_db_trigger));
+    if (r) {
+      MUTEX_LOCK(s->lock);
       r->next = e->triggers;
       r->type = t->type;
-      r->target = target;
+      r->target = t->target;
       e->triggers = r;
+      MUTEX_UNLOCK(s->lock);
       if (t->type == sen_db_index_target) {
         sen_db_store_rel_spec invrs;
         invrs.type = sen_db_before_update_trigger;
         invrs.target = e->id;
         rc = sen_db_store_add_trigger(target, &invrs);
       }
+    } else { rc = sen_memory_exhausted; }
+  }
+  return rc;
+}
+
+sen_rc
+sen_db_store_del_trigger(sen_db_store *e, sen_db_store_rel_spec *t)
+{
+  sen_rc rc;
+  sen_db *s = e->db;
+  uint32_t spec_len, newspec_len;
+  uint32_t n_triggers = 0;
+  const sen_db_store_spec *spec;
+  sen_db_store_spec *newspec;
+  sen_db_trigger *r, **rp = &e->triggers;
+  if (sen_db_lock(s, -1)) {
+    SEN_LOG(sen_log_crit, "sen_db_del_trigger: lock failed");
+    return sen_abnormal_error;
+  }
+  if (!(spec = sen_ja_ref(s->values, e->id, &spec_len))) {
+    sen_db_unlock(s);
+    return sen_invalid_argument;
+  }
+  MUTEX_LOCK(s->lock);
+  while ((r = *rp)) {
+    if (r->target == t->target) {
+      *rp = r->next;
+      SEN_GFREE(r);
+    } else {
+      n_triggers++;
+      rp = &r->next;
     }
   }
-exit :
-  SEN_FREE(newspec);
+  MUTEX_UNLOCK(s->lock);
+  newspec_len = SEN_DB_STORE_SPEC_SIZE(n_triggers);
+  if (!(newspec = SEN_GMALLOC(newspec_len))) {
+    sen_db_unlock(s);
+    return sen_memory_exhausted;
+  }
+  memcpy(newspec, spec, newspec_len);
+  newspec->n_triggers = n_triggers;
+  sen_ja_unref(s->values, e->id);
+  for (t = newspec->triggers, r = e->triggers; r; t++, r = r->next) {
+    t->type = r->type;
+    t->target = r->target;
+  }
+  rc = sen_ja_put(s->values, e->id, newspec, newspec_len, 0);
+  sen_db_unlock(s);
+  SEN_GFREE(newspec);
   return rc;
 }
 
 sen_db_store *
 sen_db_slot_class_by_id(sen_db *s, sen_id slot)
 {
-  return sen_db_slot_class(s, _sen_sym_key(s->keys, slot));
+  const char *slotname = _sen_sym_key(s->keys, slot);
+  if (!slotname) { return NULL; }
+  return sen_db_slot_class(s, slotname);
 }
 
 sen_rc
@@ -724,11 +915,12 @@ sen_db_class_add_slot(sen_db *s, sen_id class, const char *name, sen_db_store_sp
 }
 
 sen_rc
-sen_db_store_close(sen_db_store *slot, int all)
+sen_db_store_close(sen_db_store *slot)
 {
-  sen_db *s = slot->db;
   sen_db_trigger *t, *t_;
-  switch (slot->type) {
+  uint8_t type = slot->type;
+  slot->type = 0;
+  switch (type) {
   case sen_db_obj_slot :
     // sen_db_class_close(slot->u.o.class);
     sen_ra_close(slot->u.o.ra);
@@ -753,35 +945,119 @@ sen_db_store_close(sen_db_store *slot, int all)
   }
   for (t = slot->triggers; t; t = t_) {
     t_ = t->next;
-    SEN_FREE(t);
-  }
-  if (!all) {
-    sen_set_eh *ep;
-    if ((ep = sen_set_at(s->stores, &slot->id, NULL))) {
-      sen_set_del(s->stores, ep);
-    }
+    SEN_GFREE(t);
   }
   return sen_success;
 }
 
+sen_rc
+sen_db_store_remove(sen_db *s, const char *name)
+{
+  sen_rc rc;
+  sen_id id;
+  sen_db_store *e;
+  uint32_t spec_len;
+  char buffer[PATH_MAX];
+  const sen_db_store_spec *spec;
+  if (!s || !(id = sen_sym_at(s->keys, name))) { return sen_invalid_argument; }
+  if ((e = sen_array_at(&s->stores, id)) && (e->type)) { sen_db_store_close(e); }
+  if (!(spec = sen_ja_ref(s->values, id, &spec_len))) {
+    return sen_invalid_format;
+  }
+  {
+    int i;
+    sen_db_store *target;
+    const sen_db_store_rel_spec *t;
+    sen_db_store_rel_spec invrs;
+    invrs.target = id;
+    for (t = spec->triggers, i = spec->n_triggers; i; t++, i--) {
+      if (t->target && (target = sen_db_store_by_id(s, t->target))) {
+        if ((rc = sen_db_store_del_trigger(target, &invrs))) {
+          SEN_LOG(sen_log_notice, "sen_db_store_del_trigger failed(%d)", t->target);
+        }
+      }
+    }
+  }
+  if (sen_db_lock(s, -1)) {
+    SEN_LOG(sen_log_crit, "sen_db_store_remove: lock failed");
+    return sen_abnormal_error;
+  }
+  MUTEX_LOCK(s->lock);
+  gen_pathname(s->keys->io->path, buffer, id);
+  switch (spec->type) {
+  case sen_db_raw_class :
+    /* todo */
+    break;
+  case sen_db_class :
+    rc = sen_sym_remove(buffer);
+    break;
+  case sen_db_obj_slot :
+    rc = sen_ra_remove(buffer);
+    break;
+  case sen_db_ra_slot :
+    rc = sen_ra_remove(buffer);
+    break;
+  case sen_db_ja_slot :
+    rc = sen_ja_remove(buffer);
+    break;
+  case sen_db_idx_slot :
+    rc = sen_inv_remove(buffer);
+    break;
+  case sen_db_rel1 :
+    rc = sen_ra_remove(buffer);
+    break;
+  default :
+    break;
+  }
+  sen_ja_unref(s->values, id);
+  {
+    sen_ja_einfo einfo;
+    memset(&einfo, 0, sizeof(sen_ja_einfo));
+    rc = sen_ja_replace(s->values, id, &einfo);
+  }
+  rc = sen_sym_del(s->keys, name);
+  MUTEX_UNLOCK(s->lock);
+  sen_db_unlock(s);
+  return rc;
+}
+
+sen_rc
+sen_db_class_del_slot(sen_db *s, sen_id class, const char *name)
+{
+  char buf[SEN_SYM_MAX_KEY_SIZE];
+  if (sen_db_class_slotpath(s, class, name, buf)) { return sen_invalid_argument; }
+  return sen_db_store_remove(s, buf);
+}
+
 sen_rc
 sen_db_prepare_builtin_class(sen_db *db)
 {
   sen_db_store_spec spec;
   spec.type = sen_db_raw_class;
-  spec.u.c.size = sizeof(int);
+  /* 1 */
+  spec.u.c.size = sizeof(int32_t);
   if (!sen_db_store_create(db, "<int>", &spec)) { return sen_memory_exhausted; }
+  /* 2 */
+  spec.u.c.size = sizeof(uint32_t);
   if (!sen_db_store_create(db, "<uint>", &spec)) { return sen_memory_exhausted; }
+  /* 3 */
   spec.u.c.size = sizeof(int64_t);
   if (!sen_db_store_create(db, "<int64>", &spec)) { return sen_memory_exhausted; }
+  /* 4 */
   spec.u.c.size = sizeof(double);
   if (!sen_db_store_create(db, "<float>", &spec)) { return sen_memory_exhausted; }
+  /* 5 */
   spec.u.c.size = SEN_SYM_MAX_KEY_SIZE;
   if (!sen_db_store_create(db, "<shorttext>", &spec)) { return sen_memory_exhausted; }
+  /* 6 */
   spec.u.c.size = 1 << 16;
   if (!sen_db_store_create(db, "<text>", &spec)) { return sen_memory_exhausted; }
+  /* 7 */
   spec.u.c.size = JA_SEGMENT_SIZE;
   if (!sen_db_store_create(db, "<longtext>", &spec)) { return sen_memory_exhausted; }
+  /* 8 */
+  spec.u.c.size = sizeof(sen_timeval);
+  if (!sen_db_store_create(db, "<time>", &spec)) { return sen_memory_exhausted; }
   return sen_success;
 }
 
@@ -789,22 +1065,37 @@ sen_db *
 sen_db_create(const char *path, int flags, sen_encoding encoding)
 {
   sen_db *s;
-  char buffer[PATH_MAX];
-  if (strlen(path) > PATH_MAX - 14) { return NULL; }
-  if (!(s = SEN_MALLOC(sizeof(sen_db)))) { return NULL; }
-  if ((s->stores = sen_set_open(sizeof(sen_id), sizeof(sen_db_store), 0))) {
-    if ((s->keys = sen_sym_create(path, 0, flags, encoding))) {
-      gen_pathname(path, buffer, 0);
-      if ((s->values = sen_ja_create(buffer, JA_SEGMENT_SIZE))) {
-        sen_db_prepare_builtin_class(s);
-        SEN_LOG(sen_log_notice, "db created (%s) flags=%x", path, s->keys->flags);
-        return s;
+  ERRCLR(NULL);
+  if (strlen(path) <= PATH_MAX - 14) {
+    if ((s = SEN_GMALLOC(sizeof(sen_db)))) {
+      sen_array_init(&s->stores, sizeof(sen_db_store), sen_array_clear);
+      if ((s->keys = sen_sym_create(path, 0, flags, encoding))) {
+        char buffer[PATH_MAX];
+        gen_pathname(path, buffer, 0);
+        if ((s->values = sen_ja_create(buffer, JA_SEGMENT_SIZE))) {
+          MUTEX_INIT(s->lock);
+          sen_db_prepare_builtin_class(s);
+          if (!ERRP(&sen_gctx, SEN_ERROR)) {
+            SEN_LOG(sen_log_notice, "db created (%s) flags=%x", path, s->keys->flags);
+            sen_gctx.db = s;
+            sen_gctx.encoding = encoding;
+            return s;
+          }
+        } else {
+          GERR(sen_memory_exhausted, "ja create failed");
+        }
+        sen_sym_close(s->keys);
+      } else {
+        GERR(sen_memory_exhausted, "s->keys create failed");
       }
-      sen_sym_close(s->keys);
+      sen_array_fin(&s->stores);
+      SEN_GFREE(s);
+    } else {
+      GERR(sen_memory_exhausted, "sen_db alloc failed");
     }
-    sen_set_close(s->stores);
+  } else {
+    GERR(sen_invalid_argument, "too long path");
   }
-  SEN_FREE(s);
   return NULL;
 }
 
@@ -812,39 +1103,52 @@ sen_db *
 sen_db_open(const char *path)
 {
   sen_db *s;
-  char buffer[PATH_MAX];
-  if (strlen(path) > PATH_MAX - 14) { return NULL; }
-  if (!(s = SEN_MALLOC(sizeof(sen_db)))) { return NULL; }
-  if ((s->stores = sen_set_open(sizeof(sen_id), sizeof(sen_db_store), 0))) {
-    if ((s->keys = sen_sym_open(path))) {
-      gen_pathname(path, buffer, 0);
-      if ((s->values = sen_ja_open(buffer))) {
-        SEN_LOG(sen_log_notice, "db opened (%s) flags=%x", path, s->keys->flags);
-        return s;
+  ERRCLR(NULL);
+  if (strlen(path) <= PATH_MAX - 14) {
+    if ((s = SEN_GMALLOC(sizeof(sen_db)))) {
+      sen_array_init(&s->stores, sizeof(sen_db_store), sen_array_clear);
+      if ((s->keys = sen_sym_open(path))) {
+        char buffer[PATH_MAX];
+        gen_pathname(path, buffer, 0);
+        if ((s->values = sen_ja_open(buffer))) {
+          SEN_LOG(sen_log_notice, "db opened (%s) flags=%x", path, s->keys->flags);
+          sen_gctx.db = s;
+          sen_gctx.encoding = s->keys->encoding;
+          MUTEX_INIT(s->lock);
+          return s;
+        } else {
+          GERR(sen_memory_exhausted, "ja open failed");
+        }
+        sen_sym_close(s->keys);
+      } else {
+        GERR(sen_memory_exhausted, "s->keys open failed");
       }
-      sen_sym_close(s->keys);
+      sen_array_fin(&s->stores);
+      SEN_GFREE(s);
+    } else {
+      GERR(sen_memory_exhausted, "sen_db alloc failed");
     }
-    sen_set_close(s->stores);
+  } else {
+    GERR(sen_invalid_argument, "too long path");
   }
-  SEN_FREE(s);
   return NULL;
 }
 
 sen_rc
 sen_db_close(sen_db *s)
 {
-  sen_db_store *e;
-  sen_set_cursor *c;
+  if (!s) { return sen_invalid_argument; }
   sen_sym_close(s->keys);
   sen_ja_close(s->values);
-  if ((c = sen_set_cursor_open(s->stores))) {
-    while (sen_set_cursor_next(c, NULL, (void **) &e)) {
-      sen_db_store_close(e, 1);
-    }
-    sen_set_cursor_close(c);
+  {
+    sen_id id;
+    sen_db_store *sp;
+    SEN_ARRAY_EACH(&s->stores, id, sp, {
+      if (sp->type) { sen_db_store_close(sp); }
+    });
   }
-  sen_set_close(s->stores);
-  SEN_FREE(s);
+  sen_array_fin(&s->stores);
+  SEN_GFREE(s);
   return sen_success;
 }
 
@@ -859,10 +1163,10 @@ sen_vgram *
 sen_vgram_create(const char *path)
 {
   sen_vgram *s;
-  if (!(s = SEN_MALLOC(sizeof(sen_vgram)))) { return NULL; }
+  if (!(s = SEN_GMALLOC(sizeof(sen_vgram)))) { return NULL; }
   s->vgram = sen_sym_create(path, sizeof(sen_id) * 2, 0, sen_enc_none);
   if (!s->vgram) {
-    SEN_FREE(s);
+    SEN_GFREE(s);
     return NULL;
   }
   return s;
@@ -872,10 +1176,10 @@ sen_vgram *
 sen_vgram_open(const char *path)
 {
   sen_vgram *s;
-  if (!(s = SEN_MALLOC(sizeof(sen_vgram)))) { return NULL; }
+  if (!(s = SEN_GMALLOC(sizeof(sen_vgram)))) { return NULL; }
   s->vgram = sen_sym_open(path);
   if (!s->vgram) {
-    SEN_FREE(s);
+    SEN_GFREE(s);
     return NULL;
   }
   return s;
@@ -885,13 +1189,13 @@ sen_vgram_buf *
 sen_vgram_buf_open(size_t len)
 {
   sen_vgram_buf *b;
-  if (!(b = SEN_MALLOC(sizeof(sen_vgram_buf)))) { return NULL; }
+  if (!(b = SEN_GMALLOC(sizeof(sen_vgram_buf)))) { return NULL; }
   b->len = len;
-  b->tvs = b->tvp = SEN_MALLOC(sizeof(sen_id) * len);
-  if (!b->tvp) { SEN_FREE(b); return NULL; }
+  b->tvs = b->tvp = SEN_GMALLOC(sizeof(sen_id) * len);
+  if (!b->tvp) { SEN_GFREE(b); return NULL; }
   b->tve = b->tvs + len;
-  b->vps = b->vpp = SEN_MALLOC(sizeof(sen_vgram_vnode) * len * 2);
-  if (!b->vpp) { SEN_FREE(b->tvp); SEN_FREE(b); return NULL; }
+  b->vps = b->vpp = SEN_GMALLOC(sizeof(sen_vgram_vnode) * len * 2);
+  if (!b->vpp) { SEN_GFREE(b->tvp); SEN_GFREE(b); return NULL; }
   b->vpe = b->vps + len;
   return b;
 }
@@ -920,30 +1224,31 @@ sen_vgram_update(sen_vgram *vgram, sen_id rid, sen_vgram_buf *b, sen_set *terms)
     sen_id *t0, *tn;
     for (t0 = b->tvs; t0 < b->tvp - 1; t0++) {
       sen_vgram_vnode *v, **vp;
-      sen_set_at(terms, t0, (void **) &u);
-      vp = &(*u)->vnodes;
-      for (tn = t0 + 1; tn < b->tvp; tn++) {
-        for (v = *vp; v && v->tid != *tn; v = v->cdr) ;
-        if (!v) {
-          if (b->vpp < b->vpe) {
-            v = b->vpp++;
+      if (sen_set_at(terms, t0, (void **) &u)) {
+        vp = &(*u)->vnodes;
+        for (tn = t0 + 1; tn < b->tvp; tn++) {
+          for (v = *vp; v && v->tid != *tn; v = v->cdr) ;
+          if (!v) {
+            if (b->vpp < b->vpe) {
+              v = b->vpp++;
+            } else {
+              // todo;
+              break;
+            }
+            v->car = NULL;
+            v->cdr = *vp;
+            *vp = v;
+            v->tid = *tn;
+            v->vid = 0;
+            v->freq = 0;
+            v->len = tn - t0;
+          }
+          v->freq++;
+          if (v->vid) {
+            vp = &v->car;
           } else {
-            // todo;
             break;
           }
-          v->car = NULL;
-          v->cdr = *vp;
-          *vp = v;
-          v->tid = *tn;
-          v->vid = 0;
-          v->freq = 0;
-          v->len = tn - t0;
-        }
-        v->freq++;
-        if (v->vid) {
-          vp = &v->car;
-        } else {
-          break;
         }
       }
     }
@@ -954,36 +1259,40 @@ sen_vgram_update(sen_vgram *vgram, sen_id rid, sen_vgram_buf *b, sen_set *terms)
       for (t0 = b->tvs; t0 < b->tvp; t0++) {
         sen_id vid, vid0 = *t0, vid1 = 0;
         sen_vgram_vnode *v, *v2 = NULL, **vp;
-        sen_set_at(terms, t0, (void **) &u);
-        vp = &(*u)->vnodes;
-        for (tn = t0 + 1; tn < b->tvp; tn++) {
-          for (v = *vp; v; v = v->cdr) {
-            if (!v->vid && (v->freq < 2 || v->freq * v->len < 4)) {
-              *vp = v->cdr;
-              v->freq = 0;
+        if (sen_set_at(terms, t0, (void **) &u)) {
+          vp = &(*u)->vnodes;
+          for (tn = t0 + 1; tn < b->tvp; tn++) {
+            for (v = *vp; v; v = v->cdr) {
+              if (!v->vid && (v->freq < 2 || v->freq * v->len < 4)) {
+                *vp = v->cdr;
+                v->freq = 0;
+              }
+              if (v->tid == *tn) { break; }
+              vp = &v->cdr;
             }
-            if (v->tid == *tn) { break; }
-            vp = &v->cdr;
+            if (v) {
+              if (v->freq) {
+                v2 = v;
+                vid1 = vid0;
+                vid0 = v->vid;
+              }
+              if (v->vid) {
+                vp = &v->car;
+                continue;
+              }
+            }
+            break;
           }
-          if (v) {
-            if (v->freq) {
-              v2 = v;
-              vid1 = vid0;
-              vid0 = v->vid;
-            }
-            if (v->vid) {
-              vp = &v->car;
-              continue;
-            }
-          }
-          break;
         }
         if (v2) {
           if (!v2->vid) {
             vgram_key key;
             key.vid = vid1;
             key.tid = v2->tid;
-            v2->vid = sen_sym_get(vgram->vgram, (char *)&key);
+            if (!(v2->vid = sen_sym_get(vgram->vgram, (char *)&key))) {
+              sen_set_close(th);
+              return sen_memory_exhausted;
+            }
           }
           vid = *t0 = v2->vid * 2 + 1;
           memset(t0 + 1, 0, sizeof(sen_id) * v2->len);
@@ -993,7 +1302,10 @@ sen_vgram_update(sen_vgram *vgram, sen_id rid, sen_vgram_buf *b, sen_set *terms)
         }
         {
           int *tf;
-          sen_set_get(th, &vid, (void **) &tf);
+          if (!sen_set_get(th, &vid, (void **) &tf)) {
+            sen_set_close(th);
+            return sen_memory_exhausted;
+          }
           (*tf)++;
         }
       }
@@ -1003,7 +1315,7 @@ sen_vgram_update(sen_vgram *vgram, sen_id rid, sen_vgram_buf *b, sen_set *terms)
         int skip = 0;
         sen_set_eh *ehs, *ehp, *ehe;
         sen_set_sort_optarg arg;
-        uint8_t *ps = SEN_MALLOC(b->len * 2), *pp, *pe;
+        uint8_t *ps = SEN_GMALLOC(b->len * 2), *pp, *pe;
         if (!ps) {
           sen_set_close(th);
           return sen_memory_exhausted;
@@ -1016,7 +1328,7 @@ sen_vgram_update(sen_vgram *vgram, sen_id rid, sen_vgram_buf *b, sen_set *terms)
         arg.compar_arg0 = NULL;
         ehs = sen_set_sort(th, 0, &arg);
         if (!ehs) {
-          SEN_FREE(ps);
+          SEN_GFREE(ps);
           sen_set_close(th);
           return sen_memory_exhausted;
         }
@@ -1040,8 +1352,8 @@ sen_vgram_update(sen_vgram *vgram, sen_id rid, sen_vgram_buf *b, sen_set *terms)
         len_sum += b->len;
         img_sum += pp - ps;
         skip_sum += skip;
-        SEN_FREE(ehs);
-        SEN_FREE(ps);
+        SEN_GFREE(ehs);
+        SEN_GFREE(ps);
       }
       sen_set_close(th);
     }
@@ -1053,9 +1365,9 @@ sen_rc
 sen_vgram_buf_close(sen_vgram_buf *b)
 {
   if (!b) { return sen_invalid_argument; }
-  if (b->tvs) { SEN_FREE(b->tvs); }
-  if (b->vps) { SEN_FREE(b->vps); }
-  SEN_FREE(b);
+  if (b->tvs) { SEN_GFREE(b->tvs); }
+  if (b->vps) { SEN_GFREE(b->vps); }
+  SEN_GFREE(b);
   return sen_success;
 }
 
@@ -1065,6 +1377,6 @@ sen_vgram_close(sen_vgram *vgram)
   if (!vgram) { return sen_invalid_argument; }
   SEN_LOG(sen_log_debug, "len=%d img=%d skip=%d simple=%d", len_sum, img_sum, skip_sum, simple_sum);
   sen_sym_close(vgram->vgram);
-  SEN_FREE(vgram);
+  SEN_GFREE(vgram);
   return sen_success;
 }
