@@ -14,24 +14,20 @@
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
-#ifndef _SEN_STORE_H
-#define _SEN_STORE_H
+#ifndef SEN_STORE_H
+#define SEN_STORE_H
 
-#ifndef _SENNA_H
+#ifndef SENNA_H
 #include "senna_in.h"
-#endif /* _SENNA_H */
+#endif /* SENNA_H */
 
-#ifndef _SEN_SET_H
+#ifndef SEN_SET_H
 #include "set.h"
-#endif /* _SEN_SET_H */
+#endif /* SEN_SET_H */
 
-#ifndef _SEN_IO_H
+#ifndef SEN_IO_H
 #include "io.h"
-#endif /* _SEN_IO_H */
-
-#ifndef _SEN_STR_H
-#include "str.h"
-#endif /* _SEN_STR_H */
+#endif /* SEN_IO_H */
 
 #ifdef __cplusplus
 extern "C" {
@@ -110,6 +106,8 @@ sen_rc sen_ja_replace(sen_ja *ja, sen_id id, sen_ja_einfo *ei);
 #define SEN_DB_PSLOT_NSUBRECS   3
 #define SEN_DB_PSLOT_MASK 0xff
 
+typedef struct _sen_array sen_array;
+
 typedef struct _sen_db_store_rel_spec sen_db_store_rel_spec;
 typedef struct _sen_db_store_spec sen_db_store_spec;
 typedef struct _sen_db_store sen_db_store;
@@ -144,16 +142,29 @@ struct _sen_db_store_spec {
   sen_db_store_rel_spec triggers[1];
 };
 
+enum {
+  sen_array_clear = 1
+};
+
+struct _sen_array {
+  uint16_t element_size;
+  uint16_t flags;
+  sen_mutex lock;
+  void *elements[8];
+};
+
 struct _sen_db {
   sen_sym *keys;
   sen_ja *values;
-  sen_set *stores;
+  sen_array stores;
+  //  sen_set *stores;
+  sen_mutex lock; /* todo */
 };
 
 struct _sen_db_trigger {
   sen_db_trigger *next;
   sen_db_rel_type type;
-  sen_db_store *target;
+  sen_id target;
 };
 
 struct _sen_db_store {
@@ -201,8 +212,14 @@ sen_db_store *sen_db_slot_class_by_id(sen_db *s, sen_id slot);
 sen_db_store *sen_db_class_slot(sen_db *s, sen_id class, const char *name);
 sen_db_store *sen_db_class_add_slot(sen_db *s, sen_id class, const char *name,
                                     sen_db_store_spec *spec);
+sen_rc sen_db_class_del_slot(sen_db *s, sen_id class, const char *name);
 sen_rc sen_db_class_slotpath(sen_db *s, sen_id class, const char *name, char *buf);
 
+sen_rc sen_db_idx_slot_build(sen_db *s, sen_db_store *store);
+sen_rc sen_db_lock(sen_db *s, int timeout);
+sen_rc sen_db_unlock(sen_db *s);
+sen_rc sen_db_clear_lock(sen_db *s);
+
 /**** vgram ****/
 
 typedef struct _sen_vgram_vnode
@@ -242,4 +259,4 @@ sen_rc sen_vgram_buf_close(sen_vgram_buf *b);
 }
 #endif
 
-#endif /* _SEN_STORE_H */
+#endif /* SEN_STORE_H */
