@@ -17,9 +17,13 @@
 #include "senna_in.h"
 #include <stdio.h>
 #include <string.h>
+#include "ctx.h"
 #include "str.h"
 #include "set.h"
 
+#define __USE_ISOC99
+#include <math.h>
+
 static sen_set *prefix = NULL;
 static sen_set *suffix = NULL;
 
@@ -214,202 +218,15 @@ int
 sen_str_get_prefix_order(const char *str)
 {
   int *ip;
+  if (!str) { return -1; }
   if (!prefix) { prefix_init(); }
-  if (sen_set_at(prefix, str, (void **)&ip)) {
+  if (prefix && sen_set_at(prefix, str, (void **)&ip)) {
     return *ip;
   } else {
     return -1;
   }
 }
 
-#include <stdarg.h>
-#include <time.h>
-
-/*
-static void
-default_logger_func(int level, const char *time,
-                    const char *title, const char *msg, const char *location)
-{
-  fprintf(stderr, "%s|%s %s %s\n", time, location, msg, note);
-}
-*/
-
-static FILE *default_logger_fp = NULL;
-
-static void
-default_logger_func(int level, const char *time, const char *title,
-                    const char *msg, const char *location, void *func_arg)
-{
-  const char slev[] = " EACewnid-";
-  if (!default_logger_fp) {
-    // mutex_lock
-    default_logger_fp = fopen(SENNA_LOG_PATH, "a");
-    // mutex_unlock
-  }
-  if (default_logger_fp) {
-    fprintf(default_logger_fp, "%s|%c|%s %s %s\n",
-            time, *(slev + level), title, msg, location);
-    fflush(default_logger_fp);
-  }
-}
-
-void
-sen_logger_fin(void)
-{
-  if (default_logger_fp) {
-    fclose(default_logger_fp);
-    default_logger_fp = NULL;
-  }
-}
-
-static sen_logger_info default_logger = {
-  SEN_LOG_DEFAULT_LEVEL,
-  SEN_LOG_TIME|SEN_LOG_MESSAGE,
-  default_logger_func
-};
-
-static const sen_logger_info *sen_logger = &default_logger;
-
-sen_rc
-sen_logger_info_set(const sen_logger_info *info)
-{
-  if (info) {
-    sen_logger = info;
-  } else {
-    sen_logger = &default_logger;
-  }
-  return sen_success;
-}
-
-int
-sen_logger_pass(sen_log_level level)
-{
-  return level <= sen_logger->max_level;
-}
-
-#define TBUFSIZE 0x20
-#define MBUFSIZE 0x1000
-#define LBUFSIZE 0x400
-
-void
-sen_logger_put(sen_log_level level,
-               const char *file, int line, const char *func, char *fmt, ...)
-{
-  if (level <= sen_logger->max_level) {
-    char tbuf[TBUFSIZE];
-    char mbuf[MBUFSIZE];
-    char lbuf[LBUFSIZE];
-    if (sen_logger->flags & SEN_LOG_TIME) {
-#ifdef WIN32
-      struct tm *ltm;
-      time_t t;
-      struct _timeb tb;
-      time(&t);
-      ltm = localtime(&t);
-      _ftime(&tb);
-      snprintf(tbuf, TBUFSIZE - 1, "%02d/%02d:%02d:%02d:%02d.%03d",
-               ltm->tm_mon + 1, ltm->tm_mday,
-               ltm->tm_hour, ltm->tm_min, ltm->tm_sec, tb.millitm);
-      tbuf[TBUFSIZE - 1] = '\0';
-#else /* WIN32 */
-      struct timeval tv;
-      struct tm tm;
-      gettimeofday(&tv, NULL);
-      localtime_r((time_t *)&tv.tv_sec, &tm);
-      snprintf(tbuf, TBUFSIZE - 1, "%02d/%02d:%02d:%02d:%02d.%06d",
-               tm.tm_mon + 1, tm.tm_mday,
-               tm.tm_hour, tm.tm_min, tm.tm_sec, (int) tv.tv_usec);
-      tbuf[TBUFSIZE - 1] = '\0';
-#endif /* WIN32 */
-    } else {
-      tbuf[0] = '\0';
-    }
-    if (sen_logger->flags & SEN_LOG_MESSAGE) {
-      va_list argp;
-      va_start(argp, fmt);
-      vsnprintf(mbuf, MBUFSIZE - 1, fmt, argp);
-      va_end(argp);
-      mbuf[MBUFSIZE - 1] = '\0';
-    } else {
-      mbuf[0] = '\0';
-    }
-    if (sen_logger->flags & SEN_LOG_LOCATION) {
-      snprintf(lbuf, LBUFSIZE - 1, "%04x %s:%d %s()",
-               getpid(), file, line, func);
-      lbuf[LBUFSIZE - 1] = '\0';
-    } else {
-      lbuf[0] = '\0';
-    }
-    if (sen_logger->func) {
-      sen_logger->func(level, tbuf, "", mbuf, lbuf, sen_logger->func_arg);
-    } else {
-      default_logger_func(level, tbuf, "", mbuf, lbuf, sen_logger->func_arg);
-    }
-  }
-}
-
-// static int alloc_size = 0;
-static int alloc_count = 0;
-
-void *
-sen_malloc(size_t size, const char* file, int line)
-{
-  void *res;
-  res = malloc(size);
-  if (res) { alloc_count++; }
-  if (!res) { SEN_LOG(sen_log_alert, "malloc fail (%d)=%p (%s:%d) <%d>", size, res, file, line, alloc_count); }
-  return res;
-}
-
-void *
-sen_calloc(size_t size, const char* file, int line)
-{
-  void *res;
-  res = calloc(size, 1);
-  if (res) { alloc_count++; }
-  if (!res) { SEN_LOG(sen_log_alert, "calloc fail (%d)=%p (%s:%d) <%d>", size, res, file, line, alloc_count); }
-  return res;
-}
-
-void
-sen_free(void *ptr, const char* file, int line)
-{
-  free(ptr);
-  if (ptr) { alloc_count--; }
-  if (!ptr) { SEN_LOG(sen_log_alert, "free fail (%p) (%s:%d) <%d>", ptr, file, line, alloc_count); }
-}
-
-void *
-sen_realloc(void *ptr, size_t size, const char* file, int line)
-{
-  void *res;
-  res = realloc(ptr, size);
-  if (res) { alloc_count++; }
-  if (size && !res) {
-    SEN_LOG(sen_log_alert, "realloc fail (%p,%d)=%p (%s:%d) <%d>",
-            ptr, size, res, file, line, alloc_count);
-  }
-  return res;
-}
-
-char *
-sen_strdup(const char *s, const char* file, int line)
-{
-  char *res;
-  res = strdup(s);
-  if (res) { alloc_count++; }
-  if (!res) { SEN_LOG(sen_log_alert, "strdup(%p)=%p (%s:%d) <%d>", s, res, file, line, alloc_count); }
-  return res;
-}
-
-void
-sen_assert(int cond, const char* file, int line, const char* func)
-{
-  if (!cond) {
-    SEN_LOG(sen_log_error, "ASSERT fail on %s %s:%d", func, file, line);
-  }
-}
-
 static unsigned char symbol[] = {
   ',', '.', 0, ':', ';', '?', '!', 0, 0, 0, '`', 0, '^', '~', '_', 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, '-', '-', '/', '\\', 0, 0, '|', 0, 0, 0, '\'', 0,
@@ -440,7 +257,8 @@ normalize_euc(sen_nstr *nstr)
   static unsigned char handaku[] = {
     0xd1, 0, 0, 0xd4, 0, 0, 0xd7, 0, 0, 0xda, 0, 0, 0xdd
   };
-  int_least8_t *ch;
+  int16_t *ch;
+  sen_ctx *ctx = nstr->ctx;
   const unsigned char *s, *s_, *e;
   unsigned char *d, *d0, *d_, b;
   uint_least8_t *cp, *ctypes, ctype;
@@ -451,8 +269,9 @@ normalize_euc(sen_nstr *nstr)
   }
   d0 = (unsigned char *) nstr->norm;
   if (nstr->flags & SEN_STR_WITH_CHECKS) {
-    if (!(nstr->checks = SEN_MALLOC(size * 2 + 1))) {
+    if (!(nstr->checks = SEN_MALLOC(size * 2 * sizeof(int16_t) + 1))) {
       SEN_FREE(nstr->norm);
+      nstr->norm = NULL;
       return sen_memory_exhausted;
     }
   }
@@ -461,6 +280,8 @@ normalize_euc(sen_nstr *nstr)
     if (!(nstr->ctypes = SEN_MALLOC(size + 1))) {
       SEN_FREE(nstr->checks);
       SEN_FREE(nstr->norm);
+      nstr->checks = NULL;
+      nstr->norm = NULL;
       return sen_memory_exhausted;
     }
   }
@@ -642,7 +463,7 @@ normalize_euc(sen_nstr *nstr)
     length++;
     if (cp) { *cp++ = ctype; }
     if (ch) {
-      *ch++ = (int_least8_t)(s + 1 - s_);
+      *ch++ = (int16_t)(s + 1 - s_);
       s_ = s + 1;
       while (++d_ < d) { *ch++ = 0; }
     }
@@ -662,19 +483,20 @@ const char *sen_nfkc_map2(const unsigned char *prefix, const unsigned char *suff
 inline static sen_rc
 normalize_utf8(sen_nstr *nstr)
 {
-  int_least8_t *ch;
-  const unsigned char *s, *s_, *p, *p2, *pe, *e;
-  unsigned char *d, *d0, *d_;
+  int16_t *ch;
+  sen_ctx *ctx = nstr->ctx;
+  const unsigned char *s, *s_, *s__, *p, *p2, *pe, *e;
+  unsigned char *d, *d_;
   uint_least8_t *cp, *ctypes;
-  size_t length = 0, ls, ls_ = 0, lp, size = nstr->orig_blen;
+  size_t length = 0, ls, lp, size = nstr->orig_blen;
   int removeblankp = nstr->flags & SEN_STR_REMOVEBLANK;
   if (!(nstr->norm = SEN_MALLOC(size * 5 + 1))) { /* todo: realloc unless enough */
     return sen_memory_exhausted;
   }
-  d0 = (unsigned char *) nstr->norm;
   if (nstr->flags & SEN_STR_WITH_CHECKS) {
-    if (!(nstr->checks = SEN_MALLOC(size * 5 + 1))) { /* todo: realloc unless enough */
+    if (!(nstr->checks = SEN_MALLOC(size * 5 * sizeof(int16_t) + 1))) { /* todo: realloc unless enough */
       SEN_FREE(nstr->norm);
+      nstr->norm = NULL;
       return sen_memory_exhausted;
     }
   }
@@ -683,12 +505,15 @@ normalize_utf8(sen_nstr *nstr)
     if (!(nstr->ctypes = SEN_MALLOC(size * 3 + 1))) { /* todo: realloc unless enough */
       SEN_FREE(nstr->checks);
       SEN_FREE(nstr->norm);
+      nstr->checks = NULL;
+      nstr->norm = NULL;
       return sen_memory_exhausted;
     }
   }
   cp = ctypes = nstr->ctypes;
   e = (unsigned char *)nstr->orig + size;
-  for (s = s_ = (unsigned char *) nstr->orig, d = d0, d_ = NULL; ; s += ls) {
+  for (s = s_ = (unsigned char *)nstr->orig,
+       d = (unsigned char *)nstr->norm, d_ = NULL; ; s += ls) {
     if (!(ls = sen_str_charlen_utf8(s, e))) {
       break;
     }
@@ -704,7 +529,7 @@ normalize_utf8(sen_nstr *nstr)
       if (cp) { cp--; }
       if (ch) {
         ch -= (d - d_);
-        s_ -= ls_;
+        s_ = s__;
       }
       d = d_;
       length--;
@@ -727,14 +552,14 @@ normalize_utf8(sen_nstr *nstr)
           if (s_ == s + ls) {
             *ch++ = -1;
           } else {
-            *ch++ = (int_least8_t)(s + ls - s_);
+            *ch++ = (int16_t)(s + ls - s_);
+            s__ = s_;
             s_ = s + ls;
           }
           for (i = lp; i > 1; i--) { *ch++ = 0; }
         }
       }
     }
-    ls_ = ls;
   }
   if (cp) { *cp = sen_str_null; }
   *d = '\0';
@@ -766,7 +591,8 @@ normalize_sjis(sen_nstr *nstr)
   static unsigned char handaku[] = {
     0x70, 0, 0, 0x73, 0, 0, 0x76, 0, 0, 0x79, 0, 0, 0x7c
   };
-  int_least8_t *ch;
+  int16_t *ch;
+  sen_ctx *ctx = nstr->ctx;
   const unsigned char *s, *s_;
   unsigned char *d, *d0, *d_, b, *e;
   uint_least8_t *cp, *ctypes, ctype;
@@ -777,8 +603,9 @@ normalize_sjis(sen_nstr *nstr)
   }
   d0 = (unsigned char *) nstr->norm;
   if (nstr->flags & SEN_STR_WITH_CHECKS) {
-    if (!(nstr->checks = SEN_MALLOC(size * 2 + 1))) {
+    if (!(nstr->checks = SEN_MALLOC(size * 2 * sizeof(int16_t) + 1))) {
       SEN_FREE(nstr->norm);
+      nstr->norm = NULL;
       return sen_memory_exhausted;
     }
   }
@@ -787,6 +614,8 @@ normalize_sjis(sen_nstr *nstr)
     if (!(nstr->ctypes = SEN_MALLOC(size + 1))) {
       SEN_FREE(nstr->checks);
       SEN_FREE(nstr->norm);
+      nstr->checks = NULL;
+      nstr->norm = NULL;
       return sen_memory_exhausted;
     }
   }
@@ -959,7 +788,7 @@ normalize_sjis(sen_nstr *nstr)
     length++;
     if (cp) { *cp++ = ctype; }
     if (ch) {
-      *ch++ = (int_least8_t)(s + 1 - s_);
+      *ch++ = (int16_t)(s + 1 - s_);
       s_ = s + 1;
       while (++d_ < d) { *ch++ = 0; }
     }
@@ -974,7 +803,8 @@ normalize_sjis(sen_nstr *nstr)
 inline static sen_rc
 normalize_none(sen_nstr *nstr)
 {
-  int_least8_t *ch;
+  int16_t *ch;
+  sen_ctx *ctx = nstr->ctx;
   const unsigned char *s, *s_, *e;
   unsigned char *d, *d0, *d_;
   uint_least8_t *cp, *ctypes, ctype;
@@ -985,8 +815,9 @@ normalize_none(sen_nstr *nstr)
   }
   d0 = (unsigned char *) nstr->norm;
   if (nstr->flags & SEN_STR_WITH_CHECKS) {
-    if (!(nstr->checks = SEN_MALLOC(size + 1))) {
+    if (!(nstr->checks = SEN_MALLOC(size * sizeof(int16_t) + 1))) {
       SEN_FREE(nstr->norm);
+      nstr->norm = NULL;
       return sen_memory_exhausted;
     }
   }
@@ -995,6 +826,8 @@ normalize_none(sen_nstr *nstr)
     if (!(nstr->ctypes = SEN_MALLOC(size + 1))) {
       SEN_FREE(nstr->checks);
       SEN_FREE(nstr->norm);
+      nstr->checks = NULL;
+      nstr->norm = NULL;
       return sen_memory_exhausted;
     }
   }
@@ -1051,7 +884,7 @@ normalize_none(sen_nstr *nstr)
     length++;
     if (cp) { *cp++ = ctype; }
     if (ch) {
-      *ch++ = (int_least8_t)(s + 1 - s_);
+      *ch++ = (int16_t)(s + 1 - s_);
       s_ = s + 1;
       while (++d_ < d) { *ch++ = 0; }
     }
@@ -1067,7 +900,8 @@ normalize_none(sen_nstr *nstr)
 inline static sen_rc
 normalize_latin1(sen_nstr *nstr)
 {
-  int_least8_t *ch;
+  int16_t *ch;
+  sen_ctx *ctx = nstr->ctx;
   const unsigned char *s, *s_, *e;
   unsigned char *d, *d0, *d_;
   uint_least8_t *cp, *ctypes, ctype;
@@ -1078,8 +912,9 @@ normalize_latin1(sen_nstr *nstr)
   }
   d0 = (unsigned char *) nstr->norm;
   if (nstr->flags & SEN_STR_WITH_CHECKS) {
-    if (!(nstr->checks = SEN_MALLOC(size + 1))) {
+    if (!(nstr->checks = SEN_MALLOC(size * sizeof(int16_t) + 1))) {
       SEN_FREE(nstr->norm);
+      nstr->norm = NULL;
       return sen_memory_exhausted;
     }
   }
@@ -1088,6 +923,8 @@ normalize_latin1(sen_nstr *nstr)
     if (!(nstr->ctypes = SEN_MALLOC(size + 1))) {
       SEN_FREE(nstr->checks);
       SEN_FREE(nstr->norm);
+      nstr->checks = NULL;
+      nstr->norm = NULL;
       return sen_memory_exhausted;
     }
   }
@@ -1178,7 +1015,7 @@ normalize_latin1(sen_nstr *nstr)
     length++;
     if (cp) { *cp++ = ctype; }
     if (ch) {
-      *ch++ = (int_least8_t)(s + 1 - s_);
+      *ch++ = (int16_t)(s + 1 - s_);
       s_ = s + 1;
       while (++d_ < d) { *ch++ = 0; }
     }
@@ -1193,7 +1030,8 @@ normalize_latin1(sen_nstr *nstr)
 inline static sen_rc
 normalize_koi8r(sen_nstr *nstr)
 {
-  int_least8_t *ch;
+  int16_t *ch;
+  sen_ctx *ctx = nstr->ctx;
   const unsigned char *s, *s_, *e;
   unsigned char *d, *d0, *d_;
   uint_least8_t *cp, *ctypes, ctype;
@@ -1204,8 +1042,9 @@ normalize_koi8r(sen_nstr *nstr)
   }
   d0 = (unsigned char *) nstr->norm;
   if (nstr->flags & SEN_STR_WITH_CHECKS) {
-    if (!(nstr->checks = SEN_MALLOC(size + 1))) {
+    if (!(nstr->checks = SEN_MALLOC(size * sizeof(int16_t) + 1))) {
       SEN_FREE(nstr->norm);
+      nstr->norm = NULL;
       return sen_memory_exhausted;
     }
   }
@@ -1214,6 +1053,8 @@ normalize_koi8r(sen_nstr *nstr)
     if (!(nstr->ctypes = SEN_MALLOC(size + 1))) {
       SEN_FREE(nstr->checks);
       SEN_FREE(nstr->norm);
+      nstr->checks = NULL;
+      nstr->norm = NULL;
       return sen_memory_exhausted;
     }
   }
@@ -1293,7 +1134,7 @@ normalize_koi8r(sen_nstr *nstr)
     length++;
     if (cp) { *cp++ = ctype; }
     if (ch) {
-      *ch++ = (int_least8_t)(s + 1 - s_);
+      *ch++ = (int16_t)(s + 1 - s_);
       s_ = s + 1;
       while (++d_ < d) { *ch++ = 0; }
     }
@@ -1309,8 +1150,9 @@ sen_nstr *
 sen_nstr_open(const char *str, size_t str_len, sen_encoding encoding, int flags)
 {
   sen_rc rc;
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
   sen_nstr *nstr;
-  if (!str || !*str) { return NULL; }
+  if (!str) { return NULL; }
   if (!(nstr = SEN_MALLOC(sizeof(sen_nstr)))) {
     SEN_LOG(sen_log_alert, "memory allocation on sen_fakenstr_open failed !");
     return NULL;
@@ -1322,6 +1164,7 @@ sen_nstr_open(const char *str, size_t str_len, sen_encoding encoding, int flags)
   nstr->checks = NULL;
   nstr->ctypes = NULL;
   nstr->flags = flags;
+  nstr->ctx = ctx;
   switch (encoding) {
   case sen_enc_euc_jp :
     rc = normalize_euc(nstr);
@@ -1358,6 +1201,7 @@ sen_fakenstr_open(const char *str, size_t str_len, sen_encoding encoding, int fl
 {
   /* TODO: support SEN_STR_REMOVEBLANK flag and ctypes */
   sen_nstr *nstr;
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
 
   if (!(nstr = SEN_MALLOC(sizeof(sen_nstr)))) {
     SEN_LOG(sen_log_alert, "memory allocation on sen_fakenstr_open failed !");
@@ -1375,12 +1219,13 @@ sen_fakenstr_open(const char *str, size_t str_len, sen_encoding encoding, int fl
   nstr->norm_blen = str_len;
   nstr->ctypes = NULL;
   nstr->flags = flags;
+  nstr->ctx = ctx;
 
   if (flags & SEN_STR_WITH_CHECKS) {
-    int_least8_t f = 0;
+    int16_t f = 0;
     unsigned char c;
     size_t i;
-    if (!(nstr->checks = (int_least8_t *) SEN_MALLOC(sizeof(int_least8_t) * str_len))) {
+    if (!(nstr->checks = (int16_t *) SEN_MALLOC(sizeof(int16_t) * str_len))) {
       SEN_FREE(nstr->norm);
       SEN_FREE(nstr);
       return NULL;
@@ -1441,12 +1286,16 @@ sen_fakenstr_open(const char *str, size_t str_len, sen_encoding encoding, int fl
 sen_rc
 sen_nstr_close(sen_nstr *nstr)
 {
-  if (!nstr) { return sen_invalid_argument; }
-  if (nstr->norm) { SEN_FREE(nstr->norm); }
-  if (nstr->ctypes) { SEN_FREE(nstr->ctypes); }
-  if (nstr->checks) { SEN_FREE(nstr->checks); }
-  SEN_FREE(nstr);
-  return sen_success;
+  if (nstr) {
+    sen_ctx *ctx = nstr->ctx;
+    if (nstr->norm) { SEN_FREE(nstr->norm); }
+    if (nstr->ctypes) { SEN_FREE(nstr->ctypes); }
+    if (nstr->checks) { SEN_FREE(nstr->checks); }
+    SEN_FREE(nstr);
+    return sen_success;
+  } else {
+    return sen_invalid_argument;
+  }
 }
 
 static const char *sen_enc_string[] = {
@@ -1538,10 +1387,11 @@ sen_atoi(const char *nptr, const char *end, const char **rest)
     o = 1;
   }
   while (p < end && *p >= '0' && *p <= '9') {
-    t = v * 10 + (*p++ - '0');
-    if (t < v) { return 0; }
+    t = v * 10 + (*p - '0');
+    if (t < v) { v =0; break; }
     v = t;
     o = 0;
+    p++;
   }
   if (rest) { *rest = o ? nptr : p; }
   return n ? -v : v;
@@ -1552,14 +1402,38 @@ sen_atoui(const char *nptr, const char *end, const char **rest)
 {
   unsigned int v = 0, t;
   while (nptr < end && *nptr >= '0' && *nptr <= '9') {
-    t = v * 10 + (*nptr++ - '0');
-    if (t < v) { return 0; }
+    t = v * 10 + (*nptr - '0');
+    if (t < v) { v = 0; break; }
     v = t;
+    nptr++;
   }
   if (rest) { *rest = nptr; }
   return v;
 }
 
+int64_t
+sen_atoll(const char *nptr, const char *end, const char **rest)
+{
+  /* FIXME: INT_MIN is not supported */
+  const char *p = nptr;
+  int n = 0, o = 0;
+  int64_t v = 0, t;
+  if (p < end && *p == '-') {
+    p++;
+    n = 1;
+    o = 1;
+  }
+  while (p < end && *p >= '0' && *p <= '9') {
+    t = v * 10 + (*p - '0');
+    if (t < v) { v = 0; break; }
+    v = t;
+    o = 0;
+    p++;
+  }
+  if (rest) { *rest = o ? nptr : p; }
+  return n ? -v : v;
+}
+
 unsigned int
 sen_htoui(const char *nptr, const char *end, const char **rest)
 {
@@ -1641,6 +1515,30 @@ sen_str_itoa(int i, char *p, char *end, char **rest)
   return sen_success;
 }
 
+sen_rc
+sen_str_lltoa(int64_t i, char *p, char *end, char **rest)
+{
+  /* FIXME: INT_MIN is not supported */
+  char *q;
+  if (p >= end) { return sen_invalid_argument; }
+  if (i < 0) {
+    *p++ = '-';
+    i = -i;
+  }
+  q = p;
+  do {
+    if (p >= end) { return sen_invalid_argument; }
+    *p++ = i % 10 + '0';
+  } while ((i /= 10) > 0);
+  if (rest) { *rest = p; }
+  for (p--; q < p; q++, p--) {
+    char t = *q;
+    *q = *p;
+    *p = t;
+  }
+  return sen_success;
+}
+
 #define I2B(i) \
  ("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/"[(i) & 0x3f])
 
@@ -1772,6 +1670,8 @@ exit:
 #define UNIT_SIZE (1 << 12)
 #define UNIT_MASK (UNIT_SIZE - 1)
 
+int sen_rbuf_margin_size = 0;
+
 sen_rc
 sen_rbuf_init(sen_rbuf *buf, size_t size)
 {
@@ -1785,10 +1685,13 @@ sen_rc
 sen_rbuf_resize(sen_rbuf *buf, size_t newsize)
 {
   char *head;
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
+  newsize += sen_rbuf_margin_size + 1;
   newsize = (newsize + (UNIT_MASK)) & ~UNIT_MASK;
-  if (!(head = SEN_REALLOC(buf->head, newsize + 1))) { return sen_memory_exhausted; }
-  buf->curr = head + SEN_RBUF_VSIZE(buf);
-  buf->head = head;
+  head = buf->head - (buf->head ? sen_rbuf_margin_size : 0);
+  if (!(head = SEN_REALLOC(head, newsize))) { return sen_memory_exhausted; }
+  buf->curr = head + sen_rbuf_margin_size + SEN_RBUF_VSIZE(buf);
+  buf->head = head + sen_rbuf_margin_size;
   buf->tail = head + newsize;
   return sen_success;
 }
@@ -1813,13 +1716,20 @@ sen_rbuf_write(sen_rbuf *buf, const char *str, size_t len)
 }
 
 sen_rc
-sen_rbuf_space(sen_rbuf *buf, size_t len)
+sen_rbuf_reserve(sen_rbuf *buf, size_t len)
 {
   sen_rc rc = sen_success;
   if (SEN_RBUF_REST(buf) < len) {
     if ((rc = sen_rbuf_resize(buf, SEN_RBUF_VSIZE(buf) + len))) { return rc; }
   }
-  buf->curr += len;
+  return rc;
+}
+
+sen_rc
+sen_rbuf_space(sen_rbuf *buf, size_t len)
+{
+  sen_rc rc = sen_rbuf_reserve(buf, len);
+  if (!rc) { buf->curr += len; }
   return rc;
 }
 
@@ -1833,6 +1743,52 @@ sen_rbuf_itoa(sen_rbuf *buf, int i)
   return rc;
 }
 
+sen_rc
+sen_rbuf_lltoa(sen_rbuf *buf, int64_t i)
+{
+  sen_rc rc = sen_success;
+  while (sen_str_lltoa(i, buf->curr, buf->tail, &buf->curr)) {
+    if ((rc = sen_rbuf_resize(buf, SEN_RBUF_WSIZE(buf) + UNIT_SIZE))) { return rc; }
+  }
+  return rc;
+}
+
+sen_rc
+sen_rbuf_ftoa(sen_rbuf *buf, double d)
+{
+  size_t len = 32;
+  sen_rc rc = sen_success;
+  if (SEN_RBUF_REST(buf) < len) {
+    if ((rc = sen_rbuf_resize(buf, SEN_RBUF_VSIZE(buf) + len))) { return rc; }
+  }
+  switch (fpclassify(d)) {
+  CASE_FP_NAN
+    SEN_RBUF_PUTS(buf, "#<nan>");
+    break;
+  CASE_FP_INFINITE
+    SEN_RBUF_PUTS(buf, d > 0 ? "#i1/0" : "#i-1/0");
+    break;
+  default :
+    len = sprintf(buf->curr, "%#.15g", d);
+    if (buf->curr[len - 1] == '.') {
+      buf->curr += len;
+      SEN_RBUF_PUTC(buf, '0');
+    } else {
+      char *p, *q;
+      buf->curr[len] = '\0';
+      if ((p = strchr(buf->curr, 'e'))) {
+        for (q = p; *(q - 2) != '.' && *(q - 1) == '0'; q--) { len--; }
+        memmove(q, p, buf->curr + len - q);
+      } else {
+        for (q = buf->curr + len; *(q - 2) != '.' && *(q - 1) == '0'; q--) { len--; }
+      }
+      buf->curr += len;
+    }
+    break;
+  }
+  return rc;
+}
+
 sen_rc
 sen_rbuf_itoh(sen_rbuf *buf, int i)
 {
@@ -1892,8 +1848,9 @@ sen_rbuf_str_esc(sen_rbuf *buf, const char *s, int len, sen_encoding encoding)
 sen_rc
 sen_rbuf_fin(sen_rbuf *buf)
 {
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
   if (buf->head) {
-    SEN_REALLOC(buf->head, 0);
+    SEN_REALLOC(buf->head - sen_rbuf_margin_size, 0);
     buf->head = NULL;
   }
   return sen_success;
@@ -1916,6 +1873,7 @@ sen_lbuf_init(sen_lbuf *buf)
 void *
 sen_lbuf_add(sen_lbuf *buf, size_t size)
 {
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
   sen_lbuf_node *node = SEN_MALLOC(size + (size_t)(&((sen_lbuf_node *)0)->val));
   if (!node) { return NULL;  }
   node->next = NULL;
@@ -1928,6 +1886,7 @@ sen_lbuf_add(sen_lbuf *buf, size_t size)
 sen_rc
 sen_lbuf_fin(sen_lbuf *buf)
 {
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
   sen_lbuf_node *cur, *next;
   for (cur = buf->head; cur; cur = next) {
     next = cur->next;
