@@ -14,12 +14,12 @@
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
-#ifndef _SEN_STR_H
-#define _SEN_STR_H
+#ifndef SEN_STR_H
+#define SEN_STR_H
 
-#ifndef _SENNA_H
+#ifndef SENNA_H
 #include "senna_in.h"
-#endif /* _SENNA_H */
+#endif /* SENNA_H */
 
 #ifdef	__cplusplus
 extern "C" {
@@ -54,10 +54,11 @@ typedef struct {
   char *norm;
   size_t norm_blen;
   uint_least8_t *ctypes;
-  int_least8_t *checks;
+  int16_t *checks;
   size_t length;
   int flags;
-  sen_encoding encoding;
+  sen_ctx *ctx;
+  /* sen_encoding encoding; */
 } sen_nstr;
 
 typedef enum {
@@ -96,17 +97,13 @@ sen_rc sen_str_fin(void);
 #define SEN_NSTR_ISBLANK(c) (c & 0x80)
 #define SEN_NSTR_CTYPE(c) (c & 0x7f)
 
-void *sen_malloc(size_t size, const char* file, int line);
-void *sen_calloc(size_t size, const char* file, int line);
-void sen_free(void *ptr, const char* file, int line);
-void *sen_realloc(void *ptr, size_t size, const char* file, int line);
-char *sen_strdup(const char *s, const char* file, int line);
-void sen_assert(int cond, const char* file, int line, const char* func);
 int sen_isspace(const char *s, sen_encoding encoding);
 int sen_atoi(const char *nptr, const char *end, const char **rest);
 unsigned int sen_atoui(const char *nptr, const char *end, const char **rest);
 unsigned int sen_htoui(const char *nptr, const char *end, const char **rest);
+int64_t sen_atoll(const char *nptr, const char *end, const char **rest);
 sen_rc sen_str_itoa(int i, char *p, char *end, char **rest);
+sen_rc sen_str_lltoa(int64_t i, char *p, char *end, char **rest);
 const char *sen_enctostr(sen_encoding enc);
 sen_encoding sen_strtoenc(const char *str);
 
@@ -114,19 +111,10 @@ void sen_str_itoh(unsigned int i, char *p, unsigned int len);
 int sen_str_tok(char *str, size_t str_len, char delim, char **tokbuf, int buf_size, char **rest);
 int sen_str_getopt(int argc, char * const argv[], const sen_str_getopt_opt *opts, int *flags);
 
-#define SEN_MALLOC(s) sen_malloc(s,__FILE__,__LINE__)
-#define SEN_CALLOC(s) sen_calloc(s,__FILE__,__LINE__)
-#define SEN_REALLOC(p,s) sen_realloc(p,s,__FILE__,__LINE__)
-#define SEN_FREE(p) sen_free(p,__FILE__,__LINE__)
-#define SEN_STRDUP(s) sen_strdup(s,__FILE__,__LINE__)
-#ifdef DEBUG
-#define SEN_ASSERT(s) sen_assert((s),__FILE__,__LINE__,__FUNCTION__)
-#else
-#define SEN_ASSERT(s)
-#endif
-
 typedef struct _sen_rbuf sen_rbuf;
 
+extern int sen_rbuf_margin_size;
+
 struct _sen_rbuf {
   char *head;
   char *curr;
@@ -144,8 +132,11 @@ sen_rc sen_rbuf_init(sen_rbuf *buf, size_t size);
 sen_rc sen_rbuf_reinit(sen_rbuf *buf, size_t size);
 sen_rc sen_rbuf_resize(sen_rbuf *buf, size_t newsize);
 sen_rc sen_rbuf_write(sen_rbuf *buf, const char *str, size_t len);
+sen_rc sen_rbuf_reserve(sen_rbuf *buf, size_t len);
 sen_rc sen_rbuf_space(sen_rbuf *buf, size_t len);
 sen_rc sen_rbuf_itoa(sen_rbuf *buf, int i);
+sen_rc sen_rbuf_lltoa(sen_rbuf *buf, int64_t i);
+sen_rc sen_rbuf_ftoa(sen_rbuf *buf, double d);
 sen_rc sen_rbuf_itoh(sen_rbuf *buf, int i);
 sen_rc sen_rbuf_itob(sen_rbuf *buf, sen_id id);
 sen_rc sen_rbuf_fin(sen_rbuf *buf);
@@ -157,6 +148,7 @@ void sen_rbuf_str_esc(sen_rbuf *buf, const char *s, int len, sen_encoding encodi
 #define SEN_RBUF_WSIZE(buf) ((buf)->tail - (buf)->head)
 #define SEN_RBUF_REST(buf) ((buf)->tail - (buf)->curr)
 #define SEN_RBUF_VSIZE(buf) ((buf)->curr - (buf)->head)
+#define SEN_RBUF_EMPTYP(buf) ((buf)->curr == (buf)->head)
 
 sen_rc sen_lbuf_init(sen_lbuf *buf);
 void *sen_lbuf_add(sen_lbuf *buf, size_t size);
@@ -171,4 +163,4 @@ void sen_logger_fin(void);
 }
 #endif
 
-#endif /* _SEN_STR_H */
+#endif /* SEN_STR_H */
