@@ -253,11 +253,14 @@ sis_collect(sen_sym *sym, sen_set *h, sen_id id, uint32_t level)
   if (sl) {
     sen_id sid = sl->children;
     while (sid && sid != id) {
-      sen_set_get(h, &sid, (void **) &offset);
-      *offset = level;
-      if (level < MAX_LEVEL) { sis_collect(sym, h, sid, level + 1); }
-      if (!(sl = sis_at(sym, sid))) { break; }
-      sid = sl->sibling;
+      if (sen_set_get(h, &sid, (void **) &offset)) {
+        *offset = level;
+        if (level < MAX_LEVEL) { sis_collect(sym, h, sid, level + 1); }
+        if (!(sl = sis_at(sym, sid))) { break; }
+        sid = sl->sibling;
+      } else {
+        /* todo : must be handled */
+      }
     }
   }
 }
@@ -486,15 +489,17 @@ sen_sym_create(const char *path, uint32_t key_size,
 {
   int i;
   sen_io *io;
-  sen_sym *sym = NULL;
+  sen_sym *sym;
   pat_node *node0;
   struct sen_sym_header *header;
+  ERRCLR(NULL);
   if (flags & 0x70000) {
     return sen_sym_create08(path, key_size, flags, encoding);
   }
   io = sen_io_create(path, SEN_SYM_HEADER_SIZE, SEN_SYM_SEGMENT_SIZE,
                      SEN_SYM_MAX_SEGMENT, sen_io_auto, SEN_SYM_MAX_SEGMENT);
   if (!io) { return NULL; }
+  if (encoding == sen_enc_default) { encoding = sen_gctx.encoding; }
   header = sen_io_header(io);
   memcpy(header->idstr, SEN_IDSTR, 16);
   header->flags = flags;
@@ -512,7 +517,7 @@ sen_sym_create(const char *path, uint32_t key_size,
     header->patarray[i] = SEG_NOT_ASSIGNED;
     header->sisarray[i] = SEG_NOT_ASSIGNED;
   }
-  if (!(sym = SEN_MALLOC(sizeof(sen_sym)))) {
+  if (!(sym = SEN_GMALLOC(sizeof(sen_sym)))) {
     sen_io_close(io);
     return NULL;
   }
@@ -528,7 +533,11 @@ sen_sym_create(const char *path, uint32_t key_size,
     sym->pataddrs[i] = NULL;
     sym->sisaddrs[i] = NULL;
   }
-  node0 = pat_get(sym, 0);
+  if (!(node0 = pat_get(sym, 0))) {
+    sen_io_close(io);
+    SEN_GFREE(sym);
+    return NULL;
+  }
   node0->r = 0;
   node0->l = 0;
   node0->key = 0;
@@ -540,8 +549,9 @@ sen_sym_open(const char *path)
 {
   int i;
   sen_io *io;
-  sen_sym *sym = NULL;
+  sen_sym *sym;
   struct sen_sym_header *header;
+  ERRCLR(NULL);
   io = sen_io_open(path, sen_io_auto, 8192);
   if (!io) { return NULL; }
   header = sen_io_header(io);
@@ -550,7 +560,7 @@ sen_sym_open(const char *path)
     sen_io_close(io);
     return sen_sym_open08(path);
   }
-  if (!(sym = SEN_MALLOC(sizeof(sen_sym)))) {
+  if (!(sym = SEN_GMALLOC(sizeof(sen_sym)))) {
     sen_io_close(io);
     return NULL;
   }
@@ -566,6 +576,11 @@ sen_sym_open(const char *path)
     sym->pataddrs[i] = NULL;
     sym->sisaddrs[i] = NULL;
   }
+  if (!pat_at(sym, 0)) {
+    sen_io_close(io);
+    SEN_GFREE(sym);
+    return NULL;
+  }
   return sym;
 }
 
@@ -575,13 +590,14 @@ sen_sym_close(sen_sym *sym)
   sen_rc rc;
   if (!sym) { return sen_invalid_argument; }
   rc = sen_io_close(sym->io);
-  SEN_FREE(sym);
+  SEN_GFREE(sym);
   return rc;
 }
 
 sen_rc
 sen_sym_remove(const char *path)
 {
+  ERRCLR(NULL);
   if (!path) { return sen_invalid_argument; }
   return sen_io_remove(path);
 }
@@ -601,19 +617,22 @@ _sen_sym_get(sen_sym *sym, const uint8_t *key, uint32_t *new, uint32_t *lkey)
   if (*p0) {
     char xor, mask;
     const uint8_t *s, *d;
-    while ((r0 = *p0)) {
+    for (;;) {
+      if (!(r0 = *p0)) {
+        if (!(s = pat_node_get_key(sym, rn0))) { return 0; }
+        break;
+      }
       if (!(rn0 = pat_at(sym, r0))) { return 0; }
       if (c0 < rn0->check && rn0->check < len) {
         c1 = c0; c0 = rn0->check;
         p1 = p0; p0 = nth_bit(key, c0) ? &rn0->r : &rn0->l;
       } else {
-        if (rn0->check < len && !memcmp(pat_node_get_key(sym, rn0), key, size)) {
-          return r0;
-        }
+        if (!(s = pat_node_get_key(sym, rn0))) { return 0; }
+        if (rn0->check < len && !memcmp(s, key, size)) { return r0; }
         break;
       }
     }
-    for (c = 0, s = pat_node_get_key(sym, rn0), d = key; *s == *d; c += 8, s++, d++);
+    for (c = 0, d = key; *s == *d; c += 8, s++, d++);
     for (xor = *s ^ *d, mask = 0x80; !(xor & mask); mask >>= 1, c++);
     /* for test
 e    if (r0 && r0 != sen_sym_at(sym, pat_node_get_key(sym, rn0))) {
@@ -646,7 +665,7 @@ e    if (r0 && r0 != sen_sym_at(sym, pat_node_get_key(sym, rn0))) {
       if (sym->header->garbages[0]) {
         r = sym->header->garbages[0];
         sym->header->nrecords++;
-        rn = pat_at(sym, r);
+        if (!(rn = pat_at(sym, r))) { return 0; }
         sym->header->garbages[0] = rn->l;
       } else {
         if (!(rn = pat_node_new(sym, &r))) { return 0; }
@@ -655,11 +674,13 @@ e    if (r0 && r0 != sen_sym_at(sym, pat_node_get_key(sym, rn0))) {
       rn->key = *lkey;
     } else {
       if (sym->header->garbages[size2]) {
+        uint8_t *keybuf;
         r = sym->header->garbages[size2];
         sym->header->nrecords++;
-        rn = pat_at(sym, r);
+        if (!(rn = pat_at(sym, r))) { return 0; }
         sym->header->garbages[size2] = rn->l;
-        memcpy(pat_node_get_key(sym, rn), key, size);
+        if (!(keybuf = pat_node_get_key(sym, rn))) { return 0; }
+        memcpy(keybuf, key, size);
       } else {
         if (!(rn = pat_node_new(sym, &r))) { return 0; }
         pat_node_set_key(sym, rn, key, (unsigned int)size);
@@ -700,10 +721,11 @@ sen_sym_get(sen_sym *sym, const void *key)
 {
   uint32_t new, lkey = 0;
   sen_id r0;
-  if (!key) { return SEN_SYM_NIL; }
+  if (!sym || !key) { return SEN_SYM_NIL; }
   if (sym->v08p) { return sen_sym_get08(sym, key); }
   r0 = _sen_sym_get(sym, (uint8_t *)key, &new, &lkey);
-  if ((sym->flags & SEN_SYM_WITH_SIS) && (*((uint8_t *)key) & 0x80)) { // todo: refine!!
+  if (r0 && (sym->flags & SEN_SYM_WITH_SIS) &&
+      (*((uint8_t *)key) & 0x80)) { // todo: refine!!
     sis_node *sl, *sr;
     sen_id l = r0, r;
     if (new && (sl = sis_get(sym, l))) {
@@ -712,7 +734,7 @@ sen_sym_get(sen_sym *sym, const void *key)
       sl->sibling = 0;
       while (chop(sym, &sis, &lkey)) {
         if (!(*sis & 0x80)) { break; }
-        r = _sen_sym_get(sym, (uint8_t *)sis, &new, &lkey);
+        if (!(r = _sen_sym_get(sym, (uint8_t *)sis, &new, &lkey))) { break; }
         if (!(sr = sis_get(sym, r))) { break; }
         if (new) {
           sl->sibling = r;
@@ -737,15 +759,18 @@ sen_sym_at(sen_sym *sym, const void *key)
   sen_id r;
   pat_node *rn;
   int c = -1;
-  size_t len = sym->key_size * 8;
+  size_t len;
+  if (!sym || !key) { return SEN_SYM_NIL; }
+  len = sym->key_size * 8;
   if (sym->v08p) { return sen_sym_at08(sym, key); }
-  if (!key) { return SEN_SYM_NIL; }
   if (!len) { len = (strlen(key) + 1) * 8; }
   for (r = pat_at(sym, 0)->r; r; r = nth_bit((uint8_t *)key, c) ? rn->r : rn->l) {
     if (!(rn = pat_at(sym, r))) { break; /* corrupt? */ }
     if (len <= rn->check) { break; }
     if (rn->check <= c) {
-      if (!memcmp(pat_node_get_key(sym, rn), key, len >> 3)) { return r; }
+      const uint8_t *k = pat_node_get_key(sym, rn);
+      if (!k) { break; }
+      if (!memcmp(k, key, len >> 3)) { return r; }
       break;
     }
     c = rn->check;
@@ -753,6 +778,20 @@ sen_sym_at(sen_sym *sym, const void *key)
   return SEN_SYM_NIL;
 }
 
+sen_id
+sen_sym_nextid(sen_sym *sym, const void *key)
+{
+  sen_id r = SEN_SYM_NIL;
+  if (sym && key) {
+    size_t size = sym->key_size;
+    if (!size) { size = strlen((char *)key) + 1; }
+    if (!(r = sym->header->garbages[size > sizeof(uint32_t) ? size : 0])) {
+      r = sym->header->curr_rec + 1;
+    }
+  }
+  return r;
+}
+
 static void
 get_tc(sen_sym *sym, sen_set *h, pat_node *rn, unsigned int len)
 {
@@ -782,6 +821,7 @@ sen_rc
 sen_sym_prefix_search_with_set(sen_sym *sym, const void *key, sen_set *h)
 {
   int c = -1;
+  const uint8_t *k;
   size_t len = strlen(key) * 8;
   pat_node *rn = pat_at(sym, 0);
   sen_id r = rn->r;
@@ -792,26 +832,29 @@ sen_sym_prefix_search_with_set(sen_sym *sym, const void *key, sen_set *h)
       r = nth_bit((uint8_t *)key, c) ? rn->r : rn->l;
       continue;
     }
-    if (!memcmp(pat_node_get_key(sym, rn), key, len >> 3)) {
+    if (!(k = pat_node_get_key(sym, rn))) { break; }
+    if (!memcmp(k, key, len >> 3)) {
       sen_set_get(h, &r, NULL);
       if (rn->check >= len) { get_tc(sym, h, rn, (unsigned int)len); }
       return sen_success;
     }
     break;
   }
-  return sen_other_error;
+  return sen_internal_error;
 }
 
 sen_set *
 sen_sym_prefix_search(sen_sym *sym, const void *key)
 {
   sen_set *h;
+  ERRCLR(NULL);
+  if (!sym || !key || sym->key_size) { return NULL; } /* only for string */
   if (sym->v08p) { return sen_sym_prefix_search08(sym, key); }
-  if (!key || sym->key_size) { return NULL; } /* only for string */
-  h = sen_set_open(sizeof(sen_id), 0, 0);
-  if (sen_sym_prefix_search_with_set(sym, key, h)) {
-    sen_set_close(h);
-    h = NULL;
+  if ((h = sen_set_open(sizeof(sen_id), 0, 0))) {
+    if (sen_sym_prefix_search_with_set(sym, key, h)) {
+      sen_set_close(h);
+      h = NULL;
+    }
   }
   return h;
 }
@@ -821,27 +864,26 @@ sen_sym_suffix_search_with_set(sen_sym *sym, const void *key, sen_set *h)
 {
   sen_id r;
   uint32_t *offset;
-  if ((r = sen_sym_at(sym, key))) {
-    sen_set_get(h, &r, (void **) &offset);
+  if ((r = sen_sym_at(sym, key)) && sen_set_get(h, &r, (void **) &offset)) {
     *offset = 0;
     if (sym->flags & SEN_SYM_WITH_SIS) { sis_collect(sym, h, r, 1); }
     return sen_success;
   }
-  return sen_other_error;
+  return sen_internal_error;
 }
 
 sen_set *
 sen_sym_suffix_search(sen_sym *sym, const void *key)
 {
   sen_set *h;
+  ERRCLR(NULL);
+  if (!sym || !key || sym->key_size) { return NULL; }
   if (sym->v08p) { return sen_sym_suffix_search08(sym, key); }
-  if (!key || sym->key_size) { return NULL; }
-  h = sen_set_open(sizeof(sen_id), sizeof(uint32_t), 0);
-  SEN_ASSERT(h);
-//    if (!h) { return NULL; }
-  if (sen_sym_suffix_search_with_set(sym, key, h)) {
-    sen_set_close(h);
-    h = NULL;
+  if ((h = sen_set_open(sizeof(sen_id), sizeof(uint32_t), 0))) {
+    if (sen_sym_suffix_search_with_set(sym, key, h)) {
+      sen_set_close(h);
+      h = NULL;
+    }
   }
   return h;
 }
@@ -852,15 +894,15 @@ sen_sym_common_prefix_search(sen_sym *sym, const void *key)
   pat_node *rn, *rnl;
   sen_id r, rl, r2 = SEN_SYM_NIL;
   int c = -1, cl, len = 0, n, len2 = 8, nth_bit;
+  if (!sym || !key || sym->key_size) { return SEN_SYM_NIL; } /* only for string */
   if (sym->v08p) { return sen_sym_common_prefix_search08(sym, key); }
-  if (!key) { return SEN_SYM_NIL; }
-  if (sym->key_size) { return SEN_SYM_NIL; /* only for string */ }
   // len = (strlen(key) + 1) * 8;
   for (r = pat_at(sym, 0)->r; r;) {
     if (!(rn = pat_at(sym, r))) { break; /* corrupt? */ }
     // if (len <= rn->check) { break; }
     if (rn->check <= c) {
       uint8_t *p = pat_node_get_key(sym, rn);
+      if (!p) { break; }
       if (!memcmp(p, key, strlen((char *)p))) { return r; }
       break;
     }
@@ -876,6 +918,7 @@ sen_sym_common_prefix_search(sen_sym *sym, const void *key)
           if (len2 <= rnl->check) { break; }
           if (rnl->check <= cl) {
             uint8_t *p = pat_node_get_key(sym, rnl);
+            if (!p) { break; }
             if (!p[n] && !memcmp(p, key, n)) { r2 = rl; }
             break;
           }
@@ -911,7 +954,9 @@ _sen_sym_del(sen_sym *sym, const char *key, int shared)
     if (!(rn = pat_at(sym, r))) { return sen_invalid_format; }
     if (len <= rn->check) { return sen_invalid_argument; }
     if (c >= rn->check) {
-      if (!memcmp(pat_node_get_key(sym, rn), key, len >> 3)) {
+      const uint8_t *k = pat_node_get_key(sym, rn);
+      if (!k) { return sen_invalid_argument; }
+      if (!memcmp(k, key, len >> 3)) {
         break; /* found */
       } else {  return sen_invalid_argument; }
     }
@@ -1005,13 +1050,25 @@ _sen_sym_del(sen_sym *sym, const char *key, int shared)
 sen_rc
 sen_sym_del(sen_sym *sym, const void *key)
 {
+  if (!sym) { return sen_invalid_argument; }
   if (sym->v08p) { return sen_sym_del08(sym, key); }
+  if (sym->flags & SEN_SYM_WITH_SIS) {
+    SEN_LOG(sen_log_warning, "use sen_sym_del_with_sis for sym with SEN_SYM_WITH_SIS");
+  }
+  if (sym->flags & SEN_INDEX_SHARED_LEXICON) {
+    sen_id id = sen_sym_at(sym, key);
+    if (id) {
+      pat_node *rn = pat_at(sym, id);
+      if (rn && rn->bitfield >= 1 << 2) { return sen_success; }
+    }
+  }
   return _sen_sym_del(sym, key, 0);
 }
 
 uint32_t
 sen_sym_size(sen_sym *sym)
 {
+  if (!sym) { return sen_invalid_argument; }
   return sym->header->nrecords;
 }
 
@@ -1031,6 +1088,7 @@ sen_sym_key(sen_sym *sym, sen_id id, void *keybuf, int bufsize)
   int len;
   uint8_t *key;
   pat_node *node;
+  if (!sym) { return sen_invalid_argument; }
   if (sym->v08p) { return sen_sym_key08(sym, id, keybuf, bufsize); }
   if (!(node = pat_at(sym, id))) { return 0; }
   if (!(key = pat_node_get_key(sym, node))) { return 0; }
@@ -1043,6 +1101,7 @@ int
 sen_sym_pocket_get(sen_sym *sym, sen_id id)
 {
   pat_node *node;
+  if (!sym) { return -1; }
   if (sym->v08p) { return sen_sym_pocket_get08(sym, id); }
   node = pat_at(sym, id);
   if (!node) { return -1; }
@@ -1053,6 +1112,7 @@ sen_rc
 sen_sym_pocket_set(sen_sym *sym, sen_id id, unsigned int value)
 {
   pat_node *node;
+  if (!sym) { return sen_invalid_argument; }
   if (sym->v08p) { return sen_sym_pocket_set08(sym, id, value); }
   node = pat_at(sym, id);
   if (value > PAT_POCKET_MAX || !node) { return sen_invalid_argument; }
@@ -1060,10 +1120,43 @@ sen_sym_pocket_set(sen_sym *sym, sen_id id, unsigned int value)
   return sen_success;
 }
 
+sen_rc
+sen_sym_pocket_incr(sen_sym *sym, sen_id id)
+{
+  pat_node *node;
+  if (!sym) { return sen_invalid_argument; }
+  if (sym->v08p) { return sen_invalid_argument; }
+  node = pat_at(sym, id);
+  if (!node) { return sen_invalid_argument; }
+  node->bitfield += (1 << 2);
+  if (node->bitfield < (1 << 2)) {
+    SEN_LOG(sen_log_error, "sen_sym_pocket_incr failed %d", node->bitfield);
+    return sen_invalid_format;
+  }
+  return sen_success;
+}
+
+sen_rc
+sen_sym_pocket_decr(sen_sym *sym, sen_id id)
+{
+  pat_node *node;
+  if (!sym) { return sen_invalid_argument; }
+  if (sym->v08p) { return sen_invalid_argument; }
+  node = pat_at(sym, id);
+  if (!node) { return sen_invalid_argument; }
+  if (node->bitfield < (1 << 2)) {
+    SEN_LOG(sen_log_error, "sen_sym_pocket_decr failed %d", node->bitfield);
+    return sen_invalid_format;
+  }
+  node->bitfield -= (1 << 2);
+  return sen_success;
+}
+
 sen_rc
 sen_sym_info(sen_sym *sym, int *key_size, unsigned *flags,
              sen_encoding *encoding, unsigned *nrecords, unsigned *file_size)
 {
+  ERRCLR(NULL);
   if (!sym) { return sen_invalid_argument; }
   if (key_size) { *key_size = sym->key_size; }
   if (flags) { *flags = sym->flags; }
@@ -1089,7 +1182,12 @@ sen_sym_del_with_sis(sen_sym *sym, sen_id id,
   sis_node *sp, *ss = NULL, *si = sis_at(sym, id);
   if (sym->v08p) { return sen_sym_del_with_sis08(sym, id, func, func_arg); }
   while (id) {
+    if (sym->flags & SEN_INDEX_SHARED_LEXICON) {
+      pat_node *rn = pat_at(sym, id);
+      if (rn && rn->bitfield >= 1 << 2) { break; }
+    }
     if ((si && si->children && si->children != id) || !func(id, func_arg)) { break; }
+    //todo : if ERRP { break; }
     if (!(_key = _sen_sym_key(sym, id))) { break; }
     if (_key == key) {
       shared = 1;
@@ -1108,7 +1206,7 @@ sen_sym_del_with_sis(sen_sym *sym, sen_id id,
               *p = si->sibling;
               break;
             }
-            sp = sis_at(sym, *p);
+            if (!(sp = sis_at(sym, *p))) { break; }
           }
         }
       } else {
@@ -1129,6 +1227,7 @@ sen_sym_del_with_sis(sen_sym *sym, sen_id id,
       if (_sen_sym_key(sym, id) != key) { break; }
       {
         pat_node *rn = pat_at(sym, id);
+        if (!rn) { break; }
         if (lkey) {
           rn->key = lkey;
         } else {
@@ -1159,3 +1258,45 @@ sen_sym_curr_id(sen_sym *sym)
 {
   return sym->header->curr_rec;
 }
+
+sen_rc
+sen_sym_lock(sen_sym *sym, int timeout)
+{
+  static int _ncalls = 0, _ncolls = 0;
+  uint32_t count;
+  _ncalls++;
+  for (count = 0;; count++) {
+    uint32_t lock;
+    SEN_ATOMIC_ADD_EX(sym->lock, 1, lock);
+    if (lock) {
+      SEN_ATOMIC_ADD_EX(sym->lock, -1, lock);
+      if (!timeout || (timeout > 0 && timeout == count)) { break; }
+      if (!(++_ncolls % 1000000) && (_ncolls > _ncalls)) {
+        if (_ncolls < 0 || _ncalls < 0) {
+          _ncolls = 0; _ncalls = 0;
+        } else {
+          SEN_LOG(sen_log_notice, "sym(%p) collisions(%d/%d)", sym, _ncolls, _ncalls);
+        }
+      }
+      usleep(1000);
+      continue;
+    }
+    return sen_success;
+  }
+  return sen_abnormal_error;
+}
+
+sen_rc
+sen_sym_unlock(sen_sym *sym)
+{
+  uint32_t lock;
+  SEN_ATOMIC_ADD_EX(sym->lock, -1, lock);
+  return sen_success;
+}
+
+sen_rc
+sen_sym_clear_lock(sen_sym *sym)
+{
+  *sym->lock = 0;
+  return sen_success;
+}
