@@ -22,22 +22,22 @@
 #include <stdio.h>
 #include <signal.h>
 #include <string.h>
-#include <errno.h>
 #include <setjmp.h>
 #ifdef HAVE_UCONTEXT_H
 #include <ucontext.h>
 #endif /* HAVE_UCONTEXT_H */
 
 #ifdef WIN32
+#include <errno.h>
 #define sigsetjmp(x,y) setjmp((x))
 #define siglongjmp(x,y) longjmp((x), (y))
 typedef jmp_buf sigjmp_buf;
-#endif
+#endif /* WIN32 */
 
 /* TODO: export sen_index_clear_lock */
 sen_rc sen_index_clear_lock(sen_index *i);
 
-#define MAX_CHECK_FILES 256
+#define INITIAL_CHECK_FILES 128
 #define MAX_NSEGV 32
 #define MAX_NERR_FILE 8
 
@@ -67,8 +67,8 @@ typedef struct {
 
 typedef struct {
   senchk_file *files;
-  int n_files;
-  int max_n_files;
+  unsigned int n_files;
+  unsigned int max_n_files;
 } senchk_filelist;
 
 /* globals */
@@ -262,89 +262,89 @@ check_inv(senchk_file *f)
     if (chkflags & CHK_UNLOCK) {
       printf("  unlock inv file: '%s'\n", f->path);
       sen_index_clear_lock(i);
-    }
-
-    inv = i->inv;
-    if (chkflags & CHK_DUMP_LEXICON) {
-      puts("  Inv records infomation:\n"
-           "         tid,        df,        sf,    nposts | term");
-    }
-    while ((tid = sen_sym_next(lex, tid)) != SEN_SYM_NIL && f->cerr < MAX_NERR_FILE) {
-      if (!(c % 1000)) {
-        fprintf(stderr, "entries: %d\r", c);
+    } else {
+      inv = i->inv;
+      if (chkflags & CHK_DUMP_LEXICON) {
+        puts("  Inv records infomation:\n"
+             "         tid,        df,        sf,    nposts | term");
       }
-      if (!(cur = sen_inv_cursor_open(inv, tid, 1))) {
-        /* TODO: detect array_at fails or not
-        printf("cannot open inv cursor '%s'\n", f->path);
-        f->cerr++;
-        */
-      } else {
-        sen_inv_posting pre = {0, 0, 0, 0, 0, 0};
-        unsigned int df = 0, sf = 0, nposts = 0;
-        while (!sen_inv_cursor_next(cur) && f->cerr < MAX_NERR_FILE) {
-          sen_inv_posting *post = cur->post;
-          sen_id max_rid = sen_sym_curr_id(i->keys);
-          if (pre.rid > post->rid || post->rid > max_rid) {
-            printf("invalid rid: pre-rid:%d rid:%d max_rid:%d\n",
-                   pre.rid, post->rid, max_rid);
-            f->cerr++;
-          } else if (pre.rid == post->rid) {
-            /* TODO: get max_sid and compare it */
-            if (pre.sid >= post->sid) {
-              printf("invalid sid: rid:%d pre-sid:%d sid:%d\n",
-                     post->rid, pre.sid, post->sid);
+      while ((tid = sen_sym_next(lex, tid)) != SEN_SYM_NIL && f->cerr < MAX_NERR_FILE) {
+        if (!(c % 1000)) {
+          fprintf(stderr, "entries: %d\r", c);
+        }
+        if (!(cur = sen_inv_cursor_open(inv, tid, 1))) {
+          /* TODO: detect array_at fails or not
+          printf("cannot open inv cursor '%s'\n", f->path);
+          f->cerr++;
+          */
+        } else {
+          sen_inv_posting pre = {0, 0, 0, 0, 0, 0};
+          unsigned int df = 0, sf = 0, nposts = 0;
+          while (!sen_inv_cursor_next(cur) && f->cerr < MAX_NERR_FILE) {
+            sen_inv_posting *post = cur->post;
+            sen_id max_rid = sen_sym_curr_id(i->keys);
+            if (pre.rid > post->rid || post->rid > max_rid) {
+              printf("invalid rid: pre-rid:%d rid:%d max_rid:%d\n",
+                     pre.rid, post->rid, max_rid);
               f->cerr++;
+            } else if (pre.rid == post->rid) {
+              /* TODO: get max_sid and compare it */
+              if (pre.sid >= post->sid) {
+                printf("invalid sid: rid:%d pre-sid:%d sid:%d\n",
+                       post->rid, pre.sid, post->sid);
+                f->cerr++;
+              } else {
+                sf++;
+              }
             } else {
+              char key[SEN_SYM_MAX_KEY_SIZE];
+              if (!sen_sym_key(keys, post->rid, key, SEN_SYM_MAX_KEY_SIZE)) {
+                printf("looking up key from keys failed.\n"
+                       " rid: %d\n", post->rid);
+                f->cerr++;
+              }
+              df++;
               sf++;
             }
-          } else {
-            char key[SEN_SYM_MAX_KEY_SIZE];
-            if (!sen_sym_key(keys, post->rid, key, SEN_SYM_MAX_KEY_SIZE)) {
-              printf("looking up key from keys failed.\n"
-                     " rid: %d\n", post->rid);
-              f->cerr++;
+            pre.pos = 0;
+            while (!sen_inv_cursor_next_pos(cur) && f->cerr < MAX_NERR_FILE) {
+              post = cur->post;
+              if (pre.pos && pre.pos >= post->pos) {
+                printf("invalid position: rid:%d sid:%d pre-pos:%d pos:%d\n",
+                       post->rid, post->sid, pre.pos, post->pos);
+                f->cerr++;
+              } else if (!post->tf) {
+                printf("invalid tf: rid:%d sid:%d pos:%d tf:0\n",
+                       post->rid, post->sid, post->pos);
+                f->cerr++;
+              }
+              nposts++;
+              pre = *post;
             }
-            df++;
-            sf++;
-          }
-          pre.pos = 0;
-          while (!sen_inv_cursor_next_pos(cur) && f->cerr < MAX_NERR_FILE) {
-            post = cur->post;
-            if (pre.pos && pre.pos >= post->pos) {
-              printf("invalid position: rid:%d sid:%d pre-pos:%d pos:%d\n",
-                     post->rid, post->sid, pre.pos, post->pos);
-              f->cerr++;
-            } else if (!post->tf) {
-              printf("invalid tf: rid:%d sid:%d pos:%d tf:0\n",
-                     post->rid, post->sid, post->pos);
-              f->cerr++;
-            }
-            nposts++;
             pre = *post;
           }
-          pre = *post;
-        }
-        if (chkflags & CHK_DUMP_LEXICON) {
-          char term[SEN_SYM_MAX_KEY_SIZE];
-          if (!sen_sym_key(lex, tid, term, SEN_SYM_MAX_KEY_SIZE)) {
-            printf("looking up term from lexicon failed.\n"
-                   " termid: %d\n", tid);
-            f->cerr++;
-          } else {
-            /* FIXME: check lexicon size and print non-null terminated fix-size term */
-            printf("  %10d,%10d,%10d,%10d | %s\n", tid, df, sf, nposts, term);
+          if (chkflags & CHK_DUMP_LEXICON) {
+            char term[SEN_SYM_MAX_KEY_SIZE];
+            if (!sen_sym_key(lex, tid, term, SEN_SYM_MAX_KEY_SIZE)) {
+              printf("looking up term from lexicon failed.\n"
+                     " termid: %d\n", tid);
+              f->cerr++;
+            } else {
+              /* FIXME: check lexicon size and print non-null terminated fix-size term */
+              printf("  %10d,%10d,%10d,%10d | %s\n", tid, df, sf, nposts, term);
+            }
           }
+          tdf += df;
+          tsf += sf;
+          tnposts += nposts;
+          sen_inv_cursor_close(cur);
         }
-        tdf += df;
-        tsf += sf;
-        tnposts += nposts;
-        sen_inv_cursor_close(cur);
+        c++;
       }
-      c++;
+      printf(" checked inv file %s: total %d entiries\n"
+             " checked: %d records/%d sections/%d posts\n", f->path, c, tdf, tsf, tnposts);
+      e = f->cerr;
     }
-    printf(" checked inv file %s: total %d entiries\n"
-           " checked: %d records/%d sections/%d posts\n", f->path, c, tdf, tsf, tnposts);
-    e = f->cerr;
   } else {
     /* handle segv */
     e = -4;
@@ -379,7 +379,7 @@ cat_pathname(const char *path, const char *ext)
     size_t path_len = strlen(path);
     size_t ext_len = strlen(ext);
     if (path_len + ext_len >= PATH_MAX) { return NULL; }
-    if ((buf = SEN_MALLOC(path_len + ext_len + 1))) {
+    if ((buf = SEN_GMALLOC(path_len + ext_len + 1))) {
       memcpy(buf, path, path_len);
       memcpy(buf + path_len, ext, ext_len + 1);
     }
@@ -414,7 +414,7 @@ check_file_existence(const char *path, const char *ext)
   fclose(fp);
   ret = 1;
 exit:
-  if (ext) { SEN_FREE((char *)buf); }
+  if (ext) { SEN_GFREE((char *)buf); }
   return ret;
 }
 
@@ -422,17 +422,33 @@ senchk_filelist *
 senchk_filelist_init(unsigned int max_n_files)
 {
   senchk_filelist *fl;
-  if ((fl = SEN_MALLOC(sizeof(senchk_filelist)))) {
-    if ((fl->files = SEN_CALLOC(max_n_files * sizeof(senchk_file)))) {
+  if ((fl = SEN_GMALLOC(sizeof(senchk_filelist)))) {
+    if ((fl->files = SEN_GMALLOC(max_n_files * sizeof(senchk_file)))) {
       fl->n_files = 0;
       fl->max_n_files = max_n_files;
       return fl;
     }
-    SEN_FREE(fl);
+    SEN_GFREE(fl);
   }
   return NULL;
 }
 
+inline static void
+senchk_filelist_expand(senchk_filelist *fl)
+{
+  senchk_file *files;
+  if (fl->n_files < fl->max_n_files) { return; }
+  fl->max_n_files <<= 1;
+  if (!(files = SEN_GREALLOC(fl->files,
+                             fl->max_n_files * sizeof(senchk_file)))) {
+    puts("memory realloction failed: on senchk_filelist_expand");
+    fl->max_n_files >>= 1;
+    fl->n_files = 0; /* overwrite */
+    return;
+  }
+  fl->files = files;
+}
+
 int
 senchk_filelist_add(senchk_filelist *fl, senchk_filetype type,
                     const char *path, const char *keyspath, const char *lexpath)
@@ -448,6 +464,7 @@ senchk_filelist_add(senchk_filelist *fl, senchk_filetype type,
     }
     f->path = strdup(path);
     fl->n_files++;
+    senchk_filelist_expand(fl);
     return 1;
   }
   return 0;
@@ -456,7 +473,7 @@ senchk_filelist_add(senchk_filelist *fl, senchk_filetype type,
 int
 senchk_filelist_addi(senchk_filelist *fl, senchk_filetype type, void *instance)
 {
-  if (fl && instance && fl->max_n_files > fl->n_files) {
+  if (fl && instance) {
     senchk_file *f = fl->files + fl->n_files;
     f->type = type;
     f->keyspath = f->lexpath = NULL;
@@ -491,6 +508,7 @@ senchk_filelist_addi(senchk_filelist *fl, senchk_filetype type, void *instance)
         break;
     }
     fl->n_files++;
+    senchk_filelist_expand(fl);
     return 1;
   }
   return 0;
@@ -503,11 +521,11 @@ senchk_filelist_fin(senchk_filelist *fl)
   if (!fl) { return; }
   end = fl->files + fl->n_files;
   for (f = fl->files; f < end; f++) {
-    if (f->path) { SEN_FREE(f->path); }
-    if (f->keyspath) { SEN_FREE(f->keyspath); }
-    if (f->lexpath) { SEN_FREE(f->lexpath); }
+    if (f->path) { SEN_GFREE(f->path); }
+    if (f->keyspath) { SEN_GFREE(f->keyspath); }
+    if (f->lexpath) { SEN_GFREE(f->lexpath); }
   }
-  SEN_FREE(fl);
+  SEN_GFREE(fl);
 }
 
 int
@@ -609,33 +627,35 @@ senchk_filelist_check(senchk_filelist *fl)
   if (!fl) { return -1; }
   end = fl->files + fl->n_files;
 
-  /* sym/ra/ja check */
-  for (f = fl->files; f < end; f++) {
-    switch (f->type) {
-      case type_sym:
-        if (check_sym(f)) {
-          ngcount++;
-          printf("* BROKEN SYM FILE: '%s'\n", f->path);
-        }
-        break;
-      case type_ra:
-        /* TODO: implement */
-        show_delimiter(stdout);
-        printf(" check ra file: '%s'\n", f->path);
-        puts("* now checking ra is not implemented...");
-        break;
-      case type_ja:
-        /* TODO: implement */
-        show_delimiter(stdout);
-        printf(" check ja file: '%s'\n", f->path);
-        puts("* now checking ja is not implemented...");
-        break;
-      case type_inv:
-        /* do nothing */
-        break;
-      default:
-        /* TODO: print error */
-        break;
+  if (!(chkflags & CHK_UNLOCK)) {
+    /* sym/ra/ja check */
+    for (f = fl->files; f < end; f++) {
+      switch (f->type) {
+        case type_sym:
+          if (check_sym(f)) {
+            ngcount++;
+            printf("* BROKEN SYM FILE: '%s'\n", f->path);
+          }
+          break;
+        case type_ra:
+          /* TODO: implement */
+          show_delimiter(stdout);
+          printf(" check ra file: '%s'\n", f->path);
+          puts("* now checking ra is not implemented...");
+          break;
+        case type_ja:
+          /* TODO: implement */
+          show_delimiter(stdout);
+          printf(" check ja file: '%s'\n", f->path);
+          puts("* now checking ja is not implemented...");
+          break;
+        case type_inv:
+          /* do nothing */
+          break;
+        default:
+          /* TODO: print error */
+          break;
+      }
     }
   }
 
@@ -736,7 +756,7 @@ main(int argc, char *argv[])
   show_delimiter(stdout);
   show_seninfo(stdout);
 
-  if (!(fls = senchk_filelist_init(MAX_CHECK_FILES))) {
+  if (!(fls = senchk_filelist_init(INITIAL_CHECK_FILES))) {
     puts("cannot init filelist object");
     ret = 3;
     goto exit;
